/* Copyright 2005-2010 Edscott Wilson Garca. 
 * Distributed with GPL licence.*/

/* basic callback functions used to set up the iconview 
 *
 * */

static void
adjustment_changed (
    GtkAdjustment * adjustment,
    gpointer data
) {
    view_t *view_p = (view_t *) data;
    DBG ("geometry: adjustment_changed to %lf\n",
         gtk_adjustment_get_value (gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window)));

    rodent_done_with_rename(data);

    rodent_save_local_view_geometry_p (view_p);
    rodent_unsaturate_pixbuf (view_p);
    rodent_label_event (view_p, NULL);
}

static gboolean
invisible_bottom (
    view_t * view_p
) {
    int totally_invisible_items;
    int partially_invisible_items;
    int invisible_items = 0;
    gdouble sp = gtk_adjustment_get_value (gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window));

    totally_invisible_items = ((int) (sp) / CELLWIDTH) * view_p->grid_columns;
    partially_invisible_items = (((int) (sp) % CELLWIDTH) / view_p->icon_size) * view_p->grid_columns;

    TRACE ("partially=%d, totally=%d, invisible=%d\n", partially_invisible_items, totally_invisible_items, invisible_items);

    invisible_items = totally_invisible_items + partially_invisible_items;
    return invisible_items;
}

static gboolean
invisible_top (
    view_t * view_p
) {
    int invisible_items;
    gdouble sp = gtk_adjustment_get_value (gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window));

    invisible_items = ((int) (sp + view_p->widgets.window->allocation.height) / CELLHEIGHT) * view_p->grid_columns;

    return invisible_items;
}

static gboolean
signal_destroy_event (
    GtkWidget * window,
    GdkEvent * event,
    gpointer data
) {
    DBG (">> signal_destroy_event\n");
    view_t *view_p = rodent_get_current_view (window);
    rodent_destroy_view (view_p);
    return TRUE;
}

static gboolean
signal_on_size_paper (
    GtkWidget * widget,
    GtkAllocation * allocation,
    gpointer user_data
) {
    DBG (">> geometry: signal_on_size_paper\n");
    view_t *view_p = (view_t *) user_data;
    TRACE ("PPP on_size_paper::size allocation event for paper********...\n");
    rodent_on_size (view_p, widget->allocation.width, widget->allocation.height);
    rodent_save_root_view_geometry_p (view_p);
    rodent_restore_view_size (view_p);
    TRACE ("PPP on_size_paper:: done\n");
    return TRUE;
}

static gboolean
signal_on_configure_paper (
    GtkWidget * widget,
    GdkEventConfigure * event,
    gpointer data
) {
    DBG (">> signal_on_configure_paper\n");
    rodent_done_with_rename(data);
    view_t *view_p = (view_t *) data;
    TRACE ("PPPCCC: on_configure_paper:: configure event paper********\n");
    TRACE ("PPPCCC: window  w: %d, H:%d\n", view_p->widgets.window->allocation.width, view_p->widgets.window->allocation.height);
    TRACE ("PPPCCC: paper size=(%d,%d), allocation=(%d,%d)\n",
           view_p->widgets.paperX, view_p->widgets.paperY,
           view_p->widgets.paper->allocation.width, view_p->widgets.paper->allocation.height);
    rodent_on_size (view_p, widget->allocation.width, widget->allocation.height);
    rodent_save_root_view_geometry_p (view_p);
    rodent_restore_view_size (view_p);

    // ??  on_size_paper (widget, NULL, data);
    TRACE ("PPPCCC: on_configure_paper:: done\n");
    return TRUE;
}

static gboolean
signal_on_configure_window (
    GtkWidget * window,
    GdkEventConfigure * event,
    gpointer data
) {
    // FIXME: must do the stuff for all view_ps in window...
    DBG (">> signal_on_configure_window\n");
    view_t *view_p = rodent_get_current_view (window);
    TRACE ("WWWCCC on_configure_window:: configure event window********\n");
    TRACE ("window  w: %d, H:%d\n", window->allocation.width, window->allocation.height);
    if(view_p->widgets.rename) {
        rodent_done_with_rename ((gpointer) view_p);
    }
    //gtk_paned_set_position(GTK_PANED(view_p->widgets.vpane), 10000);

    if(!view_p->en) {
        view_geometry_t *iconview_geometry_p = rodent_get_view_geometry_p (view_p);
        if(!iconview_geometry_p
           || (window->allocation.width != iconview_geometry_p->w || window->allocation.height != iconview_geometry_p->h)) {
            DBG ("1.WWW now saving geometry...w: %d, H:%d\n", window->allocation.width, window->allocation.height);
            rodent_save_root_view_geometry_p (view_p);
        }
    } else if(view_p->en->path && rodent_get_view_geometry_p (view_p)) {
        DBG ("2.WWW now saving geometry...w: %d, H:%d\n", window->allocation.width, window->allocation.height);
        rodent_save_root_view_geometry_p (view_p);
    }

    TRACE ("on_configure_window:: done\n");
    return FALSE;
}

#if 0
static gboolean
signal_on_size_window (
    GtkWidget * widget,
    GtkAllocation * allocation,
    gpointer user_data
) {
    DBG (">> signal_on_size_window\n");
# if 0
    view_t *view_p = (view_t *) user_data;
    TRACE ("WWW on_size_window:: size allocation event for window********...\n");

    if(!view_p->en) {
        iconview_geometry_t *iconview_geometry_p = rodent_get_view_geometry_p (view_p);
        if(!iconview_geometry_p
           || (view_p->widgets.window->allocation.width !=
               iconview_geometry_p->w || view_p->widgets.window->allocation.height != iconview_geometry_p->h)) {
            TRACE ("WWW now saving geometry...w: %d, H:%d\n",
                   view_p->widgets.window->allocation.width, view_p->widgets.window->allocation.height);
            rodent_save_view_geometry_p (view_p);
        }
    } else if(view_p->en->path && rodent_get_view_geometry_p (view_p)) {
        rodent_save_view_geometry_p (view_p);
    }

    TRACE ("WWW on_size_window:: done\n");
# endif
    return TRUE;
}
#endif

static void
signal_on_expose (
    GtkWidget * widget,
    GdkEventExpose * event,
    gpointer data
) {
    view_t *view_p = (view_t *) data;
/*  if (DRAWABLE == view_p->widgets.pixmap) {
      //view_p->redraw_pixmap=TRUE;
      rodent_expose(widget, event, data);
      return;
  }*/

    TRACE (">> signal_on_expose\n");
    int bottom_row,
      bottom_column,
      top_row,
      top_column;
    if(!CELLWIDTH || !CELLHEIGHT)
        return;
    TRACE ("on_expose for %d,%d,%d,%d......\n", event->area.x, event->area.y, event->area.width, event->area.height);
    bottom_row = event->area.y / CELLHEIGHT;
    bottom_column = event->area.x / CELLWIDTH;
    top_row = (event->area.y + event->area.height) / CELLHEIGHT;
    if((event->area.y + event->area.height) % CELLHEIGHT)
        top_row++;
    top_column = (event->area.x + event->area.width) / CELLWIDTH;
    if((event->area.x + event->area.width) % CELLWIDTH)
        top_column++;
    TRACE ("graphics_expose(..,%d,%d,%d,%d)\n", bottom_row, bottom_column, top_row, top_column);
    graphics_expose (view_p, bottom_row, bottom_column, top_row, top_column);
}

static gchar *
get_current_text (
    GtkTextView * textview
) {
    // get current text
    GtkTextIter start,
      end;
    GtkTextBuffer *buffer = gtk_text_view_get_buffer (textview);

    gtk_text_buffer_get_bounds (buffer, &start, &end);
    gchar *t = gtk_text_buffer_get_text (buffer, &start, &end, TRUE);
    gchar *tt = t;
    while(*tt == ' ')
        tt++;
    gchar *ttt = g_strdup (tt);
    g_free (t);
    return ttt;
}

static void
status_grab_focus (
    view_t * view_p,
    int keyval
) {
    widgets_t *widgets_p = &(view_p->widgets);
    if(g_object_get_data (G_OBJECT (widgets_p->status), "clean")) {
        rfm_status (widgets_p, "xffm/sudo", NULL);
	view_p->sh_command_counter=g_list_length(view_p->sh_command)-1;
    } else {
        gchar *t = get_current_text ((GtkTextView *) widgets_p->status);
        rfm_status (widgets_p, "xffm/sudo", t, NULL);
        // keep unclean status...
        g_object_set_data (G_OBJECT (widgets_p->status), "clean", NULL);
        g_free (t);
    }
    gtk_widget_grab_focus (widgets_p->status);
/*    GdkWindow *win = gtk_text_view_get_window((GtkTextView *) widgets_p->status,
                                              GTK_TEXT_WINDOW_TEXT);*/
/*	XSetInputFocus(GDK_DISPLAY(),
	  GDK_WINDOW_XID(gtk_widget_get_parent_window(widgets_p->status)),
	  RevertToParent,CurrentTime);*/
/*	XGrabPointer(GDK_DISPLAY(),
	  GDK_WINDOW_XID(win),
          TRUE,
          0,
          GrabModeSync,
          GrabModeAsync,
          GDK_WINDOW_XID(win),
          None, CurrentTime);*/
    return;
}

static gboolean
signal_keyboard_event (
    GtkWidget * window,
    GdkEventKey * event,
    gpointer data
) {
    DBG (">> signal_keyboard_event\n");
    view_t *view_p = rodent_get_current_view (window);

    int selected_row;
    int selected_column;
    int selection_id;
    int last_id;
    double scroll_pos = -1;
    widgets_t *widgets_p = &(view_p->widgets);
    gboolean initial = FALSE;
    TRACE("keyboard_event: 0x%x\n", event->keyval);

    /* keybindings are generated with the popup menu: */
    if(!view_p->widgets.popup) {
	// popup menu is generated by thread at startup.
	// remote case where thread has not yet finished:
	g_warning("view_p->widgets.popup has not been created!");
	return FALSE;
    }

    /* asian Input methods */
    if(event->keyval == GDK_space && (event->state & (GDK_MOD1_MASK | GDK_SHIFT_MASK))) {
        return FALSE;
    }

    if(event->state & (GDK_SHIFT_MASK | GDK_CONTROL_MASK | GDK_MOD1_MASK)) {
        if(rodent_common_mask_key (event, &(view_p->widgets)))
            return TRUE;
    }

    if(!rfm_population_try_read_lock (view_p)) {
        return FALSE;
    }
    if(!view_p->population_pp || !view_p->population_pp[0]) {
        return FALSE;
    }
    rfm_population_read_unlock (view_p);

    // extra keys:

    if (event->keyval == GDK_Search) {
	rodent_glob_activate (NULL, (gpointer) widgets_p);
        return TRUE;
    }
    if (event->keyval == GDK_Back) {
	rodent_back_activate (NULL, (gpointer) widgets_p);
        return TRUE;
    }
    if (event->keyval == GDK_Forward) {
	rodent_jump_to_activate (NULL, (gpointer) widgets_p);
        return TRUE;
    }
    if (event->keyval == GDK_Reload) {
	rodent_refresh_activate (NULL, (gpointer) widgets_p);
        return TRUE;
    }
    if (event->keyval == GDK_HomePage) {
	rodent_home_activate (NULL, (gpointer) widgets_p);
        return TRUE;
    }
    if (event->keyval == GDK_Explorer) {
	rodent_host_activate (NULL, (gpointer) widgets_p);
        return TRUE;
    }

    //DBG ("status\n");

    if(view_p->selection_list && g_list_length (view_p->selection_list)) {
        if(event->keyval == GDK_Delete) {
            rodent_remove_activate ((GtkMenuItem *) window, widgets_p);
            return TRUE;
        }

        if(event->keyval == GDK_Return || event->keyval == GDK_KP_Enter) {
            //g_warning("return whatever");
            if(rfm_population_try_read_lock (view_p)) {
                population_t population_v;
                record_entry_t en_v;
                if(view_p->selected_p) {
                    memcpy (&population_v, view_p->selected_p, sizeof (population_t));
                    memcpy (&en_v, view_p->selected_p->en, sizeof (record_entry_t));
                    population_v.en = &en_v;
                    rfm_population_read_unlock (view_p);
                    rodent_double_click (view_p, view_p->selected_p, !(event->state & GDK_CONTROL_MASK));
                }
            }
            return TRUE;
        }
    }

    int ignore[]={
        GDK_Control_L,
        GDK_Control_R,
        GDK_Shift_L,
        GDK_Shift_R,
        GDK_Shift_Lock,
        GDK_Caps_Lock,
        GDK_Meta_L,
        GDK_Meta_R,
        GDK_Alt_L,
        GDK_Alt_R,
        GDK_Super_L,
        GDK_Super_R,
        GDK_Hyper_L,
        GDK_Hyper_R,

        0
    };
    int i;
    for (i=0; ignore[i]; i++) {
        if(event->keyval ==  ignore[i]) {
            return TRUE;
        }
    }
    /*if(event->keyval == GDK_Control_L || event->keyval == GDK_Control_R) {
        return TRUE;
    }*/

    if(rfm_population_try_read_lock (view_p)) {
        rodent_unselect_all_pixbuf (view_p);
        rodent_unsaturate_pixbuf (view_p);
        rodent_label_event (view_p, NULL);
        rfm_population_read_unlock (view_p);
    }

    if(g_object_get_data (G_OBJECT (widgets_p->status), "clean") == NULL) {
        if(event->keyval == GDK_Escape) {
            gtk_widget_grab_focus (widgets_p->paper);
            rfm_update_status_line (widgets_p->view_p);
            rfm_hide_text (widgets_p);
        } else {
            return FALSE;
        }
    }

    if(event->keyval != GDK_Escape &&
       event->keyval != GDK_Page_Up &&
       event->keyval != GDK_Page_Down &&
       event->keyval != GDK_Home &&
       event->keyval != GDK_Left &&
       event->keyval != GDK_Up &&
       event->keyval != GDK_Down && event->keyval != GDK_Right && event->keyval != GDK_End && event->keyval != GDK_Down) {
        status_grab_focus (view_p, event->keyval);
        return FALSE;
    }
    if(event->keyval == GDK_Page_Up || event->keyval == GDK_Page_Down) {
        int factor = 3 * CELLHEIGHT;
        if(event->keyval == GDK_Page_Up) {
            factor *= -1;
        }
        scroll_pos = gtk_adjustment_get_value (gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window)) + factor;
        if(scroll_pos < 0) {
            scroll_pos = 0;
        }
        if(scroll_pos > view_p->grid_rows * CELLHEIGHT) {
            scroll_pos = view_p->grid_rows * CELLHEIGHT;
        }
        rodent_set_scroll (view_p, scroll_pos);
        return TRUE;
    }

    if(!rfm_population_try_read_lock (view_p)) {
        return FALSE;
    }
    for(selection_id = 0; view_p->population_pp[selection_id]
        && view_p->population_pp[selection_id] != view_p->selected_p; selection_id++) ;

    if(!view_p->population_pp[selection_id]) {
        if(event->keyval != GDK_End)
            selection_id = 0;
        else
            selection_id = view_p->grid_rows * view_p->grid_columns - 1;
        initial = TRUE;
    }
    rfm_population_read_unlock (view_p);

    /* is selection id visible? */
    int invisible_items = invisible_bottom (view_p);

    if(selection_id < invisible_items) {
        selection_id = invisible_items;
        initial = TRUE;
    }
    TRACE ("selection_id=%d, invisibleitems=%d, listlength=%d\n",
           selection_id, invisible_items, view_p->grid_rows * view_p->grid_columns);

    if(selection_id >= view_p->grid_rows * view_p->grid_columns) {
        selection_id = view_p->grid_rows * view_p->grid_columns - 1;
        initial = TRUE;
        scroll_pos = (selection_id / view_p->grid_columns) * CELLWIDTH;
        TRACE ("scrollback, scrollpos at %lf\n", scroll_pos);
    }
    //TRACE("at %s\n", view_p->population_pp[selection_id]->en->path);

    last_id = selection_id;

    if(!initial) {
        switch (event->keyval) {
        case GDK_Escape:
            selection_id = -1;
            break;
        case GDK_Home:
            selection_id = 0;
            scroll_pos = 0;
            break;
        case GDK_Left:
            if(selection_id - 1 >= 0) {
                selection_id--;
                if(selection_id / view_p->grid_columns
                   != last_id / view_p->grid_columns && selection_id < invisible_bottom (view_p)) {
                    scroll_pos =
                        gtk_adjustment_get_value
                        (gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window)) - CELLHEIGHT;
                    if(scroll_pos < 0)
                        scroll_pos = 0;
                }
            }
            break;
        case GDK_Right:
            if(rfm_population_try_read_lock (view_p)) {
                if(view_p->population_pp[selection_id + 1]) {
                    selection_id++;
                    if(selection_id / view_p->grid_columns
                       != last_id / view_p->grid_columns && selection_id >= invisible_top (view_p)) {
                        scroll_pos =
                            CELLHEIGHT +
                            gtk_adjustment_get_value (gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window));
                    }
                }
                rfm_population_read_unlock (view_p);
            }
            break;

        case GDK_End:
            selection_id = view_p->max_elements - 1;
            scroll_pos = (selection_id / view_p->grid_columns) * CELLWIDTH;
            break;
        case GDK_Down:
            if(selection_id + view_p->grid_columns < view_p->grid_rows * view_p->grid_columns) {
                selection_id += view_p->grid_columns;
                if(selection_id >= view_p->max_elements) {
                    selection_id = view_p->max_elements - 1;
                }
                TRACE ("invisible top=%d selection_id=%d\n", invisible_top (view_p), selection_id);
                if(selection_id >= invisible_top (view_p)) {
                    scroll_pos =
                        CELLHEIGHT +
                        gtk_adjustment_get_value (gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window));
                }
            }
            break;
        case GDK_Up:
            if(selection_id - view_p->grid_columns >= 0)
                selection_id -= view_p->grid_columns;
            TRACE ("invisible bottom=%d selection_id=%d\n", invisible_bottom (view_p), selection_id);
            if(selection_id < invisible_bottom (view_p)) {
                scroll_pos =
                    gtk_adjustment_get_value (gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window)) - CELLHEIGHT;
                if(scroll_pos < 0)
                    scroll_pos = 0;
            }
            break;
        default:
            return FALSE;
        }                       // end switch
    }                           // end if (!initial)

    if(selection_id >= 0) {
        selected_column = selection_id % view_p->grid_columns;
        selected_row = selection_id / view_p->grid_columns;
        if(rfm_population_try_read_lock (view_p)) {
            rodent_saturate_pixbuf (view_p, view_p->population_pp[selection_id]);
            rodent_select_pixbuf (view_p, view_p->population_pp[selection_id]);
            rfm_population_read_unlock (view_p);
        }

        if(scroll_pos >= 0) {
            gdouble max = gtk_adjustment_get_upper (gtk_scrolled_window_get_vadjustment (view_p->widgets.scrolled_window));
            if(scroll_pos > max) {
                scroll_pos = max;
            }
            rodent_set_scroll (view_p, scroll_pos);
        }
        DBG ("KEY: item %d selected row and column=%d,%d (grid_rows=%d,grid_columns=%d)\n", selection_id,
             selected_row, selected_column, view_p->grid_rows, view_p->grid_columns);
        /* this expose is necessary because the gtk expose event is returning
         * weird expose areas, probably because the paper has been resized... */
        graphics_expose_item (view_p, selected_row, selected_column);
        graphics_expose_item (view_p, last_id / view_p->grid_columns, last_id % view_p->grid_columns);
        view_p->selected_p = view_p->population_pp[selection_id];
    }
    rfm_update_status_line (view_p);

    return TRUE;
}
