
static void
print_suggestion (
    widgets_t * widgets_p,
    const char *data,
    int i,
    gboolean cr
) {
    gchar *element = g_strdup_printf ("%d", i);
    rfm_diagnostics (widgets_p, NULL, "[", NULL);
    rfm_diagnostics (widgets_p, "xffm_tag/green", element, NULL);
    rfm_diagnostics (widgets_p, NULL, "] ", NULL);
    rfm_diagnostics (widgets_p, "xffm_tag/blue", data, NULL);
    if(cr)
        rfm_diagnostics (widgets_p, NULL, "\n", NULL);

    g_free (element);
}

static gint
ya_strcmp (
    gconstpointer a,
    gconstpointer b
) {
    return strcmp ((char *) a, (char *) b);
}

static void
suggest_command (
    widgets_t * widgets_p,
    const char *complete,
    gboolean anywhere
) {
    view_t *view_p = widgets_p->view_p;
    GList *p;
    char *suggest = NULL;
    for(p = g_list_last (view_p->sh_command); p && p->data; p = p->prev) {
        char *data = (char *) p->data;
        if((anywhere && strstr (data, complete)) || (!anywhere && strncmp (complete, data, strlen (complete)) == 0)) {
            suggest = g_strdup (data);
            break;
        }
        TRACE ("COMPLETE: ?? %s\n", data);
    }
    if(suggest) {
        rfm_status (widgets_p, "xffm/sudo", suggest, NULL);
        g_object_set_data (G_OBJECT (widgets_p->status), "clean", NULL);
        g_free (suggest);
    }
}

static int
length_equal_string(const gchar *a, const gchar *b){
    int length=0;
    int i;
    for (i = 1; i < strlen(a) && i < strlen(b); i++){
	if (strncmp(a,b,i)) {
	    length=i-1;
	    break;
	} else {
	    length=i+1;
	}
    }
 	    TRACE ("%s --- %s differ at length =%d\n",
		   a,b,length);fflush(NULL);
   return length;
}

static gchar *
list_matches (widgets_t * widgets_p, GList **matches_p){
    *matches_p = g_list_sort (*matches_p, ya_strcmp);
    GList *a;
    GList *b;
    int length;
    GList *p = g_list_first (*matches_p);
    char *suggest = g_strdup ((gchar *) (p->data));
    int equal_length=strlen(suggest);
    for(a = *matches_p; a && a->data; a = a->next) {
	for(b = a->next; b && b->data; b = b->next) {
	    length=length_equal_string(
		    (const gchar *)(a->data), (const gchar *)(b->data));
	    if(length < equal_length) {
		equal_length = length;
	    }
	}
    }
    suggest[equal_length]=0;
    TRACE("string=%s, equal_length=%d, suggest=%s\n",
	    (gchar *) (p->data), equal_length, suggest);


    rfm_show_text(widgets_p);
    rfm_diagnostics (widgets_p, "xffm_tag/red", _("Text Completion"), ":\n", NULL);
    int i;
    for(i = 1, p = g_list_first (*matches_p); p && p->data; p = p->next, i++) {
	const gchar *s = p->data;
	rfm_diagnostics (widgets_p, "xffm_tag/green", s, "\n", NULL);
    }
    return suggest;
}

static void
complete_command (widgets_t * widgets_p,  const char *complete) {
//    static gchar *last_complete = NULL;
    view_t *view_p = widgets_p->view_p;

    if(complete && strlen (complete)) {
        TRACE ("COMPLETE: %s\n", complete);
        GList *p;
        GList *matches = NULL;
        char *suggest = NULL;
        int i = 1;
        for(p = g_list_first (view_p->sh_command); p && p->data; p = p->next, i++) {
            char *data = (char *) p->data;
            if(strncmp (complete, data, strlen (complete)) == 0) {
                if(g_list_find_custom (matches, data, ya_strcmp) == NULL) {
                    TRACE ("COMPLETE: match! %s <= %s\n", complete, data);
                    matches = g_list_append (matches, p->data);
                }
            }
            DBG ("COMPLETE: ?? %s\n", data);
        }
        if(matches) {
            TRACE ("COMPLETE: matches %d\n", g_list_length (matches));
            if(g_list_length (matches) == 1) {
                suggest = g_strdup (matches->data);
            } else {
		suggest=list_matches(widgets_p, &matches);
            }

            TRACE ("COMPLETE: g_list_free\n");
            g_list_free (matches);
        }

        if(!suggest && strncmp (complete, "history", strlen (complete)) == 0) {
            suggest = g_strdup ("history");
        }
        if(suggest) {
            rfm_status (widgets_p, "xffm/sudo", suggest, NULL);
            g_object_set_data (G_OBJECT (widgets_p->status), "clean", NULL);
            g_free (suggest);
        }
    }
}

static gboolean
internal_cd (
    widgets_t * widgets_p,
    gchar ** argvp
) {
    view_t *view_p = widgets_p->view_p;
    const gchar *gg;

    if(argvp[1]) {
        gg = argvp[1];
    } else {
        gg = g_get_home_dir ();
    }
    rfm_show_text (widgets_p);
    DBG ("CD: gg=%s argv[1]=%s\n", gg, argvp[1]);
    rfm_diagnostics (widgets_p, "xffm_tag/command", "cd", " ", NULL);
    rfm_diagnostics (widgets_p, "xffm_tag/green", gg, "\n", NULL);
    gdk_flush ();

    // must allow relative paths too.
    if(view_p->en && view_p->en->path && g_file_test (view_p->en->path, G_FILE_TEST_IS_DIR)) {
        if(chdir (view_p->en->path) < 0) {
            g_warning ("gridview.c: chdir(%s)<0", view_p->en->path);
        }
    }
    if(!g_file_test (gg, G_FILE_TEST_IS_DIR)) {
        rfm_diagnostics (widgets_p, "xffm/error", gg, ": ", strerror (ENOENT), "\n", NULL);
        return TRUE;
    }
    gtk_widget_grab_focus (widgets_p->paper);
    // make path absolute
    if(chdir (gg) < 0) {
        g_warning ("chdir(%s): %s", gg, strerror (errno));
    }
    gchar *g = g_get_current_dir ();
    SETWD();
    rodent_push_view_go_history (view_p);
    record_entry_t *new_en = rfm_stat_entry (g, __LOCAL_TYPE);
    view_p->module_name = NULL;
    rfm_save_to_go_history ((gchar *) g);
    rodent_refresh (widgets_p, new_en);
    g_free (g);
    gchar *command = g_strdup_printf ("cd %s", gg);
    rfm_save_sh_command_history (view_p, command);
    return TRUE;
}

static void
print_history (
    widgets_t * widgets_p
) {
    int i;
    view_t *view_p = widgets_p->view_p;
    GList *p;
    rfm_diagnostics (widgets_p, "xffm_tag/command", "history:", "\n", NULL);

    for(i = 1, p = g_list_first (view_p->sh_command); p && p->data; p = p->next, i++) {
        print_suggestion (widgets_p, (char *) p->data, i, TRUE);
    }

}

static void
print_tab (
    widgets_t * widgets_p,
    gchar * text
) {
    int tab_len = 12;
    rfm_diagnostics (widgets_p, "xffm_tag/red", text, NULL);
    for(tab_len = tab_len - strlen (text); tab_len > 0; tab_len--)
        rfm_diagnostics (widgets_p, NULL, " ", NULL);
}

static void
print_history_help (
    widgets_t * widgets_p
) {
    rfm_diagnostics (widgets_p, "xffm_tag/command", "history help:\n", NULL);
    print_tab (widgets_p, "?");
    rfm_diagnostics (widgets_p, "xffm_tag/green", "Print this help message.", "\n", NULL);

    print_tab (widgets_p, "!");
    rfm_diagnostics (widgets_p, "xffm_tag/green", "List recorded command history.", "\n", NULL);

    print_tab (widgets_p, "!n");
    rfm_diagnostics (widgets_p, "xffm_tag/green", "Refer to command line ", NULL);
    rfm_diagnostics (widgets_p, "xffm_tag/red", "n", NULL);
    rfm_diagnostics (widgets_p, "xffm_tag/green", ".", "\n", NULL);

    print_tab (widgets_p, "!string");
    rfm_diagnostics (widgets_p, "xffm_tag/green", "Refer to the most recent command starting with ", NULL);
    rfm_diagnostics (widgets_p, "xffm_tag/red", "string", NULL);
    rfm_diagnostics (widgets_p, "xffm_tag/green", ".", "\n", NULL);

    print_tab (widgets_p, "!?string");
    rfm_diagnostics (widgets_p, "xffm_tag/green", "Refer to the most recent command containing ", NULL);
    rfm_diagnostics (widgets_p, "xffm_tag/red", "string", NULL);
    rfm_diagnostics (widgets_p, "xffm_tag/green", ".", "\n", NULL);

    print_tab (widgets_p, "string<CTRL+TAB>");
    rfm_diagnostics (widgets_p, "xffm_tag/green", "Suggest command completion for ", NULL);
    rfm_diagnostics (widgets_p, "xffm_tag/red", "string", NULL);
    rfm_diagnostics (widgets_p, "xffm_tag/green", ".", "\n", NULL);

    print_tab (widgets_p, "!!");
    rfm_diagnostics (widgets_p, "xffm_tag/green", "Clear all active history.", "\n", NULL);

    print_tab (widgets_p, "!!!");
    rfm_diagnostics (widgets_p, "xffm_tag/green", "Clear all history from disk.", "\n", NULL);
}

static gboolean
internal_history (
    widgets_t * widgets_p,
    gchar ** argvp
) {
    view_t *view_p = widgets_p->view_p;
    // list history
    if(strcmp (argvp[0], "!") == 0) {
        print_history (widgets_p);
        rfm_status (widgets_p, "xffm/sudo", "", NULL);
        g_object_set_data (G_OBJECT (widgets_p->status), "clean", NULL);
        return TRUE;
    }
    // history help
    if(strcmp (argvp[0], "!?") == 0 || strncmp (argvp[0], "?", 1) == 0) {
        print_history_help (widgets_p);
        rfm_status (widgets_p, "xffm/sudo", "", NULL);
        g_object_set_data (G_OBJECT (widgets_p->status), "clean", NULL);
        return TRUE;
    }
    if(strcmp (argvp[0], "!!") == 0 || strcmp (argvp[0], "!!!") == 0) {
        gboolean disk_too = FALSE;
        if(strcmp (argvp[0], "!!!") == 0)
            disk_too = TRUE;
        rfm_clear_sh_command_history (widgets_p->view_p, disk_too);
        rfm_diagnostics (widgets_p, "xffm_tag/command", "history clear: ", NULL);
        rfm_diagnostics (widgets_p, NULL, "[", NULL);
        rfm_diagnostics (widgets_p, "xffm_tag/green", "OK", NULL);
        rfm_diagnostics (widgets_p, NULL, "]", "\n", NULL);
        rfm_status (widgets_p, "xffm/sudo", "", NULL);
        g_object_set_data (G_OBJECT (widgets_p->status), "clean", NULL);
        return TRUE;
    }
    // 
    const char *b = argvp[0] + 1;
    //   errno = 0;
    long n = strtol (b, NULL, 10);
    DBG ("COMPLETE: n=%ld\n", n);
/*    if (errno)  {
	      rfm_diagnostics (widgets_p, "xffm/warning", b, ": ",
				 strerror (errno), NULL);
	      return TRUE;
    }*/
    if(n > 1 && n <= g_list_length (view_p->sh_command)) {
        GList *p = g_list_nth (view_p->sh_command, n - 1);
        if(p && p->data) {
            rfm_status (widgets_p, "xffm/sudo", (char *) p->data, NULL);
            g_object_set_data (G_OBJECT (widgets_p->status), "clean", NULL);
        }
    } else {
        if(*b != '?') {
            suggest_command (widgets_p, b, FALSE);
        } else {
            suggest_command (widgets_p, b + 1, TRUE);
        }
    }
    return TRUE;

}

static gboolean
process_internal_commands (
    widgets_t * widgets_p,
    gchar * command
) {
    gint argcp;
    gchar **argvp;
    GError *error = NULL;
    if(!g_shell_parse_argv (command, &argcp, &argvp, &error)) {
        rfm_diagnostics (widgets_p, "xffm/error", error->message, "\n", NULL);
        g_error_free (error);
        return TRUE;
    } else {
        // shortcircuit chdir and history commands 
        gboolean is_internal = FALSE;
        if(strcmp (argvp[0], "cd") == 0) {
            internal_cd (widgets_p, argvp);
            DBG ("CD: command=%s argv[1]=%s\n", command, argvp[1]);
            is_internal = TRUE;
        } else if(strncmp (argvp[0], "history", strlen ("history")) == 0) {
            print_history_help (widgets_p);
            is_internal = TRUE;
        } else if(strncmp (argvp[0], "!", 1) == 0 || strncmp (argvp[0], "?", 1) == 0) {
            internal_history (widgets_p, argvp);
            is_internal = TRUE;
        }
        if(is_internal) {
        }
        g_strfreev (argvp);
        return is_internal;
    }
}

static gint
on_status_key_press (
    GtkWidget * textview,
    GdkEventKey * event,
    gpointer data
) {
    widgets_t *widgets_p = (widgets_t *) data;
    view_t *view_p = widgets_p->view_p;
    if(!event) {
        DBG ("STATUS: returning on event==0\n");
        //status_grab_focus (widgets_p->view_p, 0);
        return TRUE;
    }
    // backspace must not erase icon
    DBG ("STATUS: * on_status_key_press: got key= 0x%x\n", event->keyval);
    if(event->keyval == GDK_Home) {
        GtkTextIter iter;
        GtkTextBuffer *buffer = gtk_text_view_get_buffer ((GtkTextView *) (widgets_p->status));
        gtk_text_buffer_get_iter_at_offset (buffer, &iter, 2);
        gtk_text_buffer_place_cursor (buffer, &iter);
        return TRUE;
    }
    if(event->keyval == GDK_BackSpace || event->keyval == GDK_Left) {
        gint intval;
        GtkTextBuffer *buffer = gtk_text_view_get_buffer ((GtkTextView *) widgets_p->status);
        g_object_get (G_OBJECT (buffer), "cursor-position", &intval, NULL);
        DBG ("STATUS: cursor-position =%d\n", intval);
        if(intval > 2) {
            // offset==1 is the icon and offset==2 is the space after icon
            return FALSE;
        }
        return TRUE;
    }
    // tab for VAX completion
    if(event->keyval == GDK_Tab && event->state & GDK_CONTROL_MASK) {
        // do VAX completion ...
        gchar *complete = get_current_text ((GtkTextView *) widgets_p->status);
        complete_command (widgets_p, complete);
        g_free (complete);
        g_object_set_data (G_OBJECT (widgets_p->status), "clean", NULL);
        return TRUE;
    }
    // tab for bash completion
    if(event->keyval == GDK_Tab) {
	rfm_show_text(widgets_p);
        rfm_diagnostics (widgets_p, "xffm/warning", NULL);
        rfm_diagnostics (widgets_p, "xffm_tag/green", _("Bash completion will appear in Rodent Gamma.\n"), NULL);
        rfm_diagnostics (widgets_p, "xffm/info", NULL);
        rfm_diagnostics (widgets_p, "xffm_tag/green", _("For command history completion, use CTRL+TAB (see history options with \"history\")\n"), NULL);
	

        g_object_set_data (G_OBJECT (widgets_p->status), "clean", NULL);
        return TRUE;
    }
    // up and down arrows, history command (glist)
    // 
    if(event->keyval == GDK_Up) {
        void *p = g_list_nth_data (view_p->sh_command, view_p->sh_command_counter - 1);
        DBG ("get nth sh_command_counter=%d\n", view_p->sh_command_counter - 1);
        if(p) {
            view_p->sh_command_counter--;
            rfm_status (widgets_p, "xffm/sudo", (char *) p, NULL);
            g_object_set_data (G_OBJECT (textview), "clean", NULL);
            gtk_widget_grab_focus (widgets_p->status);
        }

        return TRUE;
    }

    if(event->keyval == GDK_Down) {
        void *p = g_list_nth_data (view_p->sh_command, view_p->sh_command_counter + 1);
        if(p) {
            view_p->sh_command_counter++;
            rfm_status (widgets_p, "xffm/sudo", (char *) p, NULL);
            g_object_set_data (G_OBJECT (textview), "clean", NULL);
        }
        return TRUE;
    }
    if(event->keyval == GDK_Return || event->keyval == GDK_KP_Enter) {
        if(g_object_get_data (G_OBJECT (textview), "clean") == NULL) {
            // get the command
            gchar *command = get_current_text ((GtkTextView *) widgets_p->status);
            if(command && strlen (command)) {
                // show the lp output area        
                rfm_show_text (widgets_p);

                // first process internal commands for cd and history
                if(process_internal_commands (widgets_p, command))
                    return TRUE;

                // command is now external for /bin/sh
                // printstatus with the run icon run.png
                // this is already done in run.c
                // rfm_diagnostics(widgets_p, "run.png",command,"\n",NULL);

                // run the command (in a shell)
                if(view_p->workdir) {
                    g_free (view_p->workdir);
                }
                view_t *view_p = widgets_p->view_p;
                if (!view_p->en || !view_p->en->path || ! g_file_test(view_p->en->path, G_FILE_TEST_IS_DIR)){
                    view_p->workdir = g_strdup (g_get_home_dir());
;
                } else {
                    view_p->workdir = g_strdup (view_p->en->path);
                }
                rfm_thread_run (widgets_p, command, FALSE);
                g_free (command);
            }
        }
        rfm_status (widgets_p, "xffm/sudo", NULL);
        g_object_set_data (G_OBJECT (textview), "clean", NULL);
        return TRUE;
    }
    if(g_object_get_data (G_OBJECT (textview), "clean")) {
        rfm_status (widgets_p, "xffm/sudo", NULL);
    }
    // set unclean status...
    g_object_set_data (G_OBJECT (textview), "clean", NULL);
    return FALSE;
}
