/* Copyright 2005-2010 Edscott Wilson Garca. 
 * Distributed with GPL licence.*/

#define GRIDVIEW_C

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <pwd.h>

#include <ctype.h>
#include <glob.h>
#include <memory.h>
#include <signal.h>
#include <dirent.h>

#include <X11/Xlib.h>
#ifdef HAVE_LIBSM
# include <X11/SM/SMlib.h>
#endif

#include <gdk/gdkx.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <gmodule.h>

#include "rodent_libs.h"
#include "gridview_lib.h"

/* private symbols */

#include "gridview-graphics.i"
#include "gridview-callbacks.i"
#include "gridview-lpterm.i"
#include "gridview-notebook.i"

GtkWidget *create_gridview ( void); 

view_t *create_iconview ( record_entry_t * en);

/* public symbols (listed in gridview_lib.h) */
GtkWidget *
create_gridview (
    void
) {
    view_t *view_p = NULL;
    int type = __LOCAL_TYPE | __ROOT_TYPE;
    record_entry_t *en = NULL;
    gchar *exec = g_path_get_basename (rfm_global_p->argv[0]);

    // main thread...
    //GThread *self=g_thread_self();
    //DBG("MAIN thread self=0x%lu\n",(long unsigned)self);
    if(exec && strcmp (exec, "rodent")) {
        const gchar *path = rfm_global_p->argv[1];
        if(!path)
            path = "/";
        GSList *plugin_list = rfm_find_root_plugins ();
        en = rfm_mk_entry (type);
        en->path = g_strdup (path);
        for(; plugin_list; plugin_list = plugin_list->next) {
            const gchar *exec_name = (const gchar *) rfm_void ("rfm", 
		    plugin_list->data,
                    "exec_name");
            if(exec_name && strcmp (exec_name, rfm_global_p->argv[0]) == 0) {
                en->module = rfm_void ("rfm", 
			plugin_list->data, "module_name");
                if(rfm_natural ("rfm", en->module, en, "module_path"))
                    en->path = g_strdup (rfm_natural ("rfm", 
				en->module, en, "module_path"));
                else if(rfm_natural ("rfm", en->module, en, "module_label"))
                    en->path = g_strdup (rfm_natural ("rfm", en->module, en, "module_label"));
                TRACE ("%s is in module %s", exec_name, en->module);
                goto load;
            }
        }
        g_warning ("There is no plugin with exec_name = %s", rfm_global_p->argv[0]);
    }
    g_free (exec);

    if(rfm_global_p->argc == 2) {
        gchar *path = NULL;
        en = rfm_mk_entry (type);
        if(!g_file_test (rfm_global_p->argv[1], G_FILE_TEST_IS_DIR)) {
            path = g_build_filename (g_get_home_dir (), rfm_global_p->argv[1], NULL);
            if(!g_file_test (path, G_FILE_TEST_IS_DIR)) {
                g_warning ("%s is not a directory", path);
                g_free (path);
                exit (0);
            }
        } else if(g_path_is_absolute (rfm_global_p->argv[1])) {
            path = g_strdup (rfm_global_p->argv[1]);
        } else {
            gchar *wd = g_get_current_dir ();
            path = g_build_filename (wd, rfm_global_p->argv[1], NULL);
            g_free (wd);
        }
        en = rfm_stat_entry ((const gchar *) path, type);
        g_free (path);

    } else if(rfm_global_p->argc >= 3) {
        en = rfm_mk_entry (type);
        en->path = g_strdup (rfm_global_p->argv[1]);
        if(rfm_void ("rfm", rfm_global_p->argv[2], "is_root_module")) {
            TRACE ("%s is_root_module", rfm_global_p->argv[2]);
            en->module = rfm_global_p->argv[2];
        } else {
            en->module = rfm_natural ("rfm", rfm_global_p->argv[2], NULL, "parent_module_name");
            TRACE ("%s is parent_module_name for %s", (en->module ? en->module : "null"), rfm_global_p->argv[2]);

        }

        if(!en->module) {
            DBG ("There is no %s plugin\n", rfm_global_p->argv[2]);
            //exit(0);
            rfm_destroy_entry (en);
            en = NULL;
        }

    }
  load:
    view_p = create_iconview (en);
    
    if(en)
        view_p->module_name = en->module;
    TRACE ("module=%s", view_p->module_name);
    rodent_full_reload_view (view_p, en);
    rodent_monitor ((gpointer) g_thread_self ());
    if(!view_p)
        return NULL;
    return view_p->widgets.window;
}

static gboolean
watch_preferences (
    gpointer data
) {
    GtkWidget *window = data;
    static gchar *window_color = NULL;
    static double window_transparency = 0.0;

    if (!data) {
        return FALSE;
    }
    if (!G_IS_OBJECT (data)) {
        return FALSE;
    }
    GList *window_view_list = g_object_get_data (G_OBJECT (window), "window_view_list");


    for(; window_view_list; window_view_list = window_view_list->next) {
        view_t *view_p = window_view_list->data;
        if(getenv ("RFM_ICONVIEW_COLOR") && strlen (getenv ("RFM_ICONVIEW_COLOR"))) {
            if(!window_color || strcmp (window_color, getenv ("RFM_ICONVIEW_COLOR"))) {
                DBG ("DESK: watch_root now reloading RFM_ICONVIEW_COLOR %s\n", getenv ("RFM_ICONVIEW_COLOR"));
                g_free (window_color);
                window_color = g_strdup (getenv ("RFM_ICONVIEW_COLOR"));
                rodent_expose_all (view_p);
            }
        }
        if(getenv ("RFM_PASTEBOARD_SERIAL") && strlen (getenv ("RFM_PASTEBOARD_SERIAL"))) {
	    //DBG("got pasteboard serial=%s\n",getenv ("RFM_PASTEBOARD_SERIAL"));
            errno = 0;
            long value = strtol (getenv ("RFM_PASTEBOARD_SERIAL"), NULL, 0);
            if(errno != 0 || value != view_p->pasteboard_serial) {
                view_p->pasteboard_serial = value;
                rfm_update_pasteboard (view_p);
                rodent_update_cut_icons (view_p);
            }
        }
        // DBG("stress soft reload!!!\n");rfm_soft_reload(view_p);

    }

    if(getenv ("RFM_TRANSPARENCY") && strlen (getenv ("RFM_TRANSPARENCY"))) {
        errno = 0;
        double value = strtod (getenv ("RFM_TRANSPARENCY"), NULL);
        if(errno != 0 || value < 0.0) {
            value = 0.0;
        } else if(value > 0.75)
            value = 0.75;
        if(value != window_transparency) {
            window_transparency = value;
            gtk_window_set_opacity ((GtkWindow *) window, 1.00 - window_transparency);
            gdk_flush ();
        }
    }

    return TRUE;
}

view_t *
create_iconview (
    record_entry_t * en
) {
    DBG (">> create_iconview\n");
//   view_p->widgets.window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    GtkWidget *window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_widget_set_has_tooltip (window, TRUE);
    // or:
    // g_object_set_property (G_OBJECT (window), "has-tooltip", TRUE);
    // 
    GtkWidget *vbox1 = gtk_vbox_new (FALSE, 0);
    gtk_container_add (GTK_CONTAINER (window), vbox1);

// NOTEBOOK creation
    GtkWidget *notebook = gtk_notebook_new ();
    //   gtk_notebook_set_tab_pos ((GtkNotebook *)notebook, GTK_POS_BOTTOM);
    gtk_notebook_popup_enable ((GtkNotebook *) notebook);
    gtk_notebook_set_scrollable ((GtkNotebook *) notebook, TRUE);
    g_object_set (notebook,
                  "enable-popup", TRUE, 
                  "can-focus", FALSE,
                  "scrollable", TRUE, 
                  "homogeneous", FALSE, 
                  "show-border", FALSE,
                  "show-tabs", 
                  TRUE, "tab-pos",
                  GTK_POS_TOP, NULL);

    //   g_object_set (notebook,  "show-tabs", FALSE, NULL);
    //   g_object_set (notebook,  "homogeneous", TRUE, NULL); 
    //   "tab-label"

    gtk_box_pack_start (GTK_BOX (vbox1), notebook, TRUE, TRUE, 0);
    gtk_widget_show (vbox1);
    gtk_widget_show (notebook);
// end NOTEBOOK creation
//
// add contents to notebook page
//
    view_t *view_p = create_notebook_page (window, notebook, en);
    g_signal_connect (notebook, "switch-page", G_CALLBACK (switch_page), window);

    if(getenv ("RFM_TRANSPARENCY") && strlen (getenv ("RFM_TRANSPARENCY"))) {
        errno = 0;
        double value = strtod (getenv ("RFM_TRANSPARENCY"), NULL);
        if(errno != 0 || value < 0.0) {
            value = 0.0;
        } else if(value > 0.75)
            value = 0.75;
        gtk_window_set_opacity ((GtkWindow *) window, 1.00 - value);
    }
    //   gtk_window_set_gravity((GtkWindow *) view_p->widgets.window, GDK_GRAVITY_STATIC);
    /*   {
       GdkPixbuf *rfm_icon_pixbuf;
       TRACE("create application icon...\n");
       rfm_icon_pixbuf = rfm_get_pixbuf(rfm_resolve_icon_id(view_p->en), SIZE_ICON);
       if (rfm_icon_pixbuf) {
       gtk_window_set_icon(GTK_WINDOW(window), rfm_icon_pixbuf);
       } else
       g_warning("cannot create application icon");

       gdk_flush();
       } */

// NOTEBOOK plus page
    GtkWidget *button = gtk_button_new ();
    GdkPixbuf *pb;
    GtkWidget *image;
    pb = rfm_get_pixbuf ("xffm/stock_add", SIZE_BUTTON);
    image = gtk_image_new_from_pixbuf (pb);
    gtk_widget_show (image);
    gtk_container_add (GTK_CONTAINER (button), image);
    g_object_set (button, "can-focus", FALSE, "relief", GTK_RELIEF_NONE, NULL);

    gtk_widget_show (button);
    g_signal_connect (G_OBJECT (button), "clicked", G_CALLBACK (newpage), window);
    gtk_notebook_set_action_widget (GTK_NOTEBOOK (notebook), button, GTK_PACK_END);

// NOTEBOOK end
    g_signal_connect (G_OBJECT (window), "query-tooltip", G_CALLBACK (rodent_tip_function), NULL);
    g_signal_connect (G_OBJECT (window), "key-press-event", G_CALLBACK (signal_keyboard_event), NULL);
    g_signal_connect (G_OBJECT (window), "destroy_event", G_CALLBACK (signal_destroy_event), NULL);
    g_signal_connect (G_OBJECT (window), "delete_event", G_CALLBACK (signal_destroy_event), NULL);
    /*   g_signal_connect(G_OBJECT(window), "size-allocate",
       G_CALLBACK(signal_on_size_window), view_p); */
    /* pointer events */
    g_signal_connect (G_OBJECT (window), "configure-event", G_CALLBACK (signal_on_configure_window), NULL);

    // XXX is this data of any use?
    //g_object_set_data(G_OBJECT(window), "view_p",
    //                  (gpointer) view_p->widgets.paper);

    /* set minimum window size: */
    gtk_widget_set_size_request ((GtkWidget *) window, 1.3 * CELLHEIGHT, 1.5 * CELLHEIGHT);

    view_geometry_t *iconview_geometry_p = rodent_get_view_geometry_p (view_p);
    if(iconview_geometry_p) {
        gtk_window_set_default_size ((GtkWindow *) window, iconview_geometry_p->w, iconview_geometry_p->h);
        TRACE ("window... default size=%d,%d\n", iconview_geometry_p->w, iconview_geometry_p->h);
    } else {
        gtk_window_set_default_size ((GtkWindow *) window, DEFAULT_WIDTH, DEFAULT_HEIGHT);
        TRACE ("window... default size=%d,%d\n", DEFAULT_WIDTH, DEFAULT_HEIGHT);
    }
    gtk_window_set_resizable ((GtkWindow *) window, TRUE);
    gtk_widget_show (window);
    g_timeout_add (160, (GtkFunction) watch_preferences, window);

    gdk_flush ();
    TRACE ("----Done with create_iconview\n");
    gtk_widget_grab_focus (view_p->widgets.paper);
    DBG ("POP_SEM:  sem_init(&(view_p->population_sem),0,1)\n");

    return view_p;
}
