/*
 * Copyright (C) 2002-2010 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <limits.h>
#include <math.h>
#include <gmodule.h>

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <gdk/gdkx.h>

#include <libxml/parser.h>
#include <libxml/tree.h>

#include "rfm_libs.h"
#include "rfm_modules.h"

#include "primary-icons.i"

gpointer
rfm_preview_thread_f (gpointer data) {
    preview_child_t *preview_child_p = data;
    view_t *view_p = preview_child_p->view_p;
    int population_serial = preview_child_p->population_serial;
    population_t *population_p = preview_child_p->population_p;

    GdkPixbuf *pixbuf;
    DBG ("read_lock by rfm_preview_thread_f (0x%lx) %s\n", (long unsigned)g_thread_self (), population_p->en->path);
     
    rfm_population_read_lock (view_p);
    if(invalid_population (view_p, population_serial)) {
	DBG("invalid_population (population_serial=%d)\n", population_serial);
        rfm_population_read_unlock (view_p);
	g_free(preview_child_p);
        return NULL;
    }
    gchar *path = g_strdup(population_p->en->path);
    int size = view_p->icon_size;
    rfm_population_read_unlock (view_p);
    
    pixbuf = rfm_get_pixbuf (path, size);
    g_free(path);
    rfm_population_read_lock (view_p);
    if(invalid_population (view_p, population_serial)) {
	DBG("invalid_population (population_serial=%d)\n", population_serial);
        rfm_population_read_unlock (view_p);
	g_free(preview_child_p);
        return NULL;
    }
    if(GDK_IS_PIXBUF (pixbuf)) {
        population_p->pixbuf = pixbuf;
        population_p->pixbufW = gdk_pixbuf_get_width ((const GdkPixbuf *)pixbuf);
        population_p->pixbufH = gdk_pixbuf_get_height ((const GdkPixbuf *)pixbuf);
        population_p->normal_pixbuf = population_p->base_pixbuf = pixbuf;
        population_p->saturated_pixbuf = population_p->selected_pixbuf = NULL;
        view_p->redraw_pixmap = TRUE;
        rfm_expose_rect (view_p->widgets.paper->window, &(population_p->rect));
    } 
    // tag population_p as already previewed
    population_p->preview_done = TRUE;
    rfm_population_read_unlock (view_p);
    DBG ("rfm_preview_thread_f (0x%lx) rfm_population_read_unlock!\n", 
		(long unsigned)g_thread_self ());
    g_free(preview_child_p);
    return NULL;
}
void
rfm_fireup_previews (view_t * view_p) {
    DBG ("PREVIEWS: rfm_fireup_previews\n");
    /* update undetermined icons with file magic 
     * if (!SHOWS_IMAGES(view_p->en->type)) {
     DBG("SEMAPHORE: !SHOWS_IMAGES -> sem_post\n");
     sem_post(view_p->preview_p->semaphore); // green light
     return;
     }*/
    if(!view_p->population_pp) {
        return;
    }
    // fireup preview threaded thread manager
    preview_manager_t *preview_manager_p = 
	(preview_manager_t *) malloc (sizeof (preview_manager_t));
    memset (preview_manager_p, 0, sizeof (preview_manager_t));
    preview_manager_p->view_p = view_p;
    preview_manager_p->population_serial = view_p->population_serial;

    DBG ("THREAD: g_thread_create(thread_preview_manager) at primary-icons.c\n");
    g_thread_create (thread_preview_manager, (gpointer) (preview_manager_p), FALSE, NULL);
    return;
}

gboolean
rfm_is_image (record_entry_t * en) {
    if(!en)
        return FALSE;
    static GSList *pix_formats = NULL;
    GSList *l;
    gboolean result = FALSE;
    TRACE ("MIME call from  rfm_is_image\n");

    if(!en->mimetype){
	en->mimetype = MIME_type (en->path, en->st);
    }
    if(!en->mimemagic){
        en->mimemagic = MIME_magic (en->path);
    }
    /* check for image support types */

    if(!pix_formats)
        pix_formats = gdk_pixbuf_get_formats ();
    for(l = pix_formats; l; l = l->next) {
        gchar **pix_mimetypes;
        int i;
        GdkPixbufFormat *fmt = l->data;
        pix_mimetypes = gdk_pixbuf_format_get_mime_types (fmt);
        for(i = 0; pix_mimetypes[i]; i++) {
            TRACE ("allowable pix_format=%s\n", pix_mimetypes[i]);
            if(en->mimetype && 
		    !g_ascii_strcasecmp (pix_mimetypes[i], en->mimetype)) {
                g_strfreev (pix_mimetypes);
                return TRUE;
            }
            if(en->mimemagic && 
		    !g_ascii_strcasecmp (pix_mimetypes[i], en->mimemagic)) {
                g_strfreev (pix_mimetypes);
                return TRUE;
            }        
	}
        g_strfreev (pix_mimetypes);
        if(result) {
            return TRUE;
        }
    }
    /* let's keep this one in memory to reduce overhead */
    /*g_slist_free(pix_formats); */

    TRACE ("%s mimetype = %s, magic=%s is not image\n", en->path, en->mimetype, en->mimemagic);

    return result;
}

const gchar *
resolve_folder_icon (record_entry_t * en) {
    const gchar *basic_type;

    /* if (IS_NOACCESS_TYPE(en->type)) return "xffm/no-access"; */

    if(en->path && strcmp (g_get_home_dir (), en->path) == 0) {
        /*DBG("%s==%s\n",g_get_home_dir(),en->path); */
        basic_type = "xffm/stock_home";
    }
    //else if(IS_EXPANDED(type)) basic_type="xffm/open_folder";
    else
        basic_type = "xffm/closed_folder";

    return (basic_type);

}

const gchar *
rfm_resolve_icon_id (record_entry_t * en) {
    if(!en) {
        return ("xffm/stock_system");
    }
    if(en->parent_module) {
        const gchar *id = rfm_natural ("rfm", en->parent_module, en, "item_icon_id");
        // if module does not resolve icon, proceed as normal.
        if(id) {
            return id;
        }
    }
    if(en->module) {
        const gchar *id = rfm_natural ("rfm", en->module, en, "module_icon_id");
        DBG ("resolving module icon id=%s\n", id);
        if(id)
            return id;
    }
    if(IS_UP_TYPE (en->type)) {
        return ("xffm/go-up");
    }

    if(IS_SDIR (en->type)) {
        const gchar *t;
        if(strcmp (en->path, g_get_home_dir ()) == 0) {
            return "xffm/stock_home";
        }
        if(IS_NOWRITE_TYPE (en->type) || IS_NOACCESS_TYPE (en->type)) {
            t = "inode/directory";
        } else {
            t = "xffm/closed_folder";
        }

        if(IS_SLNK (en->type)) {
            t = "inode/symlink";
        } else if(IS_PROC_TYPE (en->type)){
            t = "xffm/chardevice";
	} else if(IS_CDFS_TYPE (en->type)){
            t = "xffm/cdrom";
	} else if(strstr (en->path, "floppy")){
            t = "xffm/floppy";
	} else if(strstr (en->path, "cdrom") &&  FSTAB_is_in_fstab(en->path)){
            t = "xffm/cdrom";
	} else if(strstr (en->path, "cdrw") &&  FSTAB_is_in_fstab(en->path)){
            t = "xffm/cdrom";
	} else if(strstr (en->path, "dvd") &&  FSTAB_is_in_fstab(en->path)){
            t = "xffm/dvd";
	}

        return t;
    }

    /* broken links : */
    if(IS_BROKEN_LNK (en->type))
        return ("xffm/broken");

    if( IS_SDIR(en->type)) {
        return (resolve_folder_icon (en));
    }
//    if (IS_NOACCESS_TYPE(en->type)) return ("xffm/no-access");

    /*if failed: */
    if(!en->mimetype)
        en->mimetype = MIME_type (en->path, en->st);
    /*magic at this step: */
    if(!en->mimetype)
        en->mimetype = MIME_magic (en->path);
    return en->mimetype;
}

gchar *
rfm_get_composite_icon_id (record_entry_t * en, const gchar * id) {
    if(!id)
        id = rfm_resolve_icon_id (en);
    gchar *p = g_strdup (id);
    if(!en)
        return p;
    TRACE ("composite_type_id(%s)\n", en->path);
    gchar *composite_type;

    if(IS_SLNK (en->type) || (IS_NOACCESS_TYPE (en->type))) {
        composite_type = g_strconcat (p, "/composite", NULL);
        g_free (p);
        if(IS_NOACCESS_TYPE (en->type)) {
            p = composite_type;
            composite_type = g_strconcat (p, "/error", NULL);
            g_free (p);
        }
        if(IS_SLNK (en->type) && !IS_SDIR (en->type)) {
        //if(IS_SLNK (en->type) && ! IS_XF_DIR(en->type)) {
            p = composite_type;
            composite_type = g_strconcat (p, "/link", NULL);
            g_free (p);
        }
        TRACE ("TRACE 1 s:%s composite type=%s type=0x%x (link=%d)\n", en->path, composite_type, en->type, IS_SLNK (en->type));

    } else if (en->path && g_file_test(en->path, G_FILE_TEST_IS_DIR) &&
	    (FSTAB_is_in_fstab(en->path) || FSTAB_is_mounted(en->path))) {
	if (FSTAB_is_mounted(en->path)) {
	    composite_type = g_strconcat (p, "/composite/stock_yes", NULL);
	} else {
	    composite_type = g_strconcat (p, "/composite/stock_no", NULL);
	}
    } /*else if ((en->mimetype && strstr(en->mimetype, "application/x-iso9660-image")) ||
	       (en->mimemagic && strstr(en->mimemagic, "application/x-iso9660-image"))){
	if (FSTAB_is_mounted(en->path)) {
	    composite_type = g_strconcat (p, "/composite/stock_yes", NULL);
	} else {
	    composite_type = g_strconcat (p, "/composite/stock_no", NULL);
	}
    } */
    else {
        return p;
    }
    return composite_type;
}

Pixmap
rfm_create_background_pixmap (char *file) {
    /* create Pixmap */
    GdkPixmap *pixmap;
    GdkPixbuf *pixbuf;
    Pixmap xpixmap = None;
    GdkGC *gc;
    gint root_w,
      root_h,
      root_d;

    if(xpixmap != None) {
        XFreePixmap (GDK_DISPLAY (), xpixmap);
    }
    gdk_window_get_geometry (gdk_get_default_root_window (), NULL, NULL, &root_w, &root_h, &root_d);
    pixbuf = create_background_pixbuf (file, root_w, root_h);
    if(!pixbuf) {
        //g_warning("cannot create pixbuf from %s", file);
    }

    pixmap = gdk_pixmap_new (gdk_get_default_root_window (), root_w, root_h, -1);
    if(!pixmap)
        g_error ("cannot create pixmap from %s", file);
    gc = gdk_gc_new ((GdkDrawable *) pixmap);
    gdk_gc_set_colormap (gc, gdk_colormap_get_system ());

    if(getenv ("RFM_DESKTOP_COLOR") && strlen (getenv ("RFM_DESKTOP_COLOR"))) {
        XColor db_def,
          hw_def;
        GdkColor user_bg_color;
        if(!XAllocNamedColor (GDK_DISPLAY (),
                              GDK_COLORMAP_XCOLORMAP (gdk_colormap_get_system
                                                      ()), getenv ("RFM_DESKTOP_COLOR"), &db_def, &hw_def)) {
            g_warning ("Cannot allocate color %s", getenv ("RFM_DESKTOP_COLOR"));
            user_bg_color.red = 0;
            user_bg_color.green = 0;
            user_bg_color.blue = 0;
            user_bg_color.pixel = 0;

        } else {
            user_bg_color.red = db_def.red * 0xff - 1;
            user_bg_color.green = db_def.green * 0xff - 1;
            user_bg_color.blue = db_def.blue * 0xff - 1;
            user_bg_color.pixel = db_def.pixel;
        }
        gdk_gc_set_foreground (gc, &user_bg_color);
    }
    gdk_draw_rectangle (pixmap, gc, TRUE, 0, 0, root_w, root_h);

    if(pixbuf) {
        gdk_pixbuf_render_to_drawable (pixbuf, (GdkDrawable *) pixmap,
                                       gc, 0, 0, 0, 0, gdk_pixbuf_get_width (pixbuf), gdk_pixbuf_get_height (pixbuf),
                                       //root_w, root_h,
                                       GDK_RGB_DITHER_NONE, 0, 0);

        TRACE ("rendering at size %d,%d\n", gdk_pixbuf_get_width (pixbuf), gdk_pixbuf_get_height (pixbuf));
        TRACE ("placing at %d,%d\n", (root_w - gdk_pixbuf_get_width (pixbuf)) / 2, (root_h - gdk_pixbuf_get_height (pixbuf)) / 2);
    }
    xpixmap = XCreatePixmap (GDK_DISPLAY (), GDK_ROOT_WINDOW (), root_w, root_h, root_d);

    XFillRectangle (GDK_DISPLAY (), xpixmap, GDK_GC_XGC (gc), 0, 0, root_w, root_h);
    if(pixbuf) {
        XCopyArea (GDK_DISPLAY (),
                   GDK_PIXMAP_XID (pixmap),
                   xpixmap,
                   GDK_GC_XGC (gc),
                   0, 0,
                   gdk_pixbuf_get_width (pixbuf),
                   gdk_pixbuf_get_height (pixbuf),
                   (root_w - gdk_pixbuf_get_width (pixbuf)) / 2, (root_h - gdk_pixbuf_get_height (pixbuf)) / 2);
        g_object_unref (pixbuf);
    }
    g_object_unref (pixmap);

    return xpixmap;
}

/******************************************************************/
//XXX: this stuff is to be new icon module...

gchar *
rfm_get_thumbnail_path (const gchar * file, gint size) {
    gchar *cache_dir;
    gchar *thumbnail_path = NULL;
    GString *gs;
    gchar key[11];

    cache_dir = g_build_filename (g_get_user_cache_dir (), "rfm", "thumbnails", NULL);
    if(g_mkdir_with_parents (cache_dir, 0700) < 0) {
        g_free (cache_dir);
        return NULL;
    }

    /* thumbnails are not subject to thumbnailization: */
    gchar *dirname = g_path_get_dirname (file);
    if(strncmp (cache_dir, dirname, strlen (cache_dir)) == 0) {
        TRACE ("thumbnails cannot be thumbnailed:%s\n", file);
        g_free (cache_dir);
        g_free (dirname);
        return NULL;
    }

    gs = g_string_new (dirname);
    sprintf (key, "%10u", g_string_hash (gs));
    g_strstrip (key);
    g_string_free (gs, TRUE);
    g_free (dirname);

    gchar *thumbnail_dir = g_build_filename (cache_dir, key, NULL);
    if(g_mkdir_with_parents (thumbnail_dir, 0700) < 0) {
        g_free (thumbnail_dir);
        return NULL;
    }

    gchar *filename = g_path_get_basename (file);

    gs = g_string_new (file);
    sprintf (key, "%10u", g_string_hash (gs));
    g_strstrip (key);
    g_string_free (gs, TRUE);
    g_free (filename);

    filename = g_strdup_printf ("%s-%d.png", key, size);
    thumbnail_path = g_build_filename (thumbnail_dir, filename, NULL);
    g_free (filename);
    g_free (cache_dir);
    g_free (thumbnail_dir);
    DBG ("thread: %s ->thumbnail_path=%s\n", file, thumbnail_path);

    return thumbnail_path;
}

GdkPixbuf *
rfm_get_from_pixbuf_hash (const gchar * key, int size) {
    pixbuf_t *pixbuf_p = find_in_pixbuf_hash (key, size);
    if(pixbuf_p)
        return pixbuf_p->pixbuf;
    return NULL;
}

GdkPixbuf *
rfm_get_pixbuf (const gchar * key, int size) {
    DBG ("rfm_get_pixbuf(%s, %d)\n", key, size);
    if(!key) {
        DBG ("primary-icons.c:733, key is NULL!");
	return NULL;
    }
    // find in pixbuf hash
    pixbuf_t *pixbuf_p = find_in_pixbuf_hash (key, size);
    if(pixbuf_p) {
        DBG ("HASH-primary-icons.c: %s found in pixbuf hash\n", key);
        return pixbuf_p->pixbuf;
    }
    pixbuf_p = (pixbuf_t *) malloc (sizeof (pixbuf_t));
    memset (pixbuf_p, 0, sizeof (pixbuf_t));

    if(g_path_is_absolute (key)) {
        if(!g_file_test (key, G_FILE_TEST_EXISTS)) {
            g_free (pixbuf_p);
            DBG ("rfm_get_pixbuf(%s): file does not exist! (absolute path)", key);
            return NULL;
        }
        pixbuf_p->pixbuf = create_preview (key, size);
        if(!pixbuf_p->pixbuf) {
            g_free (pixbuf_p);
            DBG ("create_preview(%s, %d): returns NULL!\n", key, size);
            return NULL;
        }
        if(g_file_test (key, G_FILE_TEST_IS_SYMLINK)) {
            GdkPixbuf *tag = rfm_get_pixbuf ("xffm/link", size);
            insert_pixbuf_tag (tag, pixbuf_p->pixbuf, "SW", 2.0);
        }
        struct stat st;
        stat (key, &st);
        pixbuf_p->mtime = st.st_mtime;
    } else {
        pixbuf_p->pixbuf = make_pixbuf_from_id (key, size);
        if(!pixbuf_p->pixbuf) {
            g_free (pixbuf_p);
            DBG ("rfm_get_pixbuf(%s, %d): returns NULL!\n", key, size);
            return NULL;
        }
    }
    pixbuf_p->path = g_strdup (key);
    pixbuf_p->size = size;
    if(size <= MAX_PIXBUF_HASH_SIZE) {
        gchar *hash_key = get_hash_key (key, size);
        g_mutex_lock (pixbuf_hash_mutex);
        g_object_ref (pixbuf_p->pixbuf);
        DBG ("HASH: inserting into pixbuf hash: %s --> %s\n", key, hash_key);
        g_hash_table_insert (pixbuf_hash, hash_key, pixbuf_p);
        g_mutex_unlock (pixbuf_hash_mutex);
    }
    return pixbuf_p->pixbuf;
}

////////////////   module wraparounds:
//

gint
rfm_svg_supported (void) {
    void *value = rfm_void ("rfm", "icons", "svg_supported");
    return GPOINTER_TO_INT (value);
}
