
typedef struct thread_run_t {
    char *wd;
    char **argv;
} thread_run_t;

typedef struct run_data_t {
    widgets_t *widgets_p;
    pid_t pid;
    gchar *command;
    gchar *workdir;
    GtkWidget *button;
    gboolean terminating; 
} run_data_t;

/***************  threaded ****************/

static
  gboolean
check_sudo (void
    ) {
    // 1. check for sudo askpass option
    FILE *pipe;
    gboolean askpass = FALSE;
    pipe = popen ("sudo -L", "r");
    if(pipe) {
#define PAGE_LINE 2048*8
        char line[PAGE_LINE];
        line[PAGE_LINE - 1] = 0;
        while(fgets (line, PAGE_LINE - 1, pipe) && !feof (pipe)) {
            if(strstr (line, "askpass"))
                askpass = TRUE;
        }
        pclose (pipe);
    }
    if(!askpass) {
        g_warning ("Sudo version obsolete. Please update sudo!\n");
    }
    return askpass;
}

static
    void
run_fork_finished_function (void *user_data) {
#if 0
    widgets_t *widgets_p = user_data;

    gchar *g = g_strdup_printf ("%d>", Tubo_id () - 1);
    if(rfm_diagnostics_is_visible (widgets_p)) {
        GDK_THREADS_ENTER ();
        rfm_diagnostics (widgets_p, "xffm_tag/command_id", g, NULL);
        rfm_diagnostics (widgets_p, "xffm/stock_no", "\n", NULL);
        GDK_THREADS_LEAVE ();
    }
    g_free (g);
#endif
    TRACE ("TRACE:fstab over by fork_finished_function\n");
}

static
    void
fork_function (void *data) {
    char **argv = (char **)data;

    int i = 0;
    int j = 0;
    char *exec_argv[MAX_COMMAND_ARGS];

    exec_argv[i++] = argv[j++];

    // sudo checks:
    gchar *sudo = g_find_program_in_path ("sudo");
    if(sudo && (strcmp (argv[0], "sudo") == 0 || strcmp (argv[0], sudo) == 0)) {
        check_sudo ();
        if(argv[1] && strcmp (argv[1], "-A") == 0) {
            exec_argv[i++] = "-A";
        }
    }
    g_free (sudo);
    // end of sudo checks
    // put the rest of argv into place:

    for(; i < MAX_COMMAND_ARGS - 2 && argv[j]; i++, j++) {
        exec_argv[i] = argv[j];
    }
    exec_argv[i] = NULL;
    if (i==MAX_COMMAND_ARGS - 1) {
    	fprintf(stderr,"%s: (> %d)\n", strerror(E2BIG), MAX_COMMAND_ARGS);
    }

    TRACE ("FORK: ");
    for(i = 0; exec_argv[i] != NULL; i++)
        TRACE (" %s", exec_argv[i]);
    TRACE ("\n");

    execvp (exec_argv[0], exec_argv);
    fprintf (stderr, "CHILD could not execvp: this should not happen\n");
    fprintf (stderr, "Do you have %s in your path?\n", exec_argv[0]);
    fflush (stderr);
    rfm_threadwait ();
    _exit (123);
}

static gboolean
destroy_run_button (GtkWidget * widget, GdkEvent * event, gpointer data) {
    run_data_t *run_data_p = (run_data_t *) data;
    TRACE ("destroy_run_button now!\n");
    if(!data)
        return TRUE;
    run_data_p->button = NULL;
    TRACE ("run_data_p->button = NULL\n");
    return TRUE;
}

static void
show_run_info (GtkButton * button, gpointer data) {
    run_data_t *run_data_p = data;
    run_data_p->terminating=TRUE;
}

static GtkWidget *
make_run_data_button (run_data_t * run_data_p) {

    if(run_data_p->widgets_p->button_space) {
	gchar *tip=g_strdup_printf("%s (%d)",  
		run_data_p->command, run_data_p->pid);
        run_data_p->button =
            rfm_mk_little_button (run_data_p->widgets_p, "xffm/stock_stop",
                                  (gpointer) show_run_info, run_data_p,
				  tip);
	g_free(tip);
        gtk_box_pack_end (GTK_BOX (run_data_p->widgets_p->button_space), run_data_p->button, FALSE, FALSE, 0);
        g_signal_connect (G_OBJECT (run_data_p->button), "destroy_event", G_CALLBACK (destroy_run_button), run_data_p);
        gtk_widget_show (run_data_p->button);
        DBG ("DIAGNOSTICS:srun_button made for grandchildPID=%d\n", (int)run_data_p->pid);
    }
    return run_data_p->button;
}

static gpointer
thread_run_f (gpointer data) {
    run_data_t *run_data_p = (run_data_t *) data;
    int status;

    /* create little button (note: thread protected within subroutine) */
GDK_THREADS_ENTER ();
    make_run_data_button (run_data_p);
GDK_THREADS_LEAVE ();
    DBG ("grand---thread waitpid for %d on (%s/%s)\n", run_data_p->pid, run_data_p->command, run_data_p->workdir);
    gboolean terminated = FALSE;
    pid_t pid;
    do {
	if (run_data_p->terminating) {
	    // don't confirm for diagnostics windows.
	    // currently, this is only for find operations
	    if(run_data_p->widgets_p->diagnostics_window) {
		gchar *g = g_strdup_printf ("%d", pid);
            GDK_THREADS_ENTER ();
		rfm_diagnostics (run_data_p->widgets_p, "xffm/warning", NULL);
		rfm_diagnostics (run_data_p->widgets_p, "xffm_tag/command_id", strerror (ECANCELED), ": ", g, "\n", NULL);
            GDK_THREADS_LEAVE ();
		g_free (g);
		terminated=TRUE;
	    } else {
		gchar *text = g_strdup_printf ("%s: %s (%d)\n%s\n", _("Terminate (TERM)"), run_data_p->command, (int) run_data_p->pid, _("Confirm"));
	    
            GDK_THREADS_ENTER ();
		if(rfm_confirm (run_data_p->widgets_p, GTK_MESSAGE_QUESTION, text, _("No"), _("Yes"))) {
		    gchar *g = g_strdup_printf ("%d", (int)(pid));
		    rfm_diagnostics (run_data_p->widgets_p, "xffm/warning", NULL);
		    rfm_diagnostics (run_data_p->widgets_p, "xffm_tag/command_id", _("Terminate (TERM)"), ": ", g, "\n", NULL);
		    g_free (g);
		    terminated=TRUE;
		}
            GDK_THREADS_LEAVE ();
	    } 
	    run_data_p->terminating=FALSE;
	} else if(terminated) {        //KO ETIMEDOUT
	    // reshow little red button
	    gtk_widget_show(GTK_WIDGET (run_data_p->button));
            gchar *text2 =g_strdup_printf ("%s %s: %s\n\n%s %s (%d)?",
                    _("Kill (KILL)"), run_data_p->command, 
		    strerror(ETIMEDOUT), 
                    _("Kill"), run_data_p->command, run_data_p->pid);
            GDK_THREADS_ENTER ();
            if(rfm_confirm (run_data_p->widgets_p, GTK_MESSAGE_QUESTION, text2, _("No"), _("Yes"))) {
                // SIGUSR2 will pass on a SIGKILL signal
                gchar *gg = g_strdup_printf ("%d", (int)(run_data_p->pid));
                rfm_diagnostics (run_data_p->widgets_p, "xffm/warning", NULL);
                rfm_diagnostics (run_data_p->widgets_p, "xffm_tag/command_id", _("Kill (KILL)"), " ", gg, "\n", NULL);
                GDK_THREADS_LEAVE ();
                g_free (gg);
                kill (run_data_p->pid, SIGUSR2);
                rfm_threadwait ();
                waitpid (run_data_p->pid, &status, 0);
                g_free (text2);
		gtk_widget_hide(GTK_WIDGET (run_data_p->button));
                break;
            } else {
		GDK_THREADS_LEAVE ();
                terminated = FALSE;
            }

            g_free (text2);
        } 
        if(terminated) {
                kill (run_data_p->pid, SIGUSR1);
		gtk_widget_hide(GTK_WIDGET (run_data_p->button));
                sleep (2);
        } 
        rfm_threadwait ();
        pid = waitpid (run_data_p->pid, &status, WNOHANG);
        TRACE ("waitpid=%d, status=%d\n", pid, status);
       
    } while(!WIFEXITED (status) || pid == 0);
    DBG ("grand---thread waitpid for %d complete!\n", run_data_p->pid);
    /* remove little button (thread protect gtk here) */
    GDK_THREADS_ENTER ();
    if(run_data_p->button && GTK_IS_WIDGET (run_data_p->button))
        gtk_widget_destroy (GTK_WIDGET (run_data_p->button));
    GDK_THREADS_LEAVE ();
    g_free (run_data_p->command);
    g_free (run_data_p->workdir);
    g_free (run_data_p);

    return NULL;
}

gchar *
rfm_what_term (void) {
    gchar *term = NULL;
    if(getenv ("TERMINAL_CMD") && strlen (getenv ("TERMINAL_CMD"))) {
        term = g_find_program_in_path (getenv ("TERMINAL_CMD"));
        if(!term) {
            g_warning ("TERMINAL_CMD=%s: %s", getenv ("TERMINAL_CMD"), strerror (ENOENT));
        }
    }
    if(!term) {
        // fallback:
        char **p,
         *terminals[] = { "xterm", "gnome-terminal", "konsole", "Terminal", "eterm", "dtterm", NULL };
        for(p = terminals; *p; p++) {
            term = g_find_program_in_path (*p);
            if(term)
                break;
        }
    }
    if(!term)
        g_warning ("no terminal command found!");
    return (gchar *) term;
}

static void
setup_run_button_thread (widgets_t * widgets_p, const gchar * exec_command, pid_t child) {
    view_t *view_p=widgets_p->view_p;
    DBG ("setup_run_button_thread grandchildPID=%d\n", (int)child);
    run_data_t *run_data_p = (run_data_t *) malloc (sizeof (run_data_t));
    memset (run_data_p, 0, sizeof (run_data_t));
    run_data_p->button = NULL;
    run_data_p->pid = child;
    run_data_p->command = g_strdup (exec_command);
    if (view_p) {
	run_data_p->workdir = g_strdup (view_p->workdir);
    } else {
	run_data_p->workdir = g_strdup (g_get_home_dir());
    }
    run_data_p->widgets_p = widgets_p;
    DBG ("THREAD: g_thread_create(thread_run_f) at primary-run.i\n");
    g_thread_create (thread_run_f, (gpointer) run_data_p, FALSE, NULL);
}

static pid_t
thread_run (
	widgets_t * widgets_p, 
	gchar ** argv, 
	void (*stdout_f) (void *stdout_data,
                      void *stream,
                      int childFD)
	) {
    view_t *view_p=(widgets_p)?widgets_p->view_p:NULL;
    if(view_p && !g_file_test (view_p->workdir, G_FILE_TEST_IS_DIR)) {
        g_free (view_p->workdir);
        view_p->workdir = g_strdup (g_get_home_dir ());
    }
    if(view_p && chdir (view_p->workdir) < 0) {
        g_warning ("chdir(%s): %s", view_p->workdir, strerror (errno));
    }
    gchar *g = g_find_program_in_path (argv[0]);
    if(g)
        g_free (g);
    else {
        rfm_show_text(widgets_p);
        rfm_diagnostics (widgets_p, "xffm/error", NULL);
        rfm_diagnostics (widgets_p, "xffm_tag/stderr", argv[0], ": ", strerror (ENOENT), "\n", NULL);
        return -1;
    }

    
    if (stdout_f==NULL) {
	stdout_f=rfm_operate_stdout;
    }
    pid_t child = Tubo_threads (fork_function,
                                (void *)argv,
                                NULL,   //&stdin_fd,
                                stdout_f,
                                rfm_operate_stderr,
                                run_fork_finished_function,
                                (void *)widgets_p,
                                FALSE,
                                TRUE);
    if (view_p) {
	view_p->children = g_list_append (view_p->children, (void *)((long)child));
        
	g_free (view_p->workdir);
	view_p->workdir = g_strdup (GETWD);
    }

    SETWD();
    return child;

}

static
void
private_rfm_thread_run_argv (
	widgets_t * widgets_p, 
	gchar ** argv, gboolean interm, 
	void (*stdout_f) (void *stdout_data,
                      void *stream,
                      int childFD)
	) {
    int i = 0;
    gchar *v_argv[MAX_COMMAND_ARGS];
    gchar *term = NULL;
    if(interm) {
	gchar *t = NULL;
        char *exec_option = "-e";
        t = rfm_what_term ();
        term = g_path_get_basename (t);
        if(strcmp (t, "gnome-terminal") == 0 || strcmp (t, "Terminal") == 0){
            exec_option = "-x";
	}
	g_free(t);
        v_argv[i++] = term;
        v_argv[i++] = exec_option;
    }
    gchar **p;
    for(p = argv; p && *p && i < MAX_COMMAND_ARGS - 2; p++) {
        v_argv[i++] = *p;
    }
    v_argv[i] = NULL;
    if (i==MAX_COMMAND_ARGS - 1) {
	rfm_show_text(widgets_p);
    	rfm_diagnostics(widgets_p,"xffm/error",NULL);
        gchar *max=g_strdup_printf("%d",MAX_COMMAND_ARGS);
        rfm_diagnostics (widgets_p, "xffm_tag/stderr", strerror(E2BIG)," (> ",max,")","\n", NULL);
        g_free(max);
    }
    int id=Tubo_id ();
    gchar *command = g_strdup (v_argv[0]);
    gchar *qq;
    for(i = 1; v_argv[i]; i++) {
            qq = command;
            command = g_strconcat (qq, " ", v_argv[i], NULL);
            g_free (qq);
    }

    pid_t child = thread_run (widgets_p, v_argv, stdout_f);
    if(rfm_diagnostics_is_visible (widgets_p)) {
        gchar *g = g_strdup_printf ("%d> (%d):", id, child);
        rfm_diagnostics (widgets_p, "xffm/stock_run", NULL);
        rfm_diagnostics (widgets_p, "xffm_tag/command_id", g, NULL);
        rfm_diagnostics (widgets_p, "xffm_tag/command", command, "\n", NULL);
        g_free (g);
    }

    if(widgets_p && child > 0) {
        setup_run_button_thread (widgets_p, command, child);
    }
    g_free (command);
    g_free (term);
    return;
}


static gchar *
substitute(const gchar *string, const gchar *expression, const gchar *replacement) {
   GRegex *regex = g_regex_new (expression, 0, 0, NULL);
   gchar *newstring = g_regex_replace_literal (regex,
                                   string, -1, 0,
				   replacement,
				    0 , NULL);
		g_regex_unref (regex);
    if (!newstring) newstring=g_strdup(string);
    return newstring;
}

static gchar *
compact_line(const gchar *line){
    gchar **p;
    gchar *charlist[]={">","<",  "&&", "||",  ";",NULL};
    //gchar *newline= g_strdup(line); 
    gchar *newline= substitute(line, "\\s\\s", " "); 

    for (p=charlist; p && *p; p++) {
        if (strstr(line, *p)){
	     gchar *exp=g_strdup_printf("\\s%s", *p);
	     gchar *newerline= substitute(newline, exp, *p);
	     g_free(exp);
	     g_free(newline);

	     newline=newerline;
	     exp=g_strdup_printf("%s\\s", *p);
	     newerline= substitute(newline, exp, *p);
	     g_free(exp);
	     g_free(newline);

	     newline=newerline;
	     exp=g_strdup_printf(" %s ", *p);
	     newerline= substitute(newline, *p, exp);
	     g_free(exp);
	     g_free(newline);
	     newline=newerline;

	    TRACE("%s ===>>> %s\n", line, newline);fflush(NULL);
        }
    }
    g_strstrip(newline);
    return newline;
	    
}

static gboolean
is_in_history_list(GList *list, const gchar *string){
    GList *tmp=g_list_first(list);
    for (;tmp && tmp->data; tmp=tmp->next){
	if (strcmp((gchar *)tmp->data, string)==0) {
	    return TRUE;
	}
    }
    return FALSE;
}


/*******************************************/
