/*
 * 
 * copywrite 2003-2010 under GNU/GPL which means what it means.
 * Edscott Wilson Garcia 
 *
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define CPY_CANCEL      0
#define CPY_YES         1
#define CPY_YES_TO_ALL  2
#define CPY_NO          3
#define CPY_NO_TO_ALL   4

static gboolean
on_destroy_event (GtkWidget * widget, GdkEvent * event, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    gtk_main_quit ();
    gtk_widget_destroy (widgets_p->remove);
    widgets_p->remove = GINT_TO_POINTER (CPY_CANCEL);
    return TRUE;
}

static void
on_copy_no_activate (GtkButton * button, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    GtkWidget *check = rodent_get_widget_by_name (widgets_p->remove, "togglebutton1");
    int result = CPY_NO;
    if(gtk_toggle_button_get_active ((GtkToggleButton *) check)) {
        result = CPY_CANCEL;
    }
    gtk_main_quit ();
    gtk_widget_destroy (widgets_p->remove);

    widgets_p->remove = GINT_TO_POINTER (result);

}

static void
on_copy_yes_activate (GtkButton * button, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    GtkWidget *check = rodent_get_widget_by_name (widgets_p->remove, "togglebutton1");
    int result = CPY_YES;
    if(gtk_toggle_button_get_active ((GtkToggleButton *) check)) {
        result = CPY_YES_TO_ALL;
    }
    gtk_main_quit ();
    gtk_widget_destroy (widgets_p->remove);
    widgets_p->remove = GINT_TO_POINTER (result);
}

static int
make_overwrite_dialog (widgets_t * widgets_p, const gchar * target, const gchar * src) {
    gchar *mess;
    struct stat st;

    widgets_p->remove = rodent_create_remove (widgets_p);

    g_signal_connect (rodent_get_widget_by_name (widgets_p->remove, "cancelbutton"),
                      "clicked", G_CALLBACK (on_copy_no_activate), widgets_p);
    g_signal_connect (rodent_get_widget_by_name (widgets_p->remove, "removebutton"),
                      "clicked", G_CALLBACK (on_copy_yes_activate), widgets_p);

    g_signal_connect (widgets_p->remove, "delete-event", G_CALLBACK (on_destroy_event), widgets_p);
    g_signal_connect (widgets_p->remove, "destroy-event", G_CALLBACK (on_destroy_event), widgets_p);

    /* dialog specifics */


        gtk_label_set_text ((GtkLabel *)
                            rodent_get_widget_by_name (widgets_p->remove, "label16"), _("Warning"));
        HIDE_IT (widgets_p->remove, "warning");

    gtk_label_set_text ((GtkLabel *)
                        rodent_get_widget_by_name (widgets_p->remove, "label22"), _("Overwrite"));

    lstat (target, &st);
    gchar *ss1 = rfm_time_to_string (st.st_mtime);
    gchar *ss2 = rfm_sizetag ((off_t) st.st_size, -1);
    if(src) {
        struct stat s_st;
        gchar *t1 = NULL,
            *t2 = NULL;
        gchar *s1 = NULL,
            *s2 = NULL;
        if(lstat (src, &s_st) < 0) {
            g_warning ("cannot stat %s", src);
        }
        {
            s1 = rfm_time_to_string (s_st.st_mtime);
            s2 = rfm_sizetag ((off_t) s_st.st_size, -1);
            t1 = rfm_utf_string (target);
            t2 = rfm_utf_string (src);
        }

        mess = g_strdup_printf ("%s:\n%s\n(%s %s)\n--->\n%s\n(%s %s)",
                _("Overwrite"), 
                t2, s1, s2,
                t1, ss1, ss2);
        g_free (s1);
        g_free (s2);
        g_free (t1);
        g_free (t2);
    } else {
        mess = g_strdup_printf ("%s\n(%s %s)", target, ss1, ss2);
    }

    g_free (ss1);
    g_free (ss2);
    gtk_label_set_text ((GtkLabel *)
                        rodent_get_widget_by_name (widgets_p->remove, "label20"), mess);
    g_free (mess);
    mess = NULL;

    gtk_widget_realize (widgets_p->remove);
    gtk_widget_show (widgets_p->remove);

    gtk_widget_grab_focus (rodent_get_widget_by_name (widgets_p->remove, "removebutton"));
    view_t *view_p=widgets_p->view_p;
    if(view_p->type != DESKVIEW_TYPE) {
        gtk_window_set_transient_for (GTK_WINDOW (widgets_p->remove), GTK_WINDOW (widgets_p->window));
    }

    gtk_main ();
    int result = GPOINTER_TO_INT (widgets_p->remove);
    widgets_p->remove = NULL;
    return result;
}

static gchar *
mktgpath (const gchar * target_dir, char *source) {
    gchar *name = g_path_get_basename (source);
    gchar *target = g_build_filename (target_dir, name, NULL);
    g_free (name);
    return target;
}

static int
warn_target_exists (widgets_t * widgets_p, const gchar * target, const gchar * src, gboolean force_override) {
    int result = CPY_YES;
    if(!force_override) {
        result = make_overwrite_dialog (widgets_p, target, src);
    }
    return result;
}

static int
ok_input (widgets_t * widgets_p, char *target, record_entry_t * s_en, gboolean force_override) {
    if(IS_SFIFO (s_en->type) || IS_SCHR (s_en->type)
       || IS_SBLK (s_en->type) || IS_SSOCK (s_en->type)) {
        rfm_show_text(widgets_p);
        rfm_diagnostics (widgets_p, "xffm/error", NULL);
        rfm_diagnostics (widgets_p, "xffm_tag/stderr", strerror (EBADF), " : ", s_en->path, " --> ", target, "\n", NULL);
        return (CPY_NO);
    }

    if(!g_file_test (target, G_FILE_TEST_EXISTS)) {
        return (CPY_YES);
    }
    if(getenv ("VERSION_CONTROL")
       && strlen (getenv ("VERSION_CONTROL"))
       && strcmp (getenv ("VERSION_CONTROL"), "none") == 0) {
        return warn_target_exists (widgets_p, target, s_en->path, force_override);
    }
    return CPY_YES;
}

static void
verify_list (widgets_t * widgets_p, GList * in_list, GList ** out_list, const gchar * target_path) {
    gboolean force_override = FALSE;
    *out_list = NULL;
    if(g_list_length (in_list) > 1 && !g_file_test (target_path, G_FILE_TEST_IS_DIR)) {
        rfm_diagnostics (widgets_p, "xffm/warning", NULL);
        rfm_diagnostics (widgets_p, "xffm_tag/stderr", target_path, ": ", strerror (ENOTDIR), "\n", NULL);
        //g_warning("target must be a directory: %s",target_path);
        return;
    }
    GList *list;
    for(list = g_list_first (in_list); list; list = list->next) {
        gchar *url = list->data;
        if(!url) {
            g_warning ("url !=NULL not met\n");
            continue;
        }
        record_entry_t *s_en = rfm_stat_entry (url, 0);
        if(!s_en) {
            g_warning ("rodent_cpy: s_en is NULL!\n");
            continue;
        }

        gchar *target = mktgpath (target_path, s_en->path);
        DBG ("PASTE CP: target=%s\n", target_path);

        switch (ok_input (widgets_p, target, s_en, force_override)) {
        case CPY_NO:
            rfm_diagnostics (widgets_p, "xffm_tag/green", _("Cancel"), " :", s_en->path, "\n", NULL);
            rfm_destroy_entry (s_en);
            break;
        case CPY_CANCEL:       /* dnd cancelled */
            rfm_diagnostics (widgets_p, "xffm/warning", strerror (ECANCELED), "\n", NULL);
            g_list_free (*out_list);
            *out_list = NULL;
            rfm_destroy_entry (s_en);
            g_free (target);
            return;
        case CPY_YES:
        case CPY_YES_TO_ALL:
        default:
            *out_list = g_list_append (*out_list, s_en);
            break;
        }
        g_free (target);
    }

}
