
/*
 * Copyright (C) 2002-2010 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*************************************************************************/
/******************   dnd functions *************************************/
/************************************************************************/

#define MAXURILEN 4096          /* Longest URI to allow */

#define DRAG_TYPE_UNDEFINED	0
#define DRAG_TYPE_LOCAL		0x01
#define DRAG_TYPE_NET		0x02
#define DRAG_TYPE_INCONSISTENT	0x04

static GtkTargetEntry target_table[] = {
    {"text/uri-list", 0, TARGET_URI_LIST},
    {"text/x-moz-url", 0, TARGET_MOZ_URL},
    {"text/plain", 0, TARGET_PLAIN},
    {"UTF8_STRING", 0, TARGET_UTF8},
    {"STRING", 0, TARGET_STRING}
};

#define NUM_TARGETS (sizeof(target_table)/sizeof(GtkTargetEntry))
static view_t *drag_view_p=NULL;

static GMutex *tooltip_text_mutex=NULL;
enum {
    DND_NONE,
    DND_MOVE,
    DND_COPY,
    DND_LINK
};


static gpointer
tip_preview_thread_f (gpointer data);

typedef struct activate_tip_t {
    view_t *view_p;
    population_t *population_p;
    int population_serial;
} activate_tip_t;
    

/*static GList *selection_list = NULL;*/
static char *dnd_data = NULL;
/*************   core *****************/
#if 0
void
on_drag_data_delete (GtkWidget * widget, GdkDragContext * context, gpointer data) {
    DBG ("TRACE: on_drag_data_delete!\n\n");

    return;
}
#endif

static GdkPixbuf *
fix_pixbuf_scale(GdkPixbuf *in_pixbuf){
    if (!GDK_IS_PIXBUF(in_pixbuf)) return NULL;
    GdkPixbuf *out_pixbuf=in_pixbuf;
    int height = gdk_pixbuf_get_height (in_pixbuf);
    int width = gdk_pixbuf_get_width (in_pixbuf);

    // this is to fix paper size previews for text files and pdfs
    if((width < height && height != PREVIEW_IMAGE_SIZE) || 
	(width >= height && width != PREVIEW_IMAGE_SIZE)) {
        //scale it to size
	if (width < height && height != PREVIEW_IMAGE_SIZE) {
	    width = gdk_pixbuf_get_width (in_pixbuf) * PREVIEW_IMAGE_SIZE / height;
	    out_pixbuf = gdk_pixbuf_scale_simple (in_pixbuf, width, PREVIEW_IMAGE_SIZE, GDK_INTERP_HYPER);
	} else {
	    height = gdk_pixbuf_get_height (in_pixbuf) * PREVIEW_IMAGE_SIZE / width;
	    out_pixbuf = gdk_pixbuf_scale_simple (in_pixbuf, PREVIEW_IMAGE_SIZE, height, GDK_INTERP_HYPER);
	}
        g_object_unref(in_pixbuf);
    }

    return out_pixbuf;
}

static GdkPixbuf *
load_preview_pixbuf_from_disk (const gchar * path, const gchar * thumbnail) {
    GError *error = NULL;
    g_mutex_lock (thumbnail_mutex);
    GdkPixbuf *pixbuf = gdk_pixbuf_new_from_file (thumbnail, &error);
    // pixbuf has ref==1
    g_mutex_unlock (thumbnail_mutex);
    if(error) {
        g_warning ("gdk_pixbuf_new_from_file(%s -> %s): %s", path, thumbnail, error->message);
        g_error_free (error);
        return NULL;
    }
    if (!GDK_IS_PIXBUF(pixbuf)) return NULL;
    GdkPixbuf *original=pixbuf;
    pixbuf=fix_pixbuf_scale(pixbuf);
    if (original != pixbuf) {
        g_mutex_lock (thumbnail_mutex);
        gdk_pixbuf_save (pixbuf, thumbnail, "png", NULL, "tEXt::Software", "Rodent", NULL);
        g_mutex_unlock (thumbnail_mutex);
    }
    return pixbuf;
}

static void
gui_drag_end (GtkWidget * widget, GdkDragContext * context, gpointer data) {
    DBG ("*DND drag end\n");
    /* check if this is in treeviewlib, then delete:
     * if (scrolltimer)g_source_remove(scrolltimer);
     scrolltimer=0,scrollx=-1, scrolly=-1;*/
    if(!widget)
        return;

    if(dnd_data) {
        g_free (dnd_data);
        dnd_data = NULL;
    }
    DBG ("DND drag end done\n");
    return;
}

/*
 * DND sender: prepare data for the remote
 * event: drag_data_get
 */
static void
gui_drag_data_get (widgets_t * widgets_p,
                   GList * drag_entry_list, GdkDragContext * context, GtkSelectionData * selection_data, guint info, guint time) {
    char *files;
    GList *tmp;
    record_entry_t *en;
    int selection_len;
    int drag_type;
    gchar *format = NULL;
    gchar *me,
     *she;

    if(!drag_entry_list || !g_list_length (drag_entry_list)
       || !drag_entry_list->data) {
        DBG ("--DND !SELECTION_LIST\n");
        return;
    }
    en = (record_entry_t *) drag_entry_list->data;

    me = g_strdup (g_get_host_name ());
    she = rfm_host_name (GDK_WINDOW_XID (context->dest_window));

    view_t *view_p = widgets_p->view_p;
    if(view_p->en) {
        DBG ("*DND on_drag_data_get (%s), me=%s --> she=%s\n", view_p->en->path, me, she);
    }

    if(en->module) {
        const gchar *fmt = rfm_natural ("rfm", en->module, en, "get_dnd_format");
        if(fmt)
            format = g_strdup (fmt);
        DBG ("DND module format=%s\n", (format ? format : "null"));
    } else
        DBG ("DND not module format\n");
    if(!format) {
        drag_type = DRAG_TYPE_LOCAL;
        if(strcmp (me, she)) {
            struct passwd *pw = getpwuid (getuid ());
            if(pw) {
                format = g_strdup_printf ("file://%s@%s", pw->pw_name, me);
            } else {
                format = g_strdup_printf ("file://%s", me);
            }
        } else
            format = g_strdup ("file:");
    }
    g_free (me);
    g_free (she);
    DBG ("DND on_drag_data_get, format=%s\n", (format ? format : "null"));

    /* prepare data for the receiver */
    switch (info) {
    case TARGET_RAW:
        g_warning ("TARGET_RAW");
    case TARGET_UTF8:
        g_warning ("TARGET_UTF8");
    case TARGET_URI_LIST:
        DBG ("DND TARGET_URI_LIST\n");
    default:
        selection_len = 0;
        if(dnd_data) {
            g_free (dnd_data);
            dnd_data = NULL;
        }
        /* count length of bytes to be allocated */
        for(tmp = drag_entry_list; tmp; tmp = tmp->next) {
            const gchar *dndpath;
            en = (record_entry_t *) tmp->data;
            if(!en || !en->path || !strlen (en->path))
                continue;
            if(en->module) {
                dndpath = rfm_natural ("rfm", en->module, en, "get_dnd_path");
                if(!dndpath)
                    continue;
            } else {
                dndpath = en->path;
            }
            /* 2 is added for the \r\n */
            selection_len += (strlen (dndpath) + strlen (format) + 2);
        }
        /* 1 is added for terminating null character */
        dnd_data = files = g_malloc (selection_len + 1);
        memset (files, 0, selection_len + 1);
        for(tmp = drag_entry_list; tmp; tmp = tmp->next) {
            const gchar *dndpath;
            en = (record_entry_t *) tmp->data;
            if(!en || !en->path || !strlen (en->path))
                continue;
            if(en->module) {
                dndpath = rfm_natural ("rfm", en->module, en, "get_dnd_path");
                if(!dndpath)
                    continue;
            } else {
                dndpath = en->path;
            }
            sprintf (files, "%s%s\r\n", format, dndpath);
            files += (strlen (format) + strlen (dndpath) + 2);
        }
        break;
    }
    DBG ("DND drag data is:%s\n", dnd_data);

    gtk_selection_data_set (selection_data, selection_data->target, 8, (const guchar *)dnd_data, selection_len);
    /*rfm_details->comm |= COMM_IGNORE_RELEASE;
    DBG ("DND drag get done, format=%s\n", format);
    DBG ("DND ignore second releas=%d... \n", rfm_details->comm & COMM_IGNORE_RELEASE);*/
    g_free (format);
}

static gboolean
gui_drag_data (widgets_t * widgets_p,
               record_entry_t * target_en,
               GdkDragContext * context, gint x, gint y, GtkSelectionData * data, guint info, guint time) {
    int the_mode;
    int nitems,
      action;
    gchar *url;
    int mode = 0;
    GList *list = NULL;
    gboolean result = FALSE;
    gchar *he,
     *me;

    if(!target_en || data->length < 0 || data->format != 8) {
        DBG ("--DND !target_en || data->length < 0 || data->format != 8\n");
        goto drag_over;         /* of course */
    }

    me = g_strdup (g_get_host_name ());
    he = rfm_host_name (GDK_WINDOW_XID (context->source_window));

    view_t *view_p = widgets_p->view_p;
    if(view_p->en) {
        DBG ("*DND on_drag_data_get (%s), me=%s --> she=%s\n", view_p->en->path, me, he);
    }

    /* remote instance may have full format specification,
     * or borked specification. Here we must consider
     * both cases */

    if(context->action <= GDK_ACTION_DEFAULT) {
        if(getenv ("RFM_DRAG_DOES_COPY")
           && strlen (getenv ("RFM_DRAG_DOES_COPY"))) {
            action = GDK_ACTION_COPY;
        } else {
            action = GDK_ACTION_MOVE;
        }
    } else
        action = context->action;

    DBG ("DND info=%d (%d,%d)\n", info, TARGET_STRING, TARGET_URI_LIST);
    if(!(info == TARGET_STRING) && !(info == TARGET_URI_LIST)
       && !(info == TARGET_MOZ_URL)) {
        goto drag_over;         /* of course */
    }

    DBG ("DND: action=%d\n", action);
    if(action == GDK_ACTION_MOVE)
        the_mode = mode = TR_MOVE;
    else if(action == GDK_ACTION_COPY)
        the_mode = mode = TR_COPY;
    else if(action == GDK_ACTION_LINK)
        the_mode = mode = TR_LINK;
    else {
        rfm_show_text(widgets_p);
        rfm_diagnostics (widgets_p, "xffm/error", NULL);
        rfm_diagnostics (widgets_p, "xffm_tag/stderr", strerror (EINVAL), "\n", NULL);
        goto drag_over;         /* of course */
    }

    DBG ("DND drag data=%s", (const char *)data->data);

    nitems = rfm_uri_parse_list ((const char *)data->data, &list);

    DBG ("DND nitems=%d, drag data=%s", nitems, (const char *)data->data);
    if(!nitems) {
        goto drag_over;         /* of course */
    }

    /***/


    /* if target is a plugin, let the plugin take care of business. */
    if(target_en->module) {
        rfm_uri_remove_file_prefix_from_list (list, he, me);
        DBG ("en->module=%s\n", target_en->module);
        if(rfm_natural ("rfm", target_en->module, target_en, "valid_drop_site")) {
            DBG ("module: valid_drop_site for %s\n", target_en->module);
            rfm_natural ("rfm", target_en->module, target_en, "set_drop_entry");
            if(rfm_rational ("rfm", target_en->module, list, widgets_p, "process_drop")) {
                DBG ("module: process_drop ok\n");
                result = TRUE;
            }
            rfm_void ("rfm", target_en->module, "clear_drop_entry");
            list = rfm_uri_free_list (list);
            goto drag_over;
        }
    }

    /* now determine whether cp or scp should be used 
     * (we ignore the format for this determination, 
     * because other applications may bork this)*/
    rfm_uri_remove_file_prefix_from_list (list, he, me);
#if 0
    if(strcmp (me, he) != 0) {
        int l;
        char **srcs;

        DBG ("dnd: doing scp\n");

        l = g_list_length (list);
        srcs = (char **)malloc ((l + 1) * sizeof (char *));
        if(!srcs) {
            list = rfm_uri_free_list (list);
            DBG ("dnd: no srcs\n");
            goto drag_over;
        }
        srcs[l] = NULL;
        for(l = 0, tmp = list; tmp; tmp = tmp->next, l++) {
            srcs[l] = tmp->data;
        }
        rodent_scp (widgets_p, (const gchar **)srcs, target_en->path);
        g_free (srcs);
        list = rfm_uri_free_list (list);
        result = TRUE;
        goto drag_over;
    }
#endif

    /* local file cp/ln/mv */
    url = list->data;
    {
        /* nonsense check */
        struct stat st;
        lstat (url, &st);
        if(target_en->st && st.st_ino == target_en->st->st_ino) {
            list = rfm_uri_free_list (list);
            rfm_diagnostics(&(view_p->widgets),"xffm/warning",NULL);
            rfm_diagnostics (widgets_p, "xffm_tag/stderr", " ", strerror (EEXIST), ": ", target_en->path, "\n", NULL);
            /*fprintf(stderr,"dbg:nonsense 1\n"); */
            goto drag_over;
        }
        if(!S_ISDIR (st.st_mode) && strchr (url, '/')) {
            char *p;
            p = g_strdup (url);
            *(strrchr (p, '/')) = 0;
            if(target_en->path && strcmp (p, target_en->path) == 0) {
                list = rfm_uri_free_list (list);
                g_free (p);
                p = NULL;
                rfm_show_text(widgets_p);
                rfm_diagnostics (widgets_p, "xffm/error", NULL);
                rfm_diagnostics (widgets_p, "xffm_tag/stderr", strerror (EEXIST), ": ", target_en->path, "\n", NULL);

                /*fprintf(stderr,"dbg:nonsense 2\n"); */
                goto drag_over;
            }
            g_free (p);
            p = NULL;
        }
    }

    DBG ("DND: local\n");

    rodent_cp (mode, widgets_p, list, target_en->path);

    list = rfm_uri_free_list (list);
    result = TRUE;
  drag_over:
    g_free (me);
    g_free (he);
    gtk_drag_finish (context, TRUE, (the_mode & TR_MOVE) ? TRUE : FALSE, time);
    DBG ("DND: on drag data: drag_over\n");
    return result;
}

/*************************************************************************/
/******************   end of dnd functions *******************************/
/*************************************************************************/

/*************************************************************************/
/******************* tip functions ***************************************/
/*************************************************************************/

static void hide_tip (view_t * view_p);

//static
//void show_tip (population_t * population_p);
#ifdef FAST_PROCESSOR
# undef FAST_PROCESSOR
#endif

static gchar *
path_info (record_entry_t * en) {
    gchar *s1 = NULL,
        *s2 = NULL;
    gchar *info = NULL;
    if(!en || !en->path)
        return NULL;
    if(IS_ROOT_TYPE (en->type) && !g_file_test (en->path, G_FILE_TEST_EXISTS))
        return NULL;

#ifdef FAST_PROCESSOR
    if (!en->mimetype) en->mimetype=MIME_type (en->path, en->st);
    if (!en->mimemagic) en->mimemagic= MIME_magic(en->path);
    if (!en->filetype) en->filetype= MIME_file(en->path);
    /* printout full file information: */
#endif

    gchar *b = g_path_get_basename (en->path);
    rfm_chop_excess (b);
    gchar *vpath = rfm_utf_string (b);
    g_free (b);
    if(IS_SLNK (en->type)) {
        char lpath[_POSIX_PATH_MAX + 1];
        memset (lpath, 0, _POSIX_PATH_MAX + 1);
        if(readlink (en->path, lpath, _POSIX_PATH_MAX) > 0) {
            rfm_chop_excess (lpath);
            gchar *q = rfm_utf_string (lpath);
            gchar *linkto=g_strdup_printf (_("Link to %s"), q);
            s1 = g_strdup_printf ("%s -> %s\n\n", vpath, linkto);
            g_free(linkto);
            //s1 = g_strdup_printf (_("Symbolic link: %s -> %s\n\n"), vpath, q);
            g_free (q);
        }
    } else {
        gchar *p = g_strdup_printf ("%s %s\n\n",_("Name:"), vpath);
        if(en->module) {
            gchar *q = g_strdup_printf ("%s: %s", _("Installed Plugins"), en->module);
            s1 = g_strconcat (p, q, NULL);
            g_free (p);
            g_free (q);
        } else {
            s1 = p;
        }
    }
    g_free (vpath);

    if(g_file_test (en->path, G_FILE_TEST_EXISTS) && en->st) {
        struct group *g;
        struct passwd *p;
        char *grupo,
         *owner;
        gchar *tag = rfm_sizetag ((off_t) en->st->st_size, -1);
        if((g = getgrgid (en->st->st_gid)) != NULL)
            grupo = g->gr_name;
        else if((int)en->st->st_gid < 0)
            grupo = "";
        else
            grupo = "?";
        if((p = getpwuid (en->st->st_uid)) != NULL)
            owner = p->pw_name;
        else if((int)en->st->st_uid < 0)
            owner = "";
        else
            owner = "?";
        //    gchar *ss= rfm_time_to_string(en->st->st_mtime);   
        //    gchar *q = rfm_utf_string(ss);
        //    g_free(ss);
        gchar *dirname = g_path_get_dirname (en->path);
#ifdef FAST_PROCESSOR
        gchar **apps = MIME_apps (en->mimemagic);
        gchar *app_txt = g_strdup ("");;
        gchar *openwith=g_strdup_printf(_("Open with %s"),": ");
        if(apps) {
            gchar **p,
             *pp;
            for(p = apps; p && *p; p++) {
                pp = g_strconcat (app_txt, "\n\t", *p, NULL);
                g_free (app_txt);
                app_txt = pp;
            }

        } else {
            app_txt = openwith;
        }
        g_strfreev (apps);
        s2 = g_strdup_printf (
                "%s/%s: %s/%s\n%s: %s\n%s: %s\n\n%s: %s\n%s: %s\n%s: %s%s\n\n%s: %s\n\n%s: %s",
                _("Owner"),_("Group"), owner, grupo,
                _("Permissions"), rfm_mode_string (en->st->st_mode),
                _("Folder"), dirname, 
                _("File Type"), en->filetype,
                _("MIME Type"), (en->mimetype) ? en->mimetype : _("unknown"),
                _("MIME Type"),  " (file)", en->mimemagic,
                openwith,  app_txt,
                _("Size"),  tag);

        g_free (app_txt);
        g_free (openwith);

#else
        s2 = g_strdup_printf (
                "%s/%s: %s/%s\n%s: %s\n%s: %s\n\n%s: %s",
                 _("Owner"),_("Group"), owner, grupo,
                _("Permissions"), rfm_mode_string (en->st->st_mode),
                _("Folder"), dirname, 
                _("Size"),  tag);
#endif
        //    g_free(q);
        g_free (tag);
        g_free (dirname);
    }
    if(g_file_test (en->path, G_FILE_TEST_EXISTS) && en->st) {
        gchar buf[1024];
#ifdef HAVE_LOCALTIME_R
        struct tm t_r;
#endif
        struct tm *t;
#ifdef HAVE_LOCALTIME_R
        t = localtime_r (&en->st->st_ctime, &t_r);
#else
        t = localtime (&en->st->st_ctime);
#endif
        sprintf (buf, "Status change= %04d/%02d/%02d  %02d:%02d", t->tm_year + 1900,
                 t->tm_mon + 1, t->tm_mday, t->tm_hour, t->tm_min);

        gchar *s3 = g_strconcat (s2, "\n", buf, NULL);
        g_free (s2);
        s2 = s3;
#ifdef HAVE_LOCALTIME_R
        t = localtime_r (&en->st->st_mtime, &t_r);
#else
        t = localtime (&en->st->st_mtime);
#endif
        sprintf (buf, "Modification time= %04d/%02d/%02d  %02d:%02d", t->tm_year + 1900,
                 t->tm_mon + 1, t->tm_mday, t->tm_hour, t->tm_min);

        s3 = g_strconcat (s2, "\n", buf, NULL);
        g_free (s2);
        s2 = s3;
#ifdef HAVE_LOCALTIME_R
        t = localtime_r (&en->st->st_atime, &t_r);
#else
        t = localtime (&en->st->st_atime);
#endif
        sprintf (buf, "Access time= %04d/%02d/%02d  %02d:%02d", t->tm_year + 1900,
                 t->tm_mon + 1, t->tm_mday, t->tm_hour, t->tm_min);

        s3 = g_strconcat (s2, "\n", buf, NULL);
        g_free (s2);
        s2 = s3;

        s3 = g_strdup_printf ("%s\n\nNumber of hard links: %ld", s2, (long)en->st->st_nlink);
        g_free (s2);
        s2 = s3;
    }

    if(!s1)
        s1 = g_strdup ("");
    if(!s2)
        s2 = g_strdup ("");
    info = g_strconcat (s1, s2, NULL);
    g_free (s1);
    g_free (s2);
#ifdef FAST_PROCESSOR
    g_free (mimetype);
    g_free (mimemagic);
    g_free (filetype);
#endif
    return info;
}

static gchar *
get_tip_text (view_t * view_p, population_t * population_p, const gchar * prepend_txt) {
    DBG ("TIP: SHOW_TIP population_p->preview_pixbuf = 0x%lx\n", (long unsigned)view_p->tooltip_pixbuf);
    const gchar *warning = NULL;
    gchar *info = path_info (population_p->en);
    gchar *g;
    if(population_p->en && IS_SDIR (population_p->en->type)) {
        warning = _("Folder");
//        warning = _("Directory:\n");
        gchar *b = g_path_get_basename (population_p->en->path);
        int files = rfm_xfdir_count_files (population_p->en->path);
        int hidden = rfm_xfdir_count_hidden_files (population_p->en->path);
        gchar *info = path_info (population_p->en);
        gchar *hidden_string = NULL;
        gchar *files_string = NULL;

        gchar *plural_string = g_strdup_printf (ngettext (" (containing %'d item)", " (containing %'d items)", files),files);
        files_string = g_strdup_printf ("%s%s.",_("Folder"), plural_string);
        g_free(plural_string);
    
        plural_string = g_strdup_printf(ngettext ("%'u item","%'u items",hidden), hidden);
        hidden_string = g_strdup_printf ("%s: %s.",_("Hidden"), plural_string);
        g_free(plural_string);

        if(files) {
            g = g_strdup_printf ("%s %s\n%s", files_string, hidden_string, info);
        } else {
            g = g_strdup_printf ("%s: %s\n%s", _("Folder"), _("The location is empty."), info);
        }

        g_free (hidden_string);
        g_free (files_string);
        g_free (b);
    } else {
        if (warning) g = g_strdup_printf ("%s:\n%s", warning, info);
        else  g = g_strdup_printf ("%s", info);
    }
    // modules can prepend text to the tooltip
    if(prepend_txt) {
        gchar *gg = g_strconcat (prepend_txt, "\n", g, NULL);
        g_free (g);
        g = gg;
    }
    g_free (info);
    return g;
}

static void
normal_tip(population_t * population_p, gboolean full){
    DBG ("normal_tip\n");
    view_t *view_p = population_p->view_p;
    
    if(population_p == NULL || !population_p->en) {
	DBG ("population_p == NULL\n");
        return;
    }
    if(getenv ("RFM_DISABLE_TIPS") && strlen (getenv ("RFM_DISABLE_TIPS"))) {
        return;
    }

    gchar *module_txt = NULL;
    if(view_p->module_name) {
        module_txt = rfm_natural ("rfm", view_p->module_name, population_p->en, "entry_tip");
    }

    int isize;
    if (!tooltip_text_mutex){
	tooltip_text_mutex=g_mutex_new();
    }
    g_mutex_lock(tooltip_text_mutex);
    gchar *oldtext=view_p->tooltip_text;
    if (full){
        view_p->tooltip_text = get_tip_text (view_p, population_p, module_txt);
        if(rfm_svg_supported ()) {
            isize = BIG_ICON_SIZE;
        } else {
            isize = view_p->icon_size;
        }
    } else {
           gchar *basename=g_path_get_basename(population_p->en->path);
           view_p->tooltip_text = rfm_utf_string(basename);
           g_free(basename);
           isize = SIZE_BUTTON;
    }
	
    g_free (oldtext);
    g_free (module_txt);

    if (view_p->tooltip_pixbuf) g_object_unref(view_p->tooltip_pixbuf);
    if(view_p->icon_size == BIG_ICON_SIZE) {
        view_p->tooltip_pixbuf = population_p->normal_pixbuf;
    } else {
        view_p->tooltip_pixbuf = rfm_get_pixbuf (rfm_resolve_icon_id (population_p->en), isize);
	// view_p->tooltip_pixbuf: ref==1
    }
    // additional ref to pixbuf
    g_object_ref(view_p->tooltip_pixbuf);
    // view_p->tooltip_pixbuf: ref==2 (first is in pixbuf hash)
    record_entry_t *olden=view_p->tooltip_entry;
    view_p->tooltip_entry = rfm_copy_entry (population_p->en);
    rfm_destroy_entry(olden);
    TRACE ("TIP: SHOW_TIP %s -> %s\n", (population_p->en) ? population_p->en->path : "Null entry", view_p->tooltip_text);
    view_p->tooltip_active = TRUE;
    g_mutex_unlock(tooltip_text_mutex);
    
}

static gboolean 
preliminary_tip(activate_tip_t *activate_tip_p){
    DBG("preliminary_tip\n");
    view_t *view_p=activate_tip_p->view_p;
    int population_serial=activate_tip_p->population_serial;
    rfm_population_read_lock (view_p);
    if (view_p->population_serial != population_serial){
	rfm_population_read_unlock (view_p);
	return FALSE;
    }
    gboolean retval=TRUE;
    if (activate_tip_p->population_p->en->mimemagic==NULL){
	activate_tip_p->population_p->en->mimemagic=
	    MIME_magic(activate_tip_p->population_p->en->path);
    }
    population_t *population_p = (population_t *)malloc(sizeof(population_t));
    memcpy(population_p, activate_tip_p->population_p, sizeof(population_t));
    population_p->en = rfm_copy_entry(activate_tip_p->population_p->en);

    const gchar *type=(population_p->en->mimetype)?
	population_p->en->mimetype:population_p->en->mimemagic;
    if (population_p->preview_pixbuf) {
	g_object_ref(population_p->preview_pixbuf);
	// population_p->preview_pixbuf: ref==2
    }
    if (population_p->initial_preview_pixbuf) {
	g_object_ref(population_p->initial_preview_pixbuf);
	// population_p->initial_preview_pixbuf: ref==2
    }
    
    int width=gdk_pixbuf_get_width(population_p->normal_pixbuf);
    int height=gdk_pixbuf_get_height(population_p->normal_pixbuf);

    rfm_population_read_unlock (view_p);

    // is the work already done?
    gboolean preview_ok = (population_p->preview_pixbuf
		       && population_p->preview_pixbuf_timestamp <= 
		          population_p->en->st->st_mtime);
    if (!tooltip_text_mutex){
	tooltip_text_mutex=g_mutex_new();
    }
    if(preview_ok) {
	g_mutex_lock(tooltip_text_mutex);
	if (view_p->tooltip_pixbuf) {
	    g_object_unref(view_p->tooltip_pixbuf);
	}
	// population_p->preview_pixbuf is already reffed above
	view_p->tooltip_pixbuf = population_p->preview_pixbuf;
	g_object_ref(view_p->tooltip_pixbuf);
	// view_p->tooltip_pixbuf: ref==3
	// the second belongs to the public population_p
	// the third belongs to the private population_p
	record_entry_t *olden=view_p->tooltip_entry;
	view_p->tooltip_entry = rfm_copy_entry (population_p->en);
	rfm_destroy_entry(olden);
	if (view_p->tooltip_text) {
	    g_free(view_p->tooltip_text);
	    view_p->tooltip_text=NULL;
	}
        view_p->tooltip_active = TRUE;
	g_mutex_unlock(tooltip_text_mutex);
	retval=FALSE;
	goto cleanup;
    } 
    else {
	DBG("###### preview not ok %ld<%ld not met\n",
		(long)population_p->preview_pixbuf_timestamp,
		(long)population_p->en->st->st_mtime);
    }
    // initial image:
    // This creates a reffed pixbuf that must be unreffed sooner or later.
    if (population_p->initial_preview_pixbuf == NULL) {
	if (width != height){
	    // this is a definite thumbnail
	    if (width > height) {
		height = height * PREVIEW_IMAGE_SIZE / width;
		width = PREVIEW_IMAGE_SIZE;
	    } else {
		width = width * PREVIEW_IMAGE_SIZE / height;
		height = PREVIEW_IMAGE_SIZE;
	    }
	    population_p->initial_preview_pixbuf = 
		    gdk_pixbuf_scale_simple (population_p->normal_pixbuf,
			width,
			height,
			GDK_INTERP_NEAREST);
	    // population_p->initial_preview_pixbuf: ref==1

	} else {
	    // scale up svg icon (ignore square thumbnails)
	    // the ref is not held in the pixbuf hash since 
	    // the size is larger than 
	    // the maximum pixbuf hash size. 
	    GdkPixbuf *src=rfm_get_pixbuf(type, PREVIEW_IMAGE_SIZE);
	    // src: ref==1 but is not in pixbuf hash, because of size
	    // PREVIEW_IMAGE_SIZE  >= MAX_PIXBUF_HASH_SIZE 
	    // (400 > 128 last time I looked)
            if (PREVIEW_IMAGE_SIZE  <= MAX_PIXBUF_HASH_SIZE){
		// this ref will be held by the pixbuf hash, 
		// so we need an additional ref to unref below 
		g_object_ref (src);
	    }
	    
	    // is it TXT or PDF?
	    // this is deprecated, because pdf files can be in landscape format
	    // so better just use a square preliminary
	    /*
	    if (want_imagemagick_preview (population_p->en)){
		// TXT or PDF document types
		width = 
		    paper_sizes[PAPER_TYPE_US_LETTER].width * PREVIEW_IMAGE_SIZE 
		  / paper_sizes[PAPER_TYPE_US_LETTER].height;
		height = PREVIEW_IMAGE_SIZE;
		population_p->initial_preview_pixbuf =
		    gdk_pixbuf_scale_simple (src,
			width,
			PREVIEW_IMAGE_SIZE,
			GDK_INTERP_BILINEAR);
	    } else */
	    {
		population_p->initial_preview_pixbuf=src;
		g_object_ref (population_p->initial_preview_pixbuf);
	    }
	    g_object_unref(src);
	    if (population_p->initial_preview_pixbuf == NULL) {
		g_warning("population_p->initial_preview_pixbuf == NULL!");
		retval=FALSE;
		view_p->tooltip_active = retval;
		goto cleanup;
	    }
	    rfm_population_read_lock (view_p);
	    if (view_p->population_serial != population_serial){
		retval=FALSE;
		view_p->tooltip_active = retval;
		rfm_population_read_unlock (view_p);
		goto cleanup;
	    }
	    activate_tip_p->population_p->initial_preview_pixbuf=
		population_p->initial_preview_pixbuf;
	    g_object_ref(activate_tip_p->population_p->initial_preview_pixbuf);
	    rfm_population_read_unlock (view_p);
	}
    } // end of population_p->initial_preview_pixbuf == NULL

    g_mutex_lock(tooltip_text_mutex);
    if (view_p->saturated_p && view_p->saturated_p->en &&
	    view_p->saturated_p->en->path &&
	    strcmp(view_p->saturated_p->en->path,population_p->en->path)==0) { 
	// race check
      if (view_p->tooltip_text) {
        g_free(view_p->tooltip_text);
        view_p->tooltip_text=NULL;
      }
      if(view_p->tooltip_pixbuf){
       g_object_unref(view_p->tooltip_pixbuf);
      }
      // population_p->initial_preview_pixbuf has an additional ref
      view_p->tooltip_pixbuf= population_p->initial_preview_pixbuf;
      g_object_ref(view_p->tooltip_pixbuf);
      // the entry will resolve any race condition when the real
      // preview is done.
      record_entry_t *olden=view_p->tooltip_entry;
      view_p->tooltip_entry = rfm_copy_entry (population_p->en);
      rfm_destroy_entry(olden);
    }
    g_mutex_unlock(tooltip_text_mutex);
		
    view_p->tooltip_active = TRUE;

cleanup:
    // private population cleanup
    if(population_p->preview_pixbuf) {
        g_object_unref(population_p->preview_pixbuf);
    }
    if(population_p->initial_preview_pixbuf) {
        g_object_unref(population_p->initial_preview_pixbuf);
    }
    rfm_destroy_entry(population_p->en);
    g_free(population_p);
		

    return retval;
}


static gboolean 
is_normal_condition(population_t *population_p){
    gboolean normal_tips = FALSE;
    view_t *view_p=population_p->view_p;
    if (population_p->en==NULL || population_p->en->st==NULL || population_p->en->st->st_size == 0) {
	DBG("normal_tips; population_p->en==NULL || population_p->en->st==NULL\n");
	normal_tips = TRUE;
    }
    
    if (! (view_p->preferences & __SHOW_IMAGES)) {
	DBG("normal_tips; ! (view_p->preferences & __SHOW_IMAGES)\n");
	normal_tips = TRUE;
    }
    if(g_object_get_data (G_OBJECT (view_p->widgets.paper), "normal_tip")) {
	DBG("normal_tips; g_object_get_data (G_OBJECT (view_p->widgets.paper), normal_tip)\n");
	normal_tips = TRUE;
    }
    gboolean normal_condition=TRUE;
    if (normal_tips) {
	normal_condition=TRUE; // user option
    } else if (population_p->preview_pixbuf) {
	normal_condition=FALSE; // already done
	DBG("already done: %s\n", population_p->en->path);
    } else if (population_p->normal_tip || 
	    !population_p->en || 
	    !population_p->en->st || 
	    population_p->en->st->st_size == 0) {
	DBG("normal_tips; population_p->normal_tip\n");
	normal_condition=TRUE; // unable to do
    }
    else if (rfm_is_image (population_p->en)) {
	normal_condition=FALSE; // condition applies for images
	DBG("rfm_is_image: %s\n", population_p->en->path);
    }
    else if((population_p->en->mimetype && strstr(population_p->en->mimetype, "opendocument"))
	    ||
	    (population_p->en->mimemagic && strstr(population_p->en->mimemagic, "opendocument"))){
	normal_condition=FALSE; // condition applies zipped opendocument files
    }
    else if(want_imagemagick_preview (population_p->en)) {
	normal_condition=FALSE; // condition applies for PDF and TXT
	DBG("want_imagemagick_preview: %s\n", population_p->en->path);
    }
    return normal_condition;
}

static void *
activate_tip_thread(void *data){
    activate_tip_t *activate_tip_p=data;
    DBG("activate_tip_thread\n");

    view_t *view_p = activate_tip_p->view_p;
    int population_serial=activate_tip_p->population_serial;

    // thread memory setup (read lock protected)
    rfm_population_read_lock (view_p);
    if (view_p->population_serial != population_serial){
	rfm_population_read_unlock (view_p);
	return NULL;
    }

    population_t *population_p = (population_t *)malloc(sizeof(population_t));
    memcpy(population_p, activate_tip_p->population_p, sizeof(population_t));
    population_p->en = rfm_copy_entry(activate_tip_p->population_p->en);

    rfm_population_read_unlock (view_p);

    if (is_normal_condition(population_p)) {
	TRACE ("Normal tip\n");
        normal_tip(population_p, TRUE);  
	rfm_destroy_entry(population_p->en);
	g_free(population_p);
        g_free(activate_tip_p);
	return NULL;
    } 
    TRACE ("Not normal tip\n");
    rfm_destroy_entry(population_p->en);
    g_free(population_p);
   //preliminary_tip will set up the preliminary tip preview
    if (preliminary_tip(activate_tip_p)==FALSE) {
	// could not set preliminary tip or full tip is now set
	g_free(activate_tip_p);
    } else {
	// fireup the real preview thread
	DBG("THREAD: g_thread_create(tip_preview_thread_f) at rodent_mouse.c\n");
	// tip_preview_thread_f will free activate_tip_p
	g_thread_create (tip_preview_thread_f, (gpointer) activate_tip_p, FALSE, NULL);
    }
    // end of not normal tips condition
    return NULL;
}


static void
hide_tip (view_t * view_p) {
    //if(!view_p->tooltip_active) return;
    // free tip text and icon
    gtk_widget_set_tooltip_text (view_p->widgets.window, NULL);
    if (!tooltip_text_mutex){
	tooltip_text_mutex=g_mutex_new();
    }
    g_mutex_lock(tooltip_text_mutex);
    if(view_p->tooltip_pixbuf) {
        view_p->tooltip_pixbuf = NULL;
    }
    if(view_p->tooltip_entry) {
        rfm_destroy_entry (view_p->tooltip_entry);
        view_p->tooltip_entry = NULL;
    }
    if(view_p->tooltip_text) {
        g_free (view_p->tooltip_text);
        view_p->tooltip_text = NULL;
    }
    g_mutex_unlock(tooltip_text_mutex);

    DBG ("SHOW_TIP: hide_tip view_p\n");
    view_p->tooltip_active = FALSE;
    gtk_widget_trigger_tooltip_query(view_p->widgets.window);
}


/**
 * rodent_select_pixbuf:
 *
 * Selects the icon which belongs to the parameter @population_p
 * in the iconview window @view_p
 *
 * **/

gboolean
rodent_tip_function (
        GtkWidget * window, 
        gint x, 
        gint y, 
        gboolean keyboard_mode, 
        GtkTooltip * tooltip,
        gpointer user_data
) {
    view_t *view_p = rodent_get_current_view (window);
    DBG ("RODENT_TIP_FUNCTION\n");

    record_entry_t *en = NULL;
    if(view_p->tooltip_entry) {
        en = view_p->tooltip_entry;
    } else {
        DBG ("RODENT_TIP_FUNCTION; view_p->tooltip_entry is NULL\n");
        gtk_widget_set_tooltip_text (view_p->widgets.window, NULL);
	view_p->tooltip_active = FALSE;
        return FALSE;
    }

    if(IS_UP_TYPE (en->type)) {
        DBG ("RODENT_TIP_FUNCTION : IS_UP_TYPE\n");  
        gtk_widget_set_tooltip_text (view_p->widgets.window, NULL);
	view_p->tooltip_active = FALSE;
        return FALSE;           // omit up icons
    }

    if(view_p->tooltip_pixbuf && view_p->tooltip_active) {
        TRACE ("RODENT_TIP_FUNCTION (pixbuf) %s\n", view_p->tooltip_text);
        // this conditional is just to avoid a useless warning on race condition
        // with threads (not really avoid, just reduce possibility).
        if (GDK_IS_PIXBUF (view_p->tooltip_pixbuf)) {
            gtk_tooltip_set_icon (tooltip, view_p->tooltip_pixbuf);
        }
        gtk_tooltip_set_text (tooltip, view_p->tooltip_text);
        return TRUE;
    } else {
        DBG ("RODENT_TIP_FUNCTION tooltip_pixbuf==NULL\n");
        gtk_widget_set_tooltip_text (view_p->widgets.window, NULL);
	view_p->tooltip_active = FALSE;
        return FALSE;
    }
}
static gpointer
tip_preview_thread_f (gpointer data) {
    DBG ("SHOW_TIP: tip_preview_thread_f NOW...\n");
    activate_tip_t *activate_tip_p=data;
    view_t *view_p = activate_tip_p->view_p;
    int population_serial=activate_tip_p->population_serial;
    rfm_population_read_lock (view_p);
    if (view_p->population_serial != population_serial){
	rfm_population_read_unlock (view_p);
	g_free(activate_tip_p);
	return NULL;
    }
    if (activate_tip_p->population_p->preview_busy) {
	DBG ("tip_preview_thread_f activate_tip_p->population_p->preview_busy...\n");
	rfm_population_read_unlock (view_p);
	g_free(activate_tip_p);
	return NULL;
    }
    activate_tip_p->population_p->preview_busy=TRUE;
    population_t *population_p = (population_t *)malloc(sizeof(population_t));
    memcpy(population_p, activate_tip_p->population_p, sizeof(population_t));
    population_p->en = rfm_copy_entry(activate_tip_p->population_p->en);
    if (population_p->preview_pixbuf){
	g_object_ref(population_p->preview_pixbuf);
    }
    rfm_population_read_unlock (view_p);
    // Do the real preview.
    if (!tooltip_text_mutex){
	tooltip_text_mutex=g_mutex_new();
    }
    GdkPixbuf *result=rodent_preview (population_p);
    if (result) {
	// update tooltip pixbuf, subject to
        // race condition check
        gboolean norace = (view_p->tooltip_entry 
            && strcmp(view_p->tooltip_entry->path,population_p->en->path)==0);
        if (norace){
            DBG("SHOW_TIP: population_p->preview_pixbuf OK!\n");
    g_mutex_lock(tooltip_text_mutex);
	    if (view_p->saturated_p && view_p->saturated_p->en 
		    && view_p->saturated_p->en->path &&
		strcmp(view_p->saturated_p->en->path,population_p->en->path)==0) { 
		// race check
		if (view_p->tooltip_pixbuf) {
		    g_object_unref(view_p->tooltip_pixbuf);
		}
		view_p->tooltip_pixbuf = result;
		g_object_ref(view_p->tooltip_pixbuf);
	    }
    g_mutex_unlock(tooltip_text_mutex);
            //view_p->tooltip_entry = rfm_copy_entry (population_p->en);
            view_p->tooltip_active = TRUE;
        } 
    }
    // update public population_p here
    rfm_population_read_lock (view_p);
    if (view_p->population_serial != population_serial){
	rfm_population_read_unlock (view_p);
	goto cleanup;
    }
    if (!result) {
	// busy flag remains TRUE to avoid retries.
	activate_tip_p->population_p->preview_busy = TRUE;
	// normal tip is set to short circuit
	activate_tip_p->population_p->normal_tip = TRUE;
    } else {
	if (view_p->population_serial == population_serial){
	    if (activate_tip_p->population_p->preview_pixbuf) {
		g_object_unref(activate_tip_p->population_p->preview_pixbuf);
	    }
	    activate_tip_p->population_p->preview_pixbuf=result;
	    g_object_ref(activate_tip_p->population_p->preview_pixbuf);
	}
	activate_tip_p->population_p->preview_busy = FALSE;
	activate_tip_p->population_p->preview_pixbuf_timestamp = 
	    activate_tip_p->population_p->en->st->st_mtime;	
    }
    rfm_population_read_unlock (view_p);

    
    //population_p->preview_busy=0;

    // cleanup:
    DBG("SHOW_TIP: tip_preview_thread_f DONE...\n");
cleanup:
    if (result) {
	 g_object_unref(result);
    }
    rfm_destroy_entry(population_p->en);
    if(population_p->preview_pixbuf) {
        g_object_unref (population_p->preview_pixbuf);
    }
 
    g_free(population_p);
    g_free(activate_tip_p);

    return NULL;
}


/*************************************************************************/
/******************* end of tip functions ********************************/
/*************************************************************************/

static void button_popup (GdkEventButton * event, view_t * view_p, population_t * population_p);

static void setup_drag_state (view_t * view_p, GdkEventButton * event);

static void draw_window (view_t * view_p, int x, int y, gboolean draw);

static int select_population_area (view_t * view_p);

static void graphics_saturation_event (view_t * view_p, population_t * population_p, double x, double y);

static void 
enter_drag_state (view_t * view_p) {
    TRACE ("enter dragstate \n");

    if(view_p->dragstate) {
        TRACE ("dragstate true\n");
        return;
    }
    TRACE ("now entering dragstate: event=0x%lx\n", (unsigned long)(&(view_p->drag_event)));
    //TRACE("now entering dragstate: G_IS_OBJECT (event)=%d\n",G_IS_OBJECT (&(view_p->drag_event)));

    view_p->drag_event.context = gtk_drag_begin (view_p->widgets.paper,
                                                 view_p->target_list,
                                                 GDK_ACTION_MOVE |
                                                 GDK_ACTION_COPY |
                                                 GDK_ACTION_LINK, 
						 1, //drag button
						 (GdkEvent *) (&(view_p->drag_event)));

    TRACE ("drag begun...\n");
    if(!view_p->drag_event.context)
        return;
    gdk_drag_status (view_p->drag_event.context, view_p->drag_action, view_p->drag_event.time);
    TRACE ("drag status...\n");

//    TRACE ("selection count=%d\n", view_p->selection_count);

    gchar *plural_text=g_strdup_printf (ngettext ("%'u item", "%'u items",
		g_list_length(view_p->selection_list)),g_list_length(view_p->selection_list));
//                                               view_p->selection_count),view_p->selection_count);
    gchar *g = g_strdup_printf ("%s: %s", _("Selection"), plural_text);
    g_free(plural_text);

    rfm_status (&(view_p->widgets), "xffm/info", g, NULL);
    g_free (g);
    
    if(g_list_length(view_p->selection_list) > 1) {
        DBG ("selection_count > 1\n");
        gtk_drag_set_icon_stock (view_p->drag_event.context, GTK_STOCK_DND_MULTIPLE, 0, 0);
    } else if(view_p->dnd_pixbuf) {
        DBG ("setting view_p->dnd_pixbuf\n");
        gtk_drag_set_icon_pixbuf (view_p->drag_event.context, view_p->dnd_pixbuf, 0, 0);
    } else {
        DBG ("view_p->dnd_pixbuf is null, using GTK_STOCK_DND\n");
        gtk_drag_set_icon_stock (view_p->drag_event.context, GTK_STOCK_DND, 0, 0);
    }
    view_p->dragstate = TRUE;
    TRACE ("enter dragstate done\n");

}

static void
setup_drag_state (view_t * view_p, GdkEventButton * event) {
    DBG ("+DND>>setup_drag_state\n");
    if(view_p->en && view_p->en->path)
        view_p->drag_src = view_p->en->path;
    else
        view_p->drag_src = NULL;
    DBG ("+DND>>setup_drag_state: view_p->drag_src=%s\n", view_p->drag_src);
    view_p->drag_event.type = GDK_DRAG_ENTER;
    view_p->drag_event.x_root = event->x;
    view_p->drag_event.y_root = event->y;
    view_p->drag_event.time = event->time + 2;
    view_p->drag_event.window = event->window;
    view_p->drag_event.send_event = event->send_event;
    TRACE ("event time=0x%x\n", event->time);
}

static void
graphics_saturation_event (view_t * view_p, population_t * population_p, double x, double y) {
    TRACE (">> graphics_saturation_event\n");
    if(!view_p)
        return;
    if(!population_p && view_p->saturated_p) {
        TRACE ("unsaturating\n");
        rodent_unsaturate_pixbuf (view_p);
        view_p->saturated_p=NULL;
        return;
    }
    TRACE ("graphics_saturation_event\n");
    if(view_p->doing_drag_p && x > 0 && y > 0) {
        gdouble h = (x - view_p->mouseX) * (x - view_p->mouseX) + (y - view_p->mouseY) * (y - view_p->mouseY);
        TRACE ("h=%f\n", h);
        /* reverse 1:2 to change behaviour from control to no-control */
        if(h >= 4)
            enter_drag_state (view_p);
#if 0
        /* if not valid drop target, return */
        if(view_p->module_name) {
            if(!rfm_natural ("rfm", view_p->module_name, population_p->en, "valid_drop_site"))
                return;
        } else {                /* local */
            if(!IS_SDIR (population_p->en->type))
                return;
        }
#endif
    }
    /*if (!population_p || population_p->selected) { */
    if(!population_p) {
        //graphics_unset_entry_tip(view_p);
        return;
    }
    if(population_p && !view_p->saturated_p) {
        rodent_saturate_pixbuf (view_p, population_p);
    }
    TRACE ("graphics_saturation_event done\n");
}

static int
select_population_rectangle (view_t * view_p, GdkRectangle * rect) {
    DBG ("RECT>> select_population_rectangle\n");
    int result = 0;
    int row,
      column;
    /*   int bottom_row=rect->y/CELLHEIGHT;
       int bottom_column=rect->x/CELLWIDTH;
       int top_row=(rect->y+rect->height)/CELLHEIGHT;
       int top_column=(rect->x+rect->width)/CELLWIDTH; */

    int low_X = rect->x,
        low_Y = rect->y;
    int high_X = rect->x + rect->width,
        high_Y = rect->y + rect->height;

    if(low_X < 0)
        low_X = 0;
    if(low_Y < 0)
        low_Y = 0;

    DBG ("RECT>> CELLHEIGHT=%d,CELLWIDTH=%d,rect=(%d,%d,%d,%d)\n", CELLHEIGHT, CELLWIDTH, low_X, low_Y, high_X, high_Y);
    if(!rect->width || !rect->height)
        return result;

    /* first find the affected grid elements. we
     * later can figure out if the icon is actually
     * selected. */
    if(!view_p->population_pp)
        return result;

    int item;
    for(item = 0; view_p->population_pp[item]; item++) {
        population_t *population_p = view_p->population_pp[item];
        int x_spacing = 0;
        int x = population_p->x,
            y = population_p->y;
        if(view_p->type == ICONVIEW_TYPE) {
            int columns = view_p->grid_columns;
            int rows = view_p->max_elements / columns;
            if(view_p->max_elements % columns)
                rows++;
            row = item / columns;
            column = item % columns;

            x_spacing = (CELLWIDTH - population_p->pixbufW) / 2;
            if(x_spacing < 0)
                x_spacing = 0;
            x = column * CELLWIDTH;
            y = row * CELLHEIGHT;

        }
        int low_x = x + x_spacing;
        int high_x = x + x_spacing + population_p->pixbufW;
        int low_y = y;
        int high_y = y + population_p->pixbufH;
        if(population_p->en) {
            DBG ("RECT: %s low=(%d,%d) high=(%d,%d)\n", population_p->en->path, low_x, low_y, high_x, high_y);
        }

        if(low_X - high_x < -1 && low_x - high_X < -1 && low_Y - high_y < -1 && low_y - high_Y < -1) {
            if(population_p->en) {
                DBG ("RECT: found in selection box: %s\n", population_p->en->path);
            }
            rodent_select_pixbuf (view_p, population_p);
            result++;
        }
    }
    return result;
}

static void
draw_window (view_t * view_p, int x, int y, gboolean draw) {
    DBG (">> draw_window\n");
    static GdkGC *toggleGC = NULL;
    if(view_p->boxX == -1 && view_p->boxY == -1)
        return;

    if(!toggleGC) {
        GdkColor pen_color = { 7, 65535, 30000, 65535 };
        toggleGC = gdk_gc_new (view_p->widgets.paper->window);
        gdk_gc_set_colormap (toggleGC, view_p->cmap);
        gdk_colormap_alloc_color (view_p->cmap, &pen_color, TRUE, TRUE);
        gdk_gc_set_foreground (toggleGC, &pen_color);
        gdk_gc_set_function (toggleGC, GDK_XOR);
    }
    if(view_p->old_X == -1 && view_p->old_Y == -1) {
        view_p->old_X = view_p->boxX;
        view_p->old_Y = view_p->boxY;
    } else {
        int lowX,
          highX,
          lowY,
          highY;
        lowX = (view_p->old_X > view_p->boxX) ? view_p->boxX : view_p->old_X;
        lowY = (view_p->old_Y > view_p->boxY) ? view_p->boxY : view_p->old_Y;
        highX = (view_p->old_X < view_p->boxX) ? view_p->boxX : view_p->old_X;
        highY = (view_p->old_Y < view_p->boxY) ? view_p->boxY : view_p->old_Y;
        /* clean old rectangle */
        if(lowX - highX || lowY - highY)
            gdk_draw_rectangle (view_p->widgets.paper->window, toggleGC, FALSE, lowX, lowY, highX - lowX, highY - lowY);
        /* remove leftover point */
        //gdk_draw_point (view_p->widgets.paper->window,toggleGC,view_p->boxX,view_p->boxY);
        if(!draw)
            return;
        /* draw new rectangle */
        view_p->old_X = x;
        view_p->old_Y = y;
        lowX = (view_p->old_X > view_p->boxX) ? view_p->boxX : view_p->old_X;
        lowY = (view_p->old_Y > view_p->boxY) ? view_p->boxY : view_p->old_Y;
        highX = (view_p->old_X < view_p->boxX) ? view_p->boxX : view_p->old_X;
        highY = (view_p->old_Y < view_p->boxY) ? view_p->boxY : view_p->old_Y;
        if(lowX - highX || lowY - highY)
            gdk_draw_rectangle (view_p->widgets.paper->window, toggleGC, FALSE, lowX, lowY, highX - lowX, highY - lowY);

    }

}

static int
select_population_area (view_t * view_p) {
    DBG ("RECT>> select_population_area downX=%d downY=%d oldX=%d oldY=%d\n",
         view_p->boxX, view_p->boxY, view_p->old_X, view_p->old_Y);
    GdkRectangle rect;
    int lowX = (view_p->old_X > view_p->boxX) ? view_p->boxX : view_p->old_X;
    int lowY = (view_p->old_Y > view_p->boxY) ? view_p->boxY : view_p->old_Y;
    int highX = (view_p->old_X < view_p->boxX) ? view_p->boxX : view_p->old_X;
    int highY = (view_p->old_Y < view_p->boxY) ? view_p->boxY : view_p->old_Y;
    rect.x = lowX;
    rect.y = lowY;
    rect.width = highX - lowX;
    rect.height = highY - lowY;
    return select_population_rectangle (view_p, &rect);
}

/* button press */
static void
button_popup (GdkEventButton * event, view_t * view_p, population_t * population_p) {
    if(population_p) {
        if(!population_p->selected) {
            if(!(event->state & GDK_CONTROL_MASK)) {
                rodent_unselect_all_pixbuf (view_p);
                rodent_unsaturate_pixbuf (view_p);
            }
        }
        rodent_saturate_pixbuf (view_p, population_p);
        rodent_select_pixbuf (view_p, population_p);
    }
    rodent_do_popup (view_p, population_p, event);
}

static void
rodent_unselect_pixbuf (view_t * view_p, population_t * population_p) {
    DBG (">> rodent_unselect_pixbuf\n");
    if(!population_p) {
        g_warning ("!population_p");
        return;
    }
    population_p->pixbuf = population_p->base_pixbuf;

    rodent_expose_item (view_p, population_p);

/*    gtk_widget_queue_draw_area (view_p->widgets.paper, 
		population_p->column*CELLWIDTH, 
		population_p->row*CELLHEIGHT,
		CELLWIDTH,CELLHEIGHT);	*/
//    if(population_p->selected) view_p->selection_count--;
    population_p->selected = FALSE;
    if(view_p->selection_list) {
        view_p->selection_list = g_list_remove (view_p->selection_list, population_p->en);
        if(!g_list_length (view_p->selection_list)) {
            g_list_free (view_p->selection_list);
            view_p->selection_list = NULL;
        }
    }
}
