/*
 * Copyright (C) 2002-2010 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <string.h>
#include <stdlib.h>
#ifdef HAVE_PTHREAD_TRYJOIN
# ifndef _GNU_SOURCE
#  define _GNU_SOURCE
# endif
# include <pthread.h>
#endif

#include "rodent_libs.h"
#include "rodent_population.i"

gint
rodent_get_population_serial (view_t * view_p) {
    if(view_p->population_serial_mutex == NULL) {
        view_p->population_serial_mutex = g_mutex_new ();
    }
    g_mutex_lock (view_p->population_serial_mutex);
    gint population_serial = view_p->population_serial;
    g_mutex_unlock (view_p->population_serial_mutex);
    return population_serial;
}

void
rodent_layout (view_t * view_p, population_t * population_p, const gchar * label) {
    TRACE (">> rodent_layout\n");
    if(!view_p || !population_p) {
        g_warning ("!view_p || !population_p");
        return;
    }

    gchar *actual_tag = rfm_utf_string (label);
    mk_layouts (view_p, actual_tag, &(population_p->layout), &(population_p->layout2));
    g_free (actual_tag);

    mk_logical_rect (population_p->layout, &(population_p->logical_rect));
    mk_logical_rect (population_p->layout2, &(population_p->logical_rect2));
    return;
}

population_t *
rodent_find_in_population (view_t * view_p, gdouble x, gdouble y) {
    if(!view_p->population_pp)
        return NULL;
    // desktop stuff
    if(view_p->type == DESKVIEW_TYPE) {
        population_t **tmp;
        for(tmp = view_p->population_pp; *tmp; tmp++) {
            population_t *population_p = *tmp;
            if(x >= population_p->x && x < population_p->x + population_p->pixbufW &&
               y >= population_p->y && y < population_p->y + population_p->pixbufH) {
                return population_p;
            }
            TRACE ("x=%d y=%d w=%d h=%d is not it\n", population_p->x, population_p->y, population_p->w, population_p->h);
        }
        return NULL;
    }
    // //////////////////////
    // else ICONVIEW_TYPE (gridview)
    int row = y / CELLHEIGHT;
    int column = x / CELLWIDTH;
    int element = row * view_p->grid_columns + column;
    population_t *population_p;
    if(x < 0 || y < 0)
        return NULL;
    if(element >= view_p->max_elements)
        return NULL;
    if(column >= view_p->grid_columns)
        return NULL;
    if(!view_p->population_pp)
        return NULL;

    population_p = *(view_p->population_pp + element);
    if(!population_p)
        return NULL;

    TRACE ("row, column ok (%d,%d)\n", row, column);

    /* update row, column to consider posible window resizes */
    population_p->row = row;
    population_p->column = column;

    column = (int)x % CELLWIDTH;
    int x_spacing;
    x_spacing = (CELLWIDTH - population_p->pixbufW) / 2;
    if(x_spacing < 0)
        x_spacing = 0;
    if(column < x_spacing || column > x_spacing + population_p->pixbufW)
        return NULL;
    row = (int)y % CELLHEIGHT;
    if(row == 0 || row > population_p->pixbufH)
        return NULL;
    return population_p;
}

population_t *
rodent_find_in_labels (view_t * view_p, gdouble x, gdouble y) {
    int row = y / CELLHEIGHT;
    int column = x / CELLWIDTH;
    int element = row * view_p->grid_columns + column;
    population_t *population_p;

    TRACE("rodent_find_in_labels %lf,%lf\n", x,y);
    if(!view_p->population_pp)
        return NULL;
    // desktop stuff
    if(view_p->type == DESKVIEW_TYPE) {
        population_t **tmp;
        for(tmp = view_p->population_pp; *tmp; tmp++) {
            population_t *population_p = *tmp;
            int height=0;
            int width=0;
            if(population_p->layout) {
                height += population_p->logical_rect.height;
                width = population_p->logical_rect.width;
            }
            if (population_p->layout2) {
                height += population_p->logical_rect2.height;
                if (population_p->logical_rect2.width > population_p->logical_rect.width)
                    width = population_p->logical_rect2.width;
            }
            //if (height) height += TEXTSPACING;

            if(x >= population_p->labelX && x < population_p->labelX + width 
                    &&
               y >= population_p->labelY && y < population_p->labelY + height)
            {
                TRACE ("FOUND *** x=%lf y=%lf... labelX=%d labelY=%d w=%d h=%d \n", x, y, population_p->labelX, population_p->labelY, population_p->w, population_p->h);
                return population_p;
            }
            TRACE ("*** x=%lf y=%lf... X=%d--%d Y=%d--%d\n", x, y, population_p->labelX, population_p->x + population_p->w, population_p->labelY, population_p->labelY + view_p->text_height+TEXTSPACING);
        }
        return NULL;
    }
    // //////////////////////
    // else ICONVIEW_TYPE (gridview)
    if(x < 0 || y < 0)
        return NULL;
    if(element >= view_p->max_elements){
        return NULL;
    }
    if(column >= view_p->grid_columns){
        return NULL;
    }
    if(!view_p->population_pp)
        return NULL;

    population_p = *(view_p->population_pp + element);
    if(!population_p)
        return NULL;

    TRACE ("***** row, column ok (%d,%d)\n", row, column);

    /* update row, column to consider posible window resizes */
    population_p->row = row;
    population_p->column = column;

    column = (int)x % CELLWIDTH;
    row = (int)y % CELLHEIGHT;
    if(population_p->layout) {
        if(column >= (CELLWIDTH - population_p->logical_rect.width) / 2
           && column <= (CELLWIDTH + population_p->logical_rect.width) / 2) {
            if(row >= TEXTSPACING + population_p->pixbufH &&
               row <= TEXTSPACING + population_p->pixbufH + population_p->logical_rect.height) {
                return population_p;
            }
        }
    }
    if(population_p->layout2) {
        if(column >= (CELLWIDTH - population_p->logical_rect2.width) / 2
           && column <= (CELLWIDTH + population_p->logical_rect2.width) / 2) {
            if(row >=
               TEXTSPACING + population_p->pixbufH +
               population_p->logical_rect2.height
               && row <= TEXTSPACING + population_p->pixbufH + 2 * population_p->logical_rect2.height) {
                return population_p;
            }
        }
    }
    return NULL;
}

void
rodent_update_cut_icons (view_t * view_p) {
    DBG ("CUT:update_cut_status\n");
    if(!rfm_population_try_read_lock (view_p))
        return;
    // find and update record
    population_t **tmp;
    for(tmp = view_p->population_pp; *tmp; tmp++) {
        population_t *population_p = *tmp;
        gboolean expose = FALSE;
        if(population_p->en && population_p->en->path) {
            int pasteboard_status = rfm_in_pasteboard (view_p, population_p->en);
            if(pasteboard_status) {
                DBG ("CUT: pasteboard: %s\n", population_p->en->path);
                GdkPixbuf *p = population_p->cut_pixbuf;
                population_p->cut_pixbuf = NULL;
                if(p)
                    g_object_unref (p);

                population_p->pixbuf = population_p->base_pixbuf =
                    population_p->cut_pixbuf = cut_icon (view_p, population_p, pasteboard_status);
                p = population_p->saturated_pixbuf;
                population_p->saturated_pixbuf = NULL;
                if(p)
                    g_object_unref (p);
                p = population_p->selected_pixbuf;
                population_p->selected_pixbuf = NULL;
                if(p)
                    g_object_unref (p);
                expose = TRUE;
            } else {
                DBG ("CUT: not in pasteboard: %s\n", population_p->en->path);
                population_p->pixbuf = population_p->base_pixbuf = population_p->normal_pixbuf;
                if(population_p->cut_pixbuf) {
                    GdkPixbuf *p = population_p->cut_pixbuf;
                    population_p->cut_pixbuf = NULL;
                    g_object_unref (p);
                    p = population_p->saturated_pixbuf;
                    population_p->saturated_pixbuf = NULL;
                    if(p)
                        g_object_unref (p);
                    p = population_p->selected_pixbuf;
                    population_p->selected_pixbuf = NULL;
                    if(p)
                        g_object_unref (p);
                    expose = TRUE;
                }
            }
            if(expose)
                rfm_expose_rect (view_p->widgets.paper->window, &(population_p->rect));
        }
    }

    // release population semaphore
    TRACE ("POP_SEM:  sem_post(&(view_p->population_sem))\n");
    rfm_population_read_unlock (view_p);
}

/**
 *
 * rodent_create_population_t: create a population element
 *
 * @view_p:
 * @en: the entry to assign to the population element.
 * @element: the element serial id number 
 * @icon_id: the mime_icon identifier. If NULL, then icon is obtained from entry information.
 * @label: the text which is to appear under the icon. If NULL, then the text is inferred from entry information.
 * 
 * **/

population_t *
rodent_create_population_t (view_t * view_p, record_entry_t * en, int element, const gchar * icon_id, const gchar * label) {
    GdkPixbuf *pixbuf = NULL;
    population_t *population_p;

    population_p = (population_t *) malloc (sizeof (population_t));
    memset (population_p, 0, sizeof (population_t));
    population_p->view_p = view_p;

    // from deskview definitions:
    // x,y are union with row,column
    /*if (x >= 0 && y >= 0) {
       population_p->x=x;
       population_p->y=y;
       } else */
    if(view_p->grid_columns) {
        population_p->row = element / view_p->grid_columns;
        population_p->column = element % view_p->grid_columns;
    }
    population_p->serial = view_p->serial++;
    ///////////////////////////////////////

    if(en) {
        population_p->en = rfm_copy_entry (en);
        population_p->en->module = view_p->submodule_name;
    } else
        population_p->en = NULL;

    if(icon_id) {
        TRACE ("icontell: %s\n", icon_id);
        pixbuf = rfm_get_pixbuf (icon_id, view_p->icon_size);
    } else if(!en) {
        pixbuf = rfm_get_pixbuf ("xffm/stock_system", view_p->icon_size);
    } else {
        // Default is now to always show little icon previews
        //if(en->path && (view_p->preferences & __SHOW_IMAGES)) {
        if(en->path) {
            // short circuit if image is already hashed
            pixbuf = rfm_get_from_pixbuf_hash (en->path, view_p->icon_size);
            if(pixbuf) {
                // What about composite types??
                // We should have retrieved pixbuf already composited:
                // rfm_preview_thread_f should worry about compositing and
                // saving appropriate pixbuf into the hash.
                population_p->preview_done = TRUE;
            }
        }
        // image is not hashed or not an image
        // this will get the default image
        if(pixbuf == NULL) {
            // image is not hashed or not an image
            // here we ge the default icon, to be updated to
            // a preview by the rfm_preview_thread_f if option is set (this is the default).
            DBG ("HASH-population.c: %s not found in pixbuf hash\n", en->path);
            // up type icon requires special compositing (requires view_p-> preferences)
            if(IS_UP_TYPE (en->type)) {
                const gchar *id = rfm_resolve_icon_id (en);
                const gchar *hidden_tag = NULL;
                const gchar *images_tag = NULL;
                if(view_p->preferences & __SHOW_HIDDEN)
                    hidden_tag = "/show-hidden";
                if(view_p->preferences & __SHOW_IMAGES)
                    images_tag = "/show-previews";
                gchar *composite_type = g_strdup_printf ("%s%s%s%s",
                                                         id,
                                                         (hidden_tag || images_tag) ? "/composite" : "",
                                                         (hidden_tag) ? hidden_tag : "",
                                                         (images_tag) ? images_tag : "");
                pixbuf = rfm_get_pixbuf (composite_type, view_p->icon_size);
                g_free (composite_type);
            } else {
                gchar *id = rfm_get_composite_icon_id (en, NULL);
                pixbuf = rfm_get_pixbuf (id, view_p->icon_size);
                g_free (id);
            }
        }
    }

    population_p->pixbuf = population_p->base_pixbuf = population_p->normal_pixbuf = pixbuf;
    int pasteboard_status = rfm_in_pasteboard (view_p, population_p->en);
    if(pasteboard_status) {
        population_p->pixbuf = population_p->base_pixbuf = population_p->cut_pixbuf =
            cut_icon (view_p, population_p, pasteboard_status);
    }

    if(pixbuf) {
        population_p->pixbufW = gdk_pixbuf_get_width ((const GdkPixbuf *)pixbuf);
        population_p->pixbufH = gdk_pixbuf_get_height ((const GdkPixbuf *)pixbuf);
    } else {
        population_p->pixbufW = view_p->icon_size;
        population_p->pixbufH = view_p->icon_size;
    }

    /*population_p->submodule_name=NULL; */
    population_p->saturated_pixbuf = NULL;
    population_p->selected_pixbuf = NULL;
    population_p->selected = FALSE;

    rodent_layout (view_p, population_p, label);

    population_p->rect.x = population_p->column * CELLWIDTH;
    population_p->rect.y = population_p->row * CELLHEIGHT;
    population_p->rect.width = CELLWIDTH;
    population_p->rect.height = CELLHEIGHT;

    return population_p;
}

void
rodent_destroy_population (view_t * view_p) {
    if(!view_p || !view_p->population_pp)
        return;

    population_t **tmp;
    DBG ("rodent_population: rodent_destroy_population\n");
    increase_population_serial (view_p);
    for(tmp = view_p->population_pp; *tmp; tmp++) {
        population_t *population_p = *tmp;
        destroy_population_item (population_p);
        population_p = NULL;
    }

    /*memset((void *)(view_p->population_pp),
       0,
       view_p->max_elements*sizeof(population_t *)); */

    g_free (view_p->population_pp);
    view_p->population_pp = NULL;
}

population_t **
rodent_mk_grid_elements (view_t * view_p, int grid_elements) {
    DBG (">> graphics_mk_grid_elements\n");
    population_t **population_pp;
    /* we use (grid_elements+1) to use a NULL pointer as an
     * end of list marker */
    TRACE ("grid_elements=%d (elements)\n", grid_elements);
    increase_population_serial (view_p);        // at destroy and here for transfer too.
    population_pp = (population_t **) malloc ((grid_elements + 1) * sizeof (population_t *));
    memset ((void *)(population_pp), 0, (grid_elements + 1) * sizeof (population_t *));
    population_pp[grid_elements] = NULL;
    return population_pp;
}

record_entry_t *
rodent_init_view (view_t * view_p, record_entry_t * en) {
    DBG ("rodent_population: rodent_init_view\n");

    // destroy obsolete entry:
    rfm_destroy_entry (view_p->en);
    rodent_destroy_population (view_p);
    view_p->boxX = -1, view_p->boxY = -1;
    view_p->old_X = -1, view_p->old_Y = -1;
    if(view_p->selection_list) {
        g_list_free (view_p->selection_list);
        view_p->selection_list = NULL;
    }
    view_p->mouseX = 0;
    view_p->mouseY = 0;
    view_p->tip_timer = 0;
    view_p->widgets.rename = NULL;
    view_p->label_p = NULL;
    view_p->doing_drag_p = NULL;
    view_p->saturated_p = NULL;
    view_p->selected_p = NULL;
    view_p->en = en;

    view_p->xfdir_p = NULL;

    return en;
}

void
rodent_expose_item (view_t * view_p, population_t * population_p) {
    TRACE (">> rodent_expose_item\n");
    GdkRectangle rect;
    if (!population_p) return;
    if(view_p->type == DESKVIEW_TYPE) {
        view_p->redraw_pixmap = TRUE;
        rect.x = population_p->x;
        rect.y = population_p->y;
    } else {
        rect.x = population_p->column * CELLWIDTH;
        rect.y = population_p->row * CELLHEIGHT;
    }
    rect.width = CELLWIDTH;
    rect.height = CELLHEIGHT;
 //    gtk_widget_queue_draw_area (view_p->widgets.paper,
  //                              population_p->column * CELLWIDTH, population_p->row * CELLHEIGHT, CELLWIDTH, CELLHEIGHT);
  //
    gdk_window_invalidate_rect (view_p->widgets.paper->window, &rect, TRUE);
//    gdk_window_invalidate_rect (view_p->widgets.window->window, &rect, TRUE);
    return;
}

void
rodent_expose_all (view_t * view_p) {
    TRACE (">> rodent_expose_all\n");
    if(view_p->type == DESKVIEW_TYPE)
        view_p->redraw_pixmap = TRUE;
    GdkRectangle rect;
    rect.x = rect.y = 0;
    rect.width = view_p->widgets.window->allocation.width;
    rect.height = view_p->widgets.window->allocation.height;
    gdk_window_invalidate_rect (view_p->widgets.window->window, &rect, TRUE);
    TRACE ("adj, exposeall for grid\n");
}

#define TIMERVAL 150

gint
rodent_monitor (gpointer data) {
    static gboolean rodent_monitor_active=FALSE;
    // make no mistake, only one monitor and in the main thread
    if (rodent_monitor_active) return TRUE;
    rodent_monitor_active=TRUE;
    GList *tmp;
    static GThread *valid_thread = NULL;
    static gint timer = 0;
    if(!timer) {
        timer = g_timeout_add (TIMERVAL, rodent_monitor, data);
        valid_thread = (GThread *) data;
	rodent_monitor_active=FALSE;
        return TRUE;
    }

    /* thread protection (just in case, though it does not seem necessary) */
    GThread *self = g_thread_self ();
    if(self != valid_thread) {
        g_warning
            ("Dropping invalid access attempt to timeout function 0x%lu != 0x%lu\n",
             (long unsigned)self, (long unsigned)valid_thread);
	rodent_monitor_active=FALSE;
        return TRUE;
    }

    TRACE ("rodent_monitor, calling thread 0x%lu\n", (unsigned long)self);

    for(tmp = global_view_list; tmp; tmp = tmp->next) {
        view_t *view_p = (view_t *) tmp->data;
        if(view_p->monitor_loop_mutex == NULL) {
            view_p->monitor_loop_mutex = g_mutex_new ();
        }
        if(!view_p->xfdir_p)
            continue;

        if(!view_p || !view_p->en) {
            DBG ("gridview_monitor:  !view_p || !view_p->en... no thread monitor.\n");
            continue;
        }
        /* XXX work in progress? */
        if(view_p->status != STATUS_IDLE)
            continue;

        if(view_p->refresh) {
            DBG ("gridview_monitor:  view_p->xfdir_p->refresh\n");
            DBG ("monitor refresh xfdir_p 0x%lx != xfdir_p->xfdir_p 0x%lx\n", (unsigned long)((long)view_p->xfdir_p),
                 (unsigned long)((long)view_p->xfdir_p->xfdir_p));
	    // ignore refresh conditions if any read_lock is set
	    if (view_p->population_lock_count == 0) {
		 if (rodent_reload_view ((gpointer) view_p)){
		    view_p->refresh = 0;
		 }
	    } else {
		TRACE("Cannot refresh, view_p->population_lock_count=%d\n",
			view_p->population_lock_count);
	    }
        }
    }
    rodent_monitor_active=FALSE;
    return TRUE;
}
// rodent_destroy_view: destroy a single view.
gboolean
rodent_destroy_view_item (view_t * view_p) {
    DBG (">> rodent_destroy_view_item\n");
   // First, hide all widgets.
    if (view_p->widgets.notebook) {
    	gtk_widget_hide(view_p->widgets.page_child);
    } else {
    	gtk_widget_hide(view_p->widgets.window);
    }

    global_view_list = g_list_remove (global_view_list, view_p);
    GList *window_view_list = g_object_get_data (G_OBJECT (view_p->widgets.window), "window_view_list");
    window_view_list = g_list_remove (window_view_list, view_p);
    DBG("removed %ld from windowviewlist\n",(long) view_p);
    g_object_set_data (G_OBJECT (view_p->widgets.window), "window_view_list", window_view_list);

    schedule_destroy(view_p);

   // no more pages, destroy window.
    if (window_view_list == NULL){
	gtk_widget_destroy (view_p->widgets.window);
	if (view_p->widgets.diagnostics_window) {
	    gtk_widget_destroy (view_p->widgets.diagnostics_window);
	}
    }
    else DBG("window_view_list != NULL--> %ld\n",(long)window_view_list->data);
 
    // if no more views left (in other windows, for example),
    // then exit the gtk loop.
    if(!global_view_list) {
        TRACE ("No more views on screen.\n");
	// remove al rodent-desk atoms as well
	if (view_p->type == DESKVIEW_TYPE) {
	    gchar *killall=g_find_program_in_path("killall");
	    if (killall){
		DBG("Doing a killall %s\n", rfm_global_p->argv[0]);
	        execlp(killall,  killall, rfm_global_p->argv[0], NULL);
	    }
	}
        gtk_main_quit ();
    }
    else DBG("global_view_list != NULL--> %ld\n",(long)global_view_list->data);

    return TRUE;
}

// rodent_destroy_view: destroys all views in the current window.
gboolean
rodent_destroy_view (view_t * view_p) {
    DBG (">> rodent_destroy_view\n");
   // First, hide all widgets.
    gtk_widget_hide(view_p->widgets.window);
    GList *window_view_list;

    while ((window_view_list = g_object_get_data (G_OBJECT (view_p->widgets.window), "window_view_list")) != NULL) {
    	DBG("destroying %ld \n",(long)window_view_list->data);
	rodent_destroy_view_item((view_t *)window_view_list->data);
    }
    return FALSE;
}

void
rodent_add_view (view_t * view_p) {
    if (!g_list_find (global_view_list, view_p)){
        global_view_list = g_list_append (global_view_list, view_p);
    }
    GList *window_view_list = g_object_get_data (G_OBJECT (view_p->widgets.window), "window_view_list");
    if (!g_list_find (window_view_list, view_p)){
	window_view_list = g_list_append (window_view_list, view_p);
	g_object_set_data (G_OBJECT (view_p->widgets.window), "window_view_list", window_view_list);
    }

}

 // generalized short reload routine...
 // when this is triggered, monitor is in or about to enter the wait state
 // this is a soft reload.
gboolean
rodent_reload_view (gpointer data) {
    view_t *view_p = (view_t *) data;
    if (view_p->protected){
	g_warning("invalid recursive call attempted\n");
	return 0;
    } else {
	view_p->protected=TRUE;
    }

    rfm_cursor_wait (view_p->widgets.window);
    //if (!view_p->xfdir_p || !view_p->xfdir_p->xfdir_p) {
    if(!view_p->xfdir_p) {
        g_error ("short_reload_view: !view_p->xfdir_p");
    }
    DBG ("starting short_reload_view...\n");
    /*  monitor in wait state  with view_p->refresh */

    DBG ("POP_SEM: rodent_reload_view lock\n");
    rfm_population_write_lock (view_p);
    DBG ("population_sem: rodent_reload_view() obtained...\n");
    // population semaphore is ours


    /* make new population_pp and transfer old population_pp elements,
     * creating new elements for those not in old population_pp
     * and eliminating those in old population_pp no longer present
     * */
    transfer_population (view_p);
    if(view_p->type == ICONVIEW_TYPE) {
	DBG ("init_grid at short_reload...\n");
        rodent_init_grid (view_p, view_p->xfdir_p->pathc);
        // this uses initgrid info but also may generate an expose
        // so must be placed after transfer_population()
        rodent_restore_view_size (view_p);
    }

    if(view_p->type == DESKVIEW_TYPE) {
        // recalculate x,y
        // get x,y coordinates for each population_p
        rodent_recalc_population_geometry (view_p);
    }
    if(view_p->type == DESKVIEW_TYPE) {
        // update pixmap
        DBG ("DESK: redraw_pixmap\n");
        view_p->redraw_pixmap = TRUE;
        //redraw_pixmap(view_p);
    }

    /* do an expose (without erasing!) */
    rodent_expose_all (view_p);

    if(view_p->en && view_p->en->st)
        memcpy (view_p->en->st, &view_p->xfdir_p->st, sizeof (struct stat));
    rfm_cursor_reset (view_p->widgets.window);
    gdk_flush ();

    // fire off pending preview threads, (thread managed):
    rfm_fireup_previews (view_p);

    DBG ("short_reload_view: reload is done\n");
    view_p->status = STATUS_IDLE;
    // this is not necesary here, reload condition
    // generated by monitor has left monitor in wait stat
    // view_p->xfdir_p->active = 0;
    DBG ("POP_SEM:  main thread sem_post\n");
    population_t *population_p = rodent_find_in_population (view_p, view_p->current_mouseX, view_p->current_mouseY);
    if (population_p) {
        rodent_saturate_pixbuf (view_p, population_p);	
    }

    rfm_population_write_unlock (view_p);
    DBG ("population_sem: rodent_reload_view(): released!\n");
    view_p->protected=FALSE;
    return TRUE;

}

//*** end progress bar stuff ***/

void
rodent_set_icon_name (GtkWidget * widget, const gchar * path) {
    gchar *iconname;
    DBG ("rodent_set_icon_name\n");
    if(!path) {
        iconname = rfm_utf_string (g_get_host_name());
    } else if(g_file_test (path, G_FILE_TEST_EXISTS)) {
	//gchar *display_host = rfm_display_host_name(widget); 


        gchar *basename = g_path_get_basename (path);
        gchar *pathname = g_strdup (path);
        gchar *b = rfm_utf_string (basename);   // non chopped
        rfm_chop_excess (pathname);
        gchar *q = rfm_utf_string (pathname);   // non chopped

        g_free (basename);
        g_free (pathname);
	//iconname = g_strconcat (display_host, ":  ", b, " (", q, ")", NULL);
	iconname = g_strconcat (b, " (", q, ")", NULL);
        g_free (q);
        g_free (b);
    } else {
        iconname = rfm_utf_string (path);
    }
    gtk_window_set_title (GTK_WINDOW (gtk_widget_get_toplevel (widget)), iconname);
    gdk_flush ();
    g_free (iconname);
}

void
rodent_add_icon_roots (view_t * view_p) {
    int i = 0;
    population_t *population_p;
    GSList *list,
     *plugin_list;

    TRACE ("call to rodent_add_icon_roots\n");
    plugin_list = rfm_find_root_plugins ();

    DBG ("ROOT: rodent_add_icon_roots: gdk_window_resize(%d,%d) rows=%d, cols=%d\n",
         view_p->paperX, view_p->paperY, 
	 view_p->grid_rows, view_p->grid_columns);
    if (view_p->type != DESKVIEW_TYPE) { 
      gtk_widget_set_size_request (view_p->widgets.paper, view_p->paperX, view_p->paperY);
    }      
    view_p->population_pp = rodent_mk_grid_elements (view_p, g_slist_length(plugin_list) + 2);

     /*init_icon_view readjusts rows if necessary */
    if(view_p->type != DESKVIEW_TYPE) {
	rodent_init_grid (view_p, g_slist_length (plugin_list) + 2);
    }

    /* create filesystem root icon */
    population_p = view_p->population_pp[i] = create_local_root (view_p, i, NULL);
    i++;
    /* create homedir icon */
    population_p = view_p->population_pp[i] = create_local_root (view_p, i, g_get_home_dir ());
    i++;

    DBG ("ROOTS: now plugins...\n");
    for(list = plugin_list; list; list = list->next) {
        DBG ("ROOTS: creating population item for module %s\n", (char *)list->data);
        population_p = view_p->population_pp[i] = create_icon_root_element (view_p, i, list->data);
	if (!population_p) continue;
        i++;
    }
    if(view_p->type == DESKVIEW_TYPE){
        rodent_recalc_population_geometry (view_p);
    }

    rodent_expose_all(view_p);
    return;
}

//
//
//     end icon roots stuff

// hard reload: implies user click triggered reload.
// this is main thread business, so we must protect
// from gtk-loop recursiveness.

int
rodent_full_reload_view (gpointer data, record_entry_t * target_en) {
    view_t *view_p = (view_t *) data;
    widgets_t *widgets_p = &(view_p->widgets);
    
    TRACE ("rodent_full_reload_view... %s\n", 
	    (target_en && target_en->path)?target_en->path:"NULL");
    if (view_p->protected){
	g_warning("invalid recursive call attempted\n");
	return 0;
    } else {
	DBG("protecting\n");
	view_p->protected=TRUE;
    }
    if (view_p->thumbnailer_active) {
	view_p->thumbnailer_exit=TRUE;    
    }
    if(view_p->type == DESKVIEW_TYPE) {
	g_free (view_p->desktop_dir);
        if (!target_en) {
	    view_p->desktop_dir = g_strdup("");
	} else if (target_en->module) {
	    view_p->desktop_dir = g_strdup (target_en->module);
	} else {
	    view_p->desktop_dir = g_strdup (target_en->path);
	}
	rfm_setenv ( "RFM_DESKTOP_DIR", view_p->desktop_dir, TRUE);
	// save configuration too
	rfm_rational ("rfm", "settings", (void *) "RFM_DESKTOP_DIR", (void *) view_p->desktop_dir, "mcs_set_var");
	DBG(" setting RFM_DESKTOP_DIR to %s\n",view_p->desktop_dir);
    }

    DBG ("starting rodent_full_reload_view... protected=%d\n", view_p->protected);

    // if we claim write lock here, other gtk events which
    // require read lock will have to wait. That's cool.
    // This write lock request blocks until write lock granted
    // (all threads with readlocks must not deadlock before releasing lock!)
    rfm_population_write_lock (view_p);

    // past this point, the reload is nonstop.
    /* schedule termination of any running monitor threads for the view_p */
    rfm_xfdir_exit_monitor (view_p);
    rfm_cursor_wait (view_p->widgets.window);

    // first, clear the screen so that the user sees immediate response, but
    // only if the target is different, otherwise it is just an annoying flash
    // this is only applicable to gridview, since the deskview pixmap update
    // is done offscreen.
    if(view_p->type != DESKVIEW_TYPE) {
        if(target_en && view_p->en && view_p->en->path && target_en->path && strcmp (view_p->en->path, target_en->path) != 0) {
            rodent_clear_paper (view_p);
            rodent_expose_all (view_p);
        }
    }
    // init_view will destroy population, so we must have 
    // obtained the write lock by here.   
    rodent_init_view (view_p, target_en);
    reset_view_preferences (view_p);
    if(view_p->type != DESKVIEW_TYPE) {
        set_view_title (view_p);
    }
    if(!view_p->en) {
        /* root icons special case scenario */
        if(view_p->selection_list) {
            g_list_free (view_p->selection_list);
            view_p->selection_list = NULL;
        }
        rodent_clear_paper (view_p);
        rfm_cursor_reset (view_p->widgets.window);
        //graphics_set_default_tooltip(view_p,NULL);
        view_p->status = STATUS_IDLE;
        view_p->icon_size = SIZE_ICON;
        rodent_add_icon_roots (view_p);
        rfm_population_write_unlock (view_p);
	view_p->protected=FALSE;
	population_t *population_p = rodent_find_in_population (view_p, view_p->current_mouseX, view_p->current_mouseY);
	if (population_p) {
	    rodent_saturate_pixbuf (view_p, population_p);	
	}
	rfm_status (widgets_p, "xffm/stock_system", _("Installed Plugins"), NULL);

       return TRUE;
    }
    view_p->module_name = target_en->module;

    // Thread 1: load the xfdir structure.
    ///////////////////////////////////////////////////
    /* We now load the xfdir structure (there is none for roots). */
    // doing load_xfdir_structure() by thread blocks navigational deskview
    // when called by gtk timeout at deskview_monitor.i. (recursive thread bumps
    // into a GDK_THREADS mutex). Anyways, there is no need to do by thread since there
    // is no status area in deskview.
    if(view_p->type == DESKVIEW_TYPE) {
	//thread_tryjoin (load_xfdir_structure, view_p);
	load_xfdir_structure(view_p);
	//thread_tryjoin (load_xfdir_icons, view_p);
        load_xfdir_icons(view_p);
    } else {
	rfm_status (widgets_p, "xffm/info", _("Reading directory: "), target_en->path, NULL);
	// you cannot have gtk_main_iterations while waiting for thread??
	// (timeout functions should not be processed)
	//load_xfdir_structure(view_p);
	// Thread 1: load the xfdir structure.
	thread_tryjoin (load_xfdir_structure, view_p);
        rodent_init_grid (view_p, view_p->xfdir_p->pathc);
        rodent_restore_view_size (view_p);
	// Thread 2: create the population (entries and icons)
	// no point in doing this by thread if there is no status area... 
        rfm_status (widgets_p, "xffm/info", _("Loading folder..."), NULL);
        view_p->progress_count = 0;
        view_p->progress_total = view_p->xfdir_p->pathc;
        if(view_p->xfdir_p->pathc >= 512) {     /* eyecandy for long loads */
            set_progress_generic (&(view_p->widgets), 0, view_p->progress_total, 1);
        }
        //load_xfdir_icons(view_p);
        thread_tryjoin (load_xfdir_icons, view_p);

        rfm_update_status_line (view_p);
        set_progress_generic (&(view_p->widgets), -1, -1, -1);
        rodent_set_scroll_position (view_p);
    } 
    g_free (view_p->workdir);
    view_p->workdir = g_strdup (target_en->path);

    if(view_p->en && view_p->en->st)
        memcpy (view_p->en->st, &view_p->xfdir_p->st, sizeof (struct stat));
    gdk_flush ();
 
    // fire off pending preview threads, (thread managed):
    DBG ("reload_iconview: rfm_fireup_previews\n");
    rfm_fireup_previews (view_p);       // resets light to green
    view_p->status = STATUS_IDLE;

    population_t *population_p = rodent_find_in_population (view_p, view_p->current_mouseX, view_p->current_mouseY);
    if (population_p) {
        rodent_saturate_pixbuf (view_p, population_p);	
    }

    rfm_cursor_reset (view_p->widgets.window);
    rfm_population_write_unlock (view_p);
    DBG("unprotecting...\n");
    rodent_expose_all (view_p);
    view_p->protected=FALSE;
    return TRUE;
}

