/*
 * Copyright (C) 2002-2010 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define DEEPEST_DIR_MENU_LEVELS 13
#define DEEPEST_MODULE_MENU_LEVELS 13
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <pwd.h>
#include <sys/mman.h>

#include <ctype.h>
#include <glob.h>
#include <memory.h>
#include <signal.h>
#include <dirent.h>

#include <X11/Xlib.h>
#ifdef HAVE_LIBSM
# include <X11/SM/SMlib.h>
#endif

#include <gdk/gdkx.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <gmodule.h>

#include "rodent_libs.h"
#include "rodent_popup.i"

/*****************************************************************************/

GtkWidget *
rodent_get_widget_by_name (GtkWidget * parent, const gchar * name) {
    return rfm_get_widget_by_name (parent, name);

}

void
rodent_set_widget_by_name (GtkWidget * parent, const gchar * name, GtkWidget * widget) {
    return rfm_set_widget_by_name (parent, name, widget);
}

void
rodent_push_view_go_history (view_t * view_p) {
    record_entry_t *history_en;
    /*if (!view_p->en || !view_p->en->path) return; */
    if(view_p->go_list) {
        GList *last = g_list_last (view_p->go_list);
        record_entry_t *last_en;
        last_en = last->data;
        if(!last_en && !view_p->en)
            return;             /* NULL case */
        if(last_en && view_p->en) {
            if(!last_en->path && !view_p->en->path)
                return;
            if(last_en->path && view_p->en->path) {
                if(strcmp (last_en->path, view_p->en->path) == 0)
                    return;
            }
        }
    }
    history_en = rfm_copy_entry (view_p->en);
    view_p->go_list = g_list_append (view_p->go_list, history_en);
    return;
}

/******************** common menu configuration ******************************/
int
rodent_refresh (widgets_t * widgets_p, record_entry_t * en) {
    DBG ("entering rodent_refresh (%s)\n", (en) ? en->path : "en==NULL");
    g_return_val_if_fail (widgets_p != NULL, 0);
    view_t *view_p = widgets_p->view_p;
    view_p->redraw_pixmap = TRUE;
    int retval;
    if(en) {
        UNSET_UP_TYPE (en->type);
    }
    TRACE("source inode=%d target inode=%d\n",
	    (int)view_p->en->st->st_ino,
	    (int)en->st->st_ino);
    gboolean redundant_condition=en && view_p->en && en->st && view_p->en->st && view_p->en->st->st_ino == en->st->st_ino && view_p->en->st->st_dev == en->st->st_dev && strcmp(view_p->en->path, en->path)!=0;
    if (redundant_condition){
	rfm_diagnostics(widgets_p, "xffm/warning", NULL);
	rfm_diagnostics(widgets_p, "xffm_tag/stderr", strerror(EXDEV),"\n",NULL);
	return 0;
    }
    retval = rodent_full_reload_view ((gpointer) view_p, en);

    return retval;
}

/**************************************************************/

gpointer
rodent_create_popup_bythread(gpointer data){
    view_t * view_p=data;
GDK_THREADS_ENTER();
    view_p->widgets.popup = mk_popup_menu (NULL, NULL, view_p);
    gui_add_autotype_C_widgets (&(view_p->widgets), 
                                    view_p->widgets.popup, 
				    (gpointer) execute_autotype,
                                    (gpointer) rodent_mount, 
				    (gpointer) rodent_unmount);
GDK_THREADS_LEAVE();
	return NULL;
}

void
rodent_do_popup (view_t * view_p, population_t * population_p, GdkEventButton * event) {
    if(!view_p)
        return;
    widgets_t *widgets_p=&(view_p->widgets);
    /* only need to destroy and recreate if module fiddles around with it */
    if(!view_p->widgets.popup) {
        view_p->widgets.popup = mk_popup_menu (NULL, NULL, view_p);
        gui_add_autotype_C_widgets (&(view_p->widgets), 
                                    view_p->widgets.popup, (gpointer) execute_autotype,
                                    (gpointer) rodent_mount, (gpointer) rodent_unmount);

    }
    gtk_widget_hide_all (view_p->widgets.popup);
    DBG ("DO_POPUPx\n");

    if(view_p->module_name && !population_p
       && rfm_rational ("rfm", view_p->module_name, (void *)(view_p->en), widgets_p, "private_popup")) {
        DBG ("private module POPUPx menu here (no population.\n");
        return;
    }
    if(population_p && population_p->en && population_p->en->module &&
       rfm_rational ("rfm", population_p->en->module, (void *)(population_p->en), widgets_p, "private_popup")) {
        DBG ("private module POPUPx menu here.\n");
        return;
    }

    {
        ///hmmm works when menu is created, but not here.....
        GtkWidget *w,
         *inner_label = rodent_get_widget_by_name (widgets_p->paper, "sort1");
        GdkPixbuf *pb;
        if(view_p->preferences & SORT_ASCENDING)
            pb = rfm_get_pixbuf ("xffm/stock_sort-ascending", SIZE_BUTTON);
        else
            pb = rfm_get_pixbuf ("xffm/stock_sort-descending", SIZE_BUTTON);
        w = gtk_image_new_from_pixbuf (pb);
        gtk_widget_show (w);
        TRACE ("now setting image for sort order menu item...\n");
        gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (inner_label), w);
    }

    if(!population_p || !population_p->en || !population_p->en->path) {
        DBG ("refresh_item here...\n");
        refresh_item (view_p, population_p);
        set_menu_toggles (view_p);
        if(view_p->module_name) {
            rfm_rational ("rfm", view_p->module_name, NULL, widgets_p, "hide_local_menu_items");
            rfm_rational ("rfm", view_p->module_name, NULL, widgets_p, "show_local_menu_items");
        }
        gtk_menu_popup (GTK_MENU (view_p->widgets.popup), NULL, NULL, NULL, NULL, 3, event->time);
        DBG ("gtk_menu_popup now...\n");
        return;
    }
    if(IS_DUMMY_TYPE (population_p->en->type) || IS_ROOT_TYPE (population_p->en->type)) {
        refresh_item (view_p, population_p);
        if (g_file_test(population_p->en->path, G_FILE_TEST_EXISTS)) {
            gui_autostuff (widgets_p, widgets_p->popup,
                   population_p->en, (gpointer) execute_autotype, 
                   (gpointer) rodent_mount, (gpointer) rodent_unmount);
        }
        gtk_menu_popup (GTK_MENU (widgets_p->navigation_popup), NULL, NULL, NULL, NULL, 3, event->time);
        return;
    }

    DBG ("POPUPx gui_autostuff here...\n");

    // re-arrange autostuff here:
    gui_autostuff (widgets_p, widgets_p->popup,
                   population_p->en, (gpointer) execute_autotype, 
                   (gpointer) rodent_mount, (gpointer) rodent_unmount);

    DBG ("POPUPx  mount_item here...\n");
    mount_item (view_p, population_p);
    DBG ("POPUPx  remove_item here...\n");
    remove_item (view_p, population_p);
    DBG ("POPUPx open_item  here...\n");
    open_item (view_p, population_p);
    DBG ("POPUPx properties_item  here...\n");
    properties_item (view_p, population_p);
    DBG ("set_menu_toggles here...\n");
    set_menu_toggles (view_p);
    if(view_p->module_name) {
        rfm_rational ("rfm", view_p->module_name, population_p->en, widgets_p, "hide_local_menu_items");
        rfm_rational ("rfm", view_p->module_name, population_p->en, widgets_p, "show_local_menu_items");
    }

    TRACE ("view_p->module_name=%s,view_p->submodule_name=%s\n", view_p->module_name, view_p->submodule_name);
    DBG ("RODENT_MK_MODULE_POPUP_MENU here...\n");
    rodent_mk_module_popup_menu (view_p->module_name, view_p, view_p->en);
    DBG ("gtk_menu_popup now...\n");

    gtk_menu_popup (GTK_MENU (widgets_p->popup), NULL, NULL, NULL, NULL, 3, event->time);
}

void
rodent_open_with (widgets_t * widgets_p, record_entry_t * en) {
    /* open with */
    view_t *view_p=widgets_p->view_p;
    gchar *command;
    gchar *command_fmt;
    TRACE ("OPEN: rodent_open_with \n");

    if(!en || !en->path) {
        DBG ("OPEN: rodent_open_with !en || !en->path\n");
        return;
    }

    g_free (view_p->workdir);
    view_p->workdir = g_path_get_dirname ((const gchar *)en->path);

    // tests 
    // for default editor...
    if (!en->mimetype) en->mimetype=MIME_type (en->path, en->st);
    if (!en->mimemagic) en->mimemagic= MIME_magic(en->path);
    if (!en->filetype) en->filetype= MIME_file(en->path);

    gboolean is_text = ((en->mimetype && strstr (en->mimetype, "text/")) ||
                        (en->mimemagic && strstr (en->mimemagic, "text/")) || 
                        (en->filetype && strstr (en->filetype, "text")));

    command_fmt = MIME_command (en->mimetype);
    DBG ("OPEN: command_fmt(%s) = %s\n", en->mimetype, command_fmt);
    if(!command_fmt) {
        command_fmt = MIME_command (en->mimemagic);
    }
    DBG ("OPEN: command_fmt(%s) = %s\n", en->mimemagic, command_fmt);

    if(!command_fmt && is_text && getenv ("EDITOR_CMD") && strlen (getenv ("EDITOR_CMD"))) {
        command_fmt = g_strdup (getenv ("EDITOR_CMD"));
    }

    DBG ("OPEN: command_fmt = %s\n", command_fmt);
    if(command_fmt) {
        command = MIME_mk_command_line (command_fmt, en->path);
        DBG ("OPEN: command = %s\n", command);
        rfm_thread_run2argv (widgets_p, command, FALSE);
        g_free (command);
        g_free (command_fmt);
    } else {
        on_open_with_activate (NULL, (gpointer) widgets_p);
    }
    return;
}

void
rodent_remove_activate (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    rodent_remove (widgets_p, view_p->selection_list);
}

int
rodent_touch (widgets_t * widgets_p, const gchar * path) {
    if(!widgets_p || !path)
        return FALSE;
    gchar *command = g_strdup_printf ("touch \"%s\"", (gchar *) path);

    TRACE ("touch_it\n");
    rfm_diagnostics (widgets_p, NULL, "$touch \"", path, "\"\n", NULL);
    rfm_thread_run2argv (widgets_p, command, FALSE);
    g_free (command);
    return TRUE;
}

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

/****************/

extern char **environ;

/***************/

gboolean
rodent_common_mask_key (GdkEventKey * event, void *data) {
    if(!(event->state & (GDK_SHIFT_MASK | GDK_CONTROL_MASK | GDK_MOD1_MASK))) {
        TRACE ("no mask\n");
        return FALSE;
    }
    if(event->state & GDK_SHIFT_MASK)
        return gui_keybind_exec (event->keyval, GDK_SHIFT_MASK, data);
    else if(event->state & GDK_CONTROL_MASK)
        return gui_keybind_exec (event->keyval, GDK_CONTROL_MASK, data);
    else if(event->state & GDK_MOD1_MASK)
        return gui_keybind_exec (event->keyval, GDK_MOD1_MASK, data);
    return FALSE;
}

GtkWidget *
rodent_mk_menu (widgets_t * widgets_p,
                const gchar * label, const gchar * name, GtkWidget * parent, gpointer callback, const gchar * iconfile) {
    GtkWidget *menu_menu = gtk_menu_new ();
    GtkWidget *v;
    GtkWidget *w = gtk_image_menu_item_new_with_mnemonic (label);
    gtk_image_menu_item_set_always_show_image ((GtkImageMenuItem *) w, TRUE);
    if(name)
        RODENT_HOOKUP_OBJECT (widgets_p->paper, w, name);
    gtk_widget_show (w);
    if(parent)
        gtk_container_add (GTK_CONTAINER (parent), w);
    if(name) {
        gchar *g = g_strconcat (name, "_menu", NULL);
        RODENT_HOOKUP_OBJECT (widgets_p->paper, menu_menu, g);
    }
    gtk_menu_item_set_submenu (GTK_MENU_ITEM (w), menu_menu);
    v = gtk_menu_item_new_with_mnemonic (label);
    gtk_widget_show (v);
    gtk_container_add (GTK_CONTAINER (menu_menu), v);
    gtk_widget_set_sensitive (v, FALSE);
    gui_add_menu_separator (menu_menu);
    if(callback) {
        g_signal_connect ((gpointer) w, "activate", G_CALLBACK (callback), NULL);
    }
    /* invalid cast from `GtkMenuItem' to `GtkImageMenuItem'
     * (just one item, so maybe not worth using... )
     * if (iconfile) rodent_mk_pixmap_menu(window, iconfile,w,MENU_PIXMAP);*/

    return menu_menu;
}

GtkWidget *
rodent_mk_pixmap_menu (widgets_t * widgets_p, const gchar * icon_id, GtkWidget * parent, int caso) {

    GtkWidget *w;
    GdkPixbuf *pb = NULL;

    TRACE (" gui_mk_pixmap_menu, icon_id: %s\n", icon_id);
    switch (caso) {
    case MENU_PIXMAP:
        pb = rfm_get_pixbuf (icon_id, SIZE_BUTTON);
        break;
    case BUTTON_PIXMAP:
        pb = rfm_get_pixbuf (icon_id, SIZE_BUTTON);
        break;
    case BIGGER_MENU_PIXMAP:
    case BIGGER_BUTTON_PIXMAP:
        pb = rfm_get_pixbuf (icon_id, SIZE_DIALOG);
        break;
    case OTHER_PIXMAP:
        pb = rfm_get_pixbuf (icon_id, SIZE_BUTTON);
        break;
    }
    if(!pb) {
        g_warning ("pb=NULL in gui_mk_pixmap_menu()");
        return NULL;
    }
    TRACE (" gui_mk_pixmap_menu ok, icon_id: %s\n", icon_id);

    w = gtk_image_new_from_pixbuf (pb);
    gtk_widget_show (w);

    switch (caso) {
    case BIGGER_MENU_PIXMAP:
    case MENU_PIXMAP:
        TRACE ("gtk_image_menu_item_set_image now...\n");
        gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (parent), w);
        break;
    case BIGGER_BUTTON_PIXMAP:
    case BUTTON_PIXMAP:
        gtk_container_add (GTK_CONTAINER (parent), w);
        break;
    case OTHER_PIXMAP:
        gtk_box_pack_start (GTK_BOX (parent), w, FALSE, TRUE, 0);
        gtk_misc_set_padding (GTK_MISC (w), 4, 0);
        break;
    }
    return w;
}

const gchar *
rodent_get_response (widgets_t * widgets_p, const gchar * title_txt, const gchar * label_txt, const gchar * default_txt) {
    static gchar *response_txt = NULL;
    int response = GTK_RESPONSE_NONE;
    GtkWidget *hbox,
     *label,
     *entry,
     *button,
     *dialog;
    if(!default_txt)
        default_txt = "";

    dialog = gtk_dialog_new ();

    g_free (response_txt);
    response_txt = NULL;
    if(widgets_p) {
	view_t *view_p=widgets_p->view_p;
        if(view_p->type != DESKVIEW_TYPE) {
            gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
            gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (widgets_p->window));
        }
    }
    gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
    gtk_window_set_resizable (GTK_WINDOW (dialog), TRUE);
    gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
    gtk_container_set_border_width (GTK_CONTAINER (dialog), 6);

    if(label_txt)
        label = gtk_label_new (label_txt);
    else
        label = gtk_label_new (_("Preparing"));

    hbox = gtk_hbox_new (TRUE, 6);
    gtk_box_pack_start (GTK_BOX ((GTK_DIALOG (dialog))->vbox), hbox, FALSE, FALSE, 0);

    entry = gtk_entry_new ();

    gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
    gtk_box_pack_start (GTK_BOX (hbox), entry, TRUE, TRUE, 0);
    gtk_widget_show_all (hbox);

/*if (GTK_CHECK_VERSION (2,18,0)) {
    GtkEntryBuffer *EntryBuffer= (GtkEntryBuffer *)gtk_entry_get_buffer((GtkEntry *)entry);
    gtk_entry_buffer_set_text (EntryBuffer,default_txt,strlen(default_txt));
} else {*/
/* gtk < 2.18 */
    gtk_entry_set_text ((GtkEntry *) entry, default_txt);
//}

    g_signal_connect (G_OBJECT (entry), "activate", G_CALLBACK (activate_entry), dialog);

    button = gtk_button_new_from_stock (GTK_STOCK_CANCEL);
    gtk_widget_show (button);
    gtk_dialog_add_action_widget (GTK_DIALOG (dialog), button, GTK_RESPONSE_CANCEL);
    button = gtk_button_new_from_stock (GTK_STOCK_OK);
    gtk_widget_show (button);

    gtk_dialog_add_action_widget (GTK_DIALOG (dialog), button, GTK_RESPONSE_YES);
    gtk_widget_realize (dialog);

    if(title_txt)
        gtk_window_set_title (GTK_WINDOW (dialog), title_txt);
    else
        gdk_window_set_decorations (dialog->window, GDK_DECOR_BORDER);

    /* show dialog and return */
    gtk_widget_show_all (dialog);
    response = gtk_dialog_run (GTK_DIALOG (dialog));
    if(response == GTK_RESPONSE_YES) {
        const gchar *et = gtk_entry_get_text (GTK_ENTRY (entry));
        if(et && strlen (et)) {
            response_txt = g_strdup (et);
        }
    }
    gtk_widget_hide (dialog);
    gtk_widget_destroy (dialog);
    g_strstrip (response_txt);

    /* use slocate_string to select... */
    return (const gchar *)(response_txt);
}

#if 0
int
rodent_get_response_integer (widgets_t * widgets_p, const gchar * title_txt, const gchar * label_txt) {
    int response_int = 1;
    int response = GTK_RESPONSE_NONE;
    GtkWidget *hbox,
     *label,
     *button,
     *dialog;

    dialog = gtk_dialog_new ();

    if(widgets_p) {
        if(widgets_p->view_p->type != DESKVIEW_TYPE) {
            gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
            gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (widgets_p->window));
        }
    }
    gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
    gtk_window_set_resizable (GTK_WINDOW (dialog), TRUE);
    gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
    gtk_container_set_border_width (GTK_CONTAINER (dialog), 6);

    if(label_txt)
        label = gtk_label_new (label_txt);
    else
        label = gtk_label_new (_("Preparing"));

    hbox = gtk_hbox_new (TRUE, 6);
    gtk_box_pack_start (GTK_BOX ((GTK_DIALOG (dialog))->vbox), hbox, FALSE, FALSE, 0);

    GtkObject *adjustment = gtk_adjustment_new (1.0, 1.0, 20.0, 1.0, 1.0, 0);

    GtkWidget *spin_button = gtk_spin_button_new (GTK_ADJUSTMENT (adjustment), 1.0, 0);

    gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
    gtk_box_pack_start (GTK_BOX (hbox), spin_button, TRUE, TRUE, 0);
    gtk_widget_show_all (hbox);

    button = gtk_button_new_from_stock (GTK_STOCK_OK);
    gtk_widget_show (button);

    gtk_dialog_add_action_widget (GTK_DIALOG (dialog), button, GTK_RESPONSE_YES);
    gtk_widget_realize (dialog);

    if(title_txt)
        gtk_window_set_title (GTK_WINDOW (dialog), title_txt);
    else
        gdk_window_set_decorations (dialog->window, GDK_DECOR_BORDER);

    /* show dialog and return */
    gtk_widget_show_all (dialog);
    response = gtk_dialog_run (GTK_DIALOG (dialog));
    if(response == GTK_RESPONSE_YES) {
        response_int = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (spin_button));
    } else
        response_int = 1;
    gtk_widget_hide (dialog);
    gtk_widget_destroy (dialog);

    return response_int;
}
#endif
const gchar *
rodent_get_response_history (widgets_t * widgets_p,
                             const gchar * title_txt,
                             const gchar * label_txt,
                             gchar * history_file,
                             const gchar * path,
                             const gchar * entry_text,
                             gchar * flagfile, 
			     const gchar * check_label, 
			     int filechooser_action, 
			     const gchar * folder) {
    static extra_key_t extra_key;
    static gchar *response_txt = NULL;
    int response = GTK_RESPONSE_NONE;
    GtkWidget *hbox,
     *label,
     *button,
     *dialog;
    static GtkComboBoxEntry *combo = NULL;
    void *combo_info = NULL;
    filechooser_t filechooser_v;

    dialog = gtk_dialog_new ();

    g_free (response_txt);
    response_txt = NULL;
    if(widgets_p) {
        gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (widgets_p->window));
    }
    gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
    gtk_window_set_resizable (GTK_WINDOW (dialog), TRUE);
    gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
    gtk_container_set_border_width (GTK_CONTAINER (dialog), 6);

    if(label_txt)
        label = gtk_label_new (label_txt);
    else
        label = gtk_label_new (_("Preparing"));

    combo = (GtkComboBoxEntry *) gtk_combo_box_entry_new ();
    gtk_widget_set_size_request (GTK_WIDGET (combo), 350, -1);

    hbox = gtk_hbox_new (FALSE, 6);
    gtk_box_pack_start (GTK_BOX ((GTK_DIALOG (dialog))->vbox), hbox, FALSE, FALSE, 0);
    GtkWidget *vbox = gtk_vbox_new (FALSE, 6);
    gtk_box_pack_start (GTK_BOX (hbox), vbox, FALSE, FALSE, 0);
    gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
    vbox = gtk_vbox_new (FALSE, 6);
    gtk_box_pack_start (GTK_BOX (hbox), vbox, FALSE, FALSE, 0);
    gtk_box_pack_start (GTK_BOX (vbox), (GtkWidget *) combo, FALSE, FALSE, 0);


    if(combo_info == NULL) {
        TRACE ("signal widgets_p is 0x%lx\n", (unsigned long)widgets_p);
        combo_info = COMBOBOX_init_combo (combo);
    } else {
        COMBOBOX_clear_history (combo_info);
    }
    COMBOBOX_set_quick_activate(combo_info, GINT_TO_POINTER(TRUE));
    COMBOBOX_set_activate_function(combo_info, activate_entry);
    COMBOBOX_set_cancel_function(combo_info, cancel_entry);
    COMBOBOX_set_activate_user_data(combo_info, dialog);
    COMBOBOX_set_cancel_user_data(combo_info, dialog);
    COMBOBOX_set_extra_key_completion_function(combo_info, extra_key_completionR);
    COMBOBOX_set_extra_key_completion_data(combo_info, &extra_key);
    /*combo_info->activate_func = activate_entry;
    combo_info->activate_user_data = dialog;
    combo_info->cancel_func = cancel_entry;
    combo_info->cancel_user_data = dialog;
    xfc_fun->extra_key_completion = extra_key_completionR;
    xfc_fun->extra_key_data = &extra_key;*/

    COMBOBOX_read_history (combo_info, history_file);
    COMBOBOX_set_combo (combo_info);

    if(filechooser_action == GTK_FILE_CHOOSER_ACTION_OPEN || filechooser_action == GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER) {
        filechooser_v.combo_info = combo_info;
        filechooser_v.parent = dialog;
        filechooser_v.folder = folder;
        filechooser_v.entry = (GtkEntry *)COMBOBOX_get_entry_widget (combo_info);
        filechooser_v.activate_func = activate_entry;
        filechooser_v.activate_user_data = dialog;

        filechooser_v.filechooser_action = filechooser_action;
        preload (filechooser_v.folder);

        button = gtk_button_new ();
        GtkWidget *image = gtk_image_new_from_stock (GTK_STOCK_DIRECTORY, GTK_ICON_SIZE_BUTTON);
        gtk_button_set_image ((GtkButton *) button, image);
        vbox = gtk_vbox_new (FALSE, 6);
        gtk_box_pack_start (GTK_BOX (hbox), vbox, FALSE, FALSE, 0);
        gtk_box_pack_start (GTK_BOX (vbox), button, FALSE, FALSE, 0);
        g_signal_connect (button, "clicked", G_CALLBACK (filechooser), (gpointer) (&filechooser_v));
        gtk_widget_show (button);
    }
    gtk_widget_show_all (hbox);

    if(path) {
        gchar *type = MIME_type ((void *)path, NULL);
        gchar *p = NULL;
        if(!type)
            type = MIME_magic (path);
        if(type)
            p = MIME_command (type);
        g_free (type);
        if(p){
	    DBG("COMBO: setting entry to %s\n", p);
	    COMBOBOX_set_entry (combo_info, p);
	}
    }

    if (entry_text) {
	DBG ("COMBO: setting combo to %s\n", entry_text);
	COMBOBOX_set_entry (combo_info, entry_text);
    }
    if (!entry_text && ! path) {
	COMBOBOX_set_default (combo_info);
	DBG ("COMBO: setting combo to default (!entry_text && ! path)\n");
    }

    if(check_label && flagfile) {
        extra_key.check1 = (GtkWidget *) gtk_check_button_new_with_mnemonic (check_label);
        g_signal_connect (extra_key.check1, "toggled", G_CALLBACK (toggle_flag), (gpointer) & extra_key);
        extra_key.entry = GTK_ENTRY(COMBOBOX_get_entry_widget(combo_info));
        gtk_box_pack_start (GTK_BOX ((GTK_DIALOG (dialog))->action_area), GTK_WIDGET (extra_key.check1), FALSE, FALSE, 0);
        gtk_widget_show (extra_key.check1);
    }

    button = gtk_button_new_from_stock (GTK_STOCK_CANCEL);
    gtk_widget_show (button);
    gtk_dialog_add_action_widget (GTK_DIALOG (dialog), button, GTK_RESPONSE_CANCEL);
    button = gtk_button_new_from_stock (GTK_STOCK_OK);
    gtk_widget_show (button);

    gtk_dialog_add_action_widget (GTK_DIALOG (dialog), button, GTK_RESPONSE_YES);
    gtk_widget_realize (dialog);
    if(flagfile) {
        extra_key.flagfile = flagfile;
        extra_key_completionR (&extra_key);
    } else
        extra_key.flagfile = NULL;

    if(title_txt)
        gtk_window_set_title (GTK_WINDOW (dialog), title_txt);
    else
        gdk_window_set_decorations (dialog->window, GDK_DECOR_BORDER);

    /* show dialog and return */
    response = gtk_dialog_run (GTK_DIALOG (dialog));
    //DBG("response=%d yes=%d none=%d\n",response, GTK_RESPONSE_YES, GTK_RESPONSE_NONE);fflush(NULL);
    if(response == GTK_RESPONSE_YES) {
        if(response_txt)
            g_strstrip (response_txt);
        const gchar *et = COMBOBOX_get_entry (combo_info);
        if(et && strlen (et)) {
            response_txt = (gchar *) malloc ((strlen (et) + 3) * sizeof (gchar));
            memset (response_txt, 0, (strlen (et) + 3) * sizeof (gchar));
            strcpy (response_txt, et);
            COMBOBOX_save_to_history ((char *)history_file, (char *)response_txt);
            if(flagfile)
                save_flags (&extra_key);
        }
        if(flagfile) {
            gboolean active = gtk_toggle_button_get_active ((GtkToggleButton *)
                                                            extra_key.check1);
            //DBG("active=%d\n",active);
            if(active)
                response_txt[strlen (response_txt) + 1] = 1;
        }
    }
    gtk_widget_hide (dialog);
    // cleanup combobox module objects:
    COMBOBOX_destroy_combo(combo_info);
    gtk_widget_destroy (dialog);

    return (const gchar *)(response_txt);
}
/*********   find callbacks   *********/

void
rodent_glob_activate (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    DBG ("FGR: rodent_glob_activate() view_p=0x%lx\n", (unsigned long)view_p);
    record_entry_t *en = NULL;
    if(rfm_population_try_read_lock (view_p)) {
        if(view_p->selected_p && view_p->selected_p->en && g_file_test (view_p->selected_p->en->path, G_FILE_TEST_IS_DIR)
           && g_list_length (view_p->selection_list) == 1) {
            en = view_p->selected_p->en;
        } else if(view_p->en) {
            en = view_p->en;
        }

        if(!rfm_rational ("rfm", "fgr", en, widgets_p, "do_find")) {
            DBG ("FGR: unable to open fgr dialog\n");
        } else {
            DBG ("FGR: fgr dialog OK\n");
        }
        rfm_population_read_unlock (view_p);
    }
    return;
}

void
rodent_refresh_activate (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if (view_p->protected) {
        rfm_diagnostics(widgets_p, "xffm/warning",NULL);
        rfm_diagnostics(widgets_p, "xffm_tag/stderr",_("Reload"),": ", strerror(EBUSY),"\n",NULL);
        return;
    }
    if(view_p->en && view_p->en->path) {
        rfm_rm_local_cache (view_p);
        // clear any thumbnail previews:
        gchar *test_image = g_build_filename (view_p->en->path, "test.jpg", NULL);
        gchar *filename = rfm_get_thumbnail_path (test_image, 48);
        gchar *dirname = g_path_get_dirname (filename);
        gchar *rm = g_find_program_in_path ("rm");

        pid_t pid = fork ();
        if(!pid) {
            execl (rm, rm, "-rf", dirname, NULL);
            _exit (1);
        }
        g_free (test_image);
        g_free (filename);
        g_free (dirname);
        g_free (rm);
        int status;
        waitpid (pid, &status, 0);

    }
    // rodent_touch takes care of invalidating cache, but involves pipe creation
    // rodent_touch(widgets_p,view_p->en->path);

    record_entry_t *en = rfm_copy_entry (view_p->en);
    rodent_refresh (widgets_p, en);
    return;
}

void
rodent_back_activate (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if (view_p->protected) {
        rfm_diagnostics(widgets_p, "xffm/warning",NULL);
        rfm_diagnostics(widgets_p, "xffm_tag/stderr",_("Go back"),": ",strerror(EBUSY),"\n",NULL);
        return;
    }
    GList *last;
    record_entry_t *en;
    TRACE ("view goback\n");

    last = g_list_last (view_p->go_list);
    if(!last)
        return;
    en = last->data;
    if(en)
        view_p->module_name = en->module;
    else
        view_p->module_name = NULL;

    rodent_full_reload_view ((gpointer) view_p, en);
    rodent_pop_view_go_history (view_p);
}

void
rodent_jump_to_activate (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;

    gchar *f = g_build_filename (g_get_user_data_dir (), GOTO_DBH_FILE, NULL);
    const gchar *choice = rodent_get_response_history (NULL, _("Go To"), _("Path"), f,
                                                       NULL, NULL, NULL, NULL,
                                                       GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER,
                                                       g_get_home_dir ());
    g_free (f);

    record_entry_t *new_en;
    //const gchar *g=fix_relative_path(view_p,choice);
    if (!choice || strlen(choice)==0) return;
    if(!g_file_test (choice, G_FILE_TEST_IS_DIR)) {
        rfm_show_text(widgets_p);
        rfm_diagnostics (widgets_p, "xffm/error", NULL);
        rfm_diagnostics (widgets_p, "xffm_tag/stderr", choice, ": ", strerror (ENOENT), "\n", NULL);
        return;
    }
    // make path absolute
    else if(chdir (choice) < 0) {
        rfm_show_text(widgets_p);
        rfm_diagnostics (widgets_p, "xffm/error", NULL);
        rfm_diagnostics (widgets_p, "xffm_tag/stderr", choice, ": ", strerror (errno), "\n", NULL);
	return;
    } 
    gchar *g = g_get_current_dir ();
    SETWD();

    if(view_p->child_constructor) {
        (*view_p->child_constructor) (widgets_p, (char *)g);
    } else {
        rodent_push_view_go_history (view_p);
        new_en = rfm_stat_entry (g, __LOCAL_TYPE);
        view_p->module_name = NULL;
        rfm_save_to_go_history ((gchar *) g);
        rodent_refresh (widgets_p, new_en);
    }
    g_free (g);
}

void
rodent_home_activate (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    if(!view_p) {
        g_warning ("!view_p");
        return;
    }
    if (view_p->protected) {
        rfm_diagnostics(widgets_p, "xffm/warning",NULL);
        rfm_diagnostics(widgets_p, "xffm_tag/stderr",_("Home"),": ",strerror(EBUSY),"\n",NULL);
        return;
    }
    if(view_p->child_constructor) {
        (*view_p->child_constructor) (widgets_p, (char *)g_get_home_dir ());
    } else {
        record_entry_t *en = rfm_stat_entry ((gchar *) g_get_home_dir (), __ROOT_TYPE | __LOCAL_TYPE);
        rodent_push_view_go_history (view_p);
        view_p->module_name = NULL;
        rodent_refresh (widgets_p, en);
    }
}

void
rodent_host_activate (GtkMenuItem * menuitem, gpointer user_data) {
    widgets_t *widgets_p = (widgets_t *) user_data;
    view_t *view_p = widgets_p->view_p;
    record_entry_t *en = NULL;
    if(!view_p)
        return;
    if(view_p->child_constructor) {
        (*view_p->child_constructor) (widgets_p, "NULL");
    } else {
        rodent_push_view_go_history (view_p);
        rodent_refresh (widgets_p, en);
    }
}

