/*
 * Copyright (C) 2002-2010 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <dirent.h>
#include <errno.h>
#include <glob.h>
#include <limits.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include "rodent_libs.h"

#define is_digit(x) ((x)>='0' && (x)<='9')

#include "rodent_remove.i"

GtkWidget *
rodent_create_remove (widgets_t * widgets_p) {
    GtkWidget *remove;
    GtkWidget *vbox2;
    GtkWidget *hbox26;
    GtkWidget *question;
    GtkWidget *warning;
    GtkWidget *vbox12;
    GtkWidget *label16;
    GtkWidget *label20;
    GtkWidget *hbox9;
    GtkWidget *togglebutton1;
    GtkWidget *hbuttonbox1;
    GtkWidget *cancelbutton;
    GtkWidget *alignment1;
    GtkWidget *hbox27;
    GtkWidget *adicon;
    GtkWidget *label21;
    GtkWidget *removebutton;
#ifdef GNU_SHRED
    GtkWidget *shredbutton;
#endif
    GtkWidget *hbox28;
    GtkWidget *image1;
    GtkWidget *label22;
    GdkPixbuf *pb;

    remove = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title (GTK_WINDOW (remove), _("Warning"));
    gtk_window_set_modal (GTK_WINDOW (remove), TRUE);

    vbox2 = gtk_vbox_new (FALSE, 0);
    gtk_widget_show (vbox2);
    gtk_container_add (GTK_CONTAINER (remove), vbox2);

    hbox26 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox26);
    gtk_box_pack_start (GTK_BOX (vbox2), hbox26, TRUE, TRUE, 0);

    pb = rfm_get_pixbuf ("xffm/question", SIZE_ICON);
    question = gtk_image_new_from_pixbuf (pb);
    gtk_widget_show (question);
    gtk_box_pack_start (GTK_BOX (hbox26), question, TRUE, TRUE, 0);
    gtk_misc_set_padding (GTK_MISC (question), 5, 0);

    pb = rfm_get_pixbuf ("xffm/warning", SIZE_ICON);
    warning = gtk_image_new_from_pixbuf (pb);
    gtk_widget_show (warning);
    gtk_box_pack_start (GTK_BOX (hbox26), warning, TRUE, TRUE, 0);
    gtk_misc_set_padding (GTK_MISC (warning), 5, 0);

    vbox12 = gtk_vbox_new (FALSE, 0);
    gtk_widget_show (vbox12);
    gtk_box_pack_start (GTK_BOX (hbox26), vbox12, TRUE, TRUE, 0);

    label16 = gtk_label_new ("");
    gtk_widget_show (label16);
    gtk_box_pack_start (GTK_BOX (vbox12), label16, FALSE, FALSE, 0);

    label20 = gtk_label_new ("label20");
    gtk_widget_show (label20);
    gtk_box_pack_start (GTK_BOX (vbox12), label20, FALSE, FALSE, 0);

    hbox9 = gtk_hbox_new (FALSE, 0);
    gtk_widget_show (hbox9);
    gtk_box_pack_start (GTK_BOX (vbox12), hbox9, TRUE, TRUE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (hbox9), 5);

    togglebutton1 = gtk_check_button_new_with_mnemonic (_("All"));
    gtk_widget_show (togglebutton1);
    gtk_box_pack_start (GTK_BOX (hbox9), togglebutton1, FALSE, FALSE, 0);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (togglebutton1), TRUE);

    hbuttonbox1 = gtk_hbutton_box_new ();
    gtk_widget_show (hbuttonbox1);
    gtk_box_pack_start (GTK_BOX (vbox12), hbuttonbox1, TRUE, TRUE, 0);
    gtk_container_set_border_width (GTK_CONTAINER (hbuttonbox1), 5);
    gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox1), GTK_BUTTONBOX_END);
    gtk_box_set_spacing (GTK_BOX (hbuttonbox1), 5);

    cancelbutton = gtk_button_new ();
    gtk_widget_show (cancelbutton);
    gtk_container_add (GTK_CONTAINER (hbuttonbox1), cancelbutton);
    GTK_WIDGET_SET_FLAGS (cancelbutton, GTK_CAN_DEFAULT);

    alignment1 = gtk_alignment_new (0.5, 0.5, 0, 0);
    gtk_widget_show (alignment1);
    gtk_container_add (GTK_CONTAINER (cancelbutton), alignment1);

    hbox27 = gtk_hbox_new (FALSE, 2);
    gtk_widget_show (hbox27);
    gtk_container_add (GTK_CONTAINER (alignment1), hbox27);

    pb = rfm_get_pixbuf ("xffm/stock_cancel", SIZE_DIALOG);
    adicon = gtk_image_new_from_pixbuf (pb);
    gtk_widget_show (adicon);
    gtk_box_pack_start (GTK_BOX (hbox27), adicon, FALSE, FALSE, 0);

    label21 = gtk_label_new_with_mnemonic (_("Cancel"));
    gtk_widget_show (label21);
    gtk_box_pack_start (GTK_BOX (hbox27), label21, FALSE, FALSE, 0);
/****************/
#ifdef GNU_SHRED
    shredbutton = gtk_button_new ();
    gtk_widget_show (shredbutton);
    gtk_container_add (GTK_CONTAINER (hbuttonbox1), shredbutton);
    GTK_WIDGET_SET_FLAGS (shredbutton, GTK_CAN_DEFAULT);
    hbox28 = gtk_hbox_new (FALSE, 2);
    gtk_widget_show (hbox28);
    gtk_container_add (GTK_CONTAINER (shredbutton), hbox28);

    pb = rfm_get_pixbuf ("xffm/no-read", SIZE_DIALOG);
    image1 = gtk_image_new_from_pixbuf (pb);
    gtk_widget_show (image1);
    gtk_box_pack_start (GTK_BOX (hbox28), image1, FALSE, TRUE, 0);

    label22 = gtk_label_new_with_mnemonic (_("Shred"));
    gtk_widget_show (label22);
    gtk_box_pack_start (GTK_BOX (hbox28), label22, TRUE, FALSE, 0);
#endif

/****************/

    removebutton = gtk_button_new ();
    gtk_widget_show (removebutton);
    gtk_container_add (GTK_CONTAINER (hbuttonbox1), removebutton);
    GTK_WIDGET_SET_FLAGS (removebutton, GTK_CAN_DEFAULT);

    hbox28 = gtk_hbox_new (FALSE, 2);
    gtk_widget_show (hbox28);
    gtk_container_add (GTK_CONTAINER (removebutton), hbox28);

    pb = rfm_get_pixbuf ("xffm/broken", SIZE_DIALOG);
    image1 = gtk_image_new_from_pixbuf (pb);
    gtk_widget_show (image1);
    gtk_box_pack_start (GTK_BOX (hbox28), image1, FALSE, TRUE, 0);

    label22 = gtk_label_new_with_mnemonic (_("Remove"));
    gtk_widget_show (label22);
    gtk_box_pack_start (GTK_BOX (hbox28), label22, TRUE, FALSE, 0);

    /* Store pointers to all widgets, for use by rodent_get_widget_by_name(). */
    RODENT_HOOKUP_OBJECT (remove, remove, "remove");
    RODENT_HOOKUP_OBJECT (remove, vbox2, "vbox2");
    RODENT_HOOKUP_OBJECT (remove, hbox26, "hbox26");
    RODENT_HOOKUP_OBJECT (remove, question, "question");
    RODENT_HOOKUP_OBJECT (remove, warning, "warning");
    RODENT_HOOKUP_OBJECT (remove, vbox12, "vbox12");
    RODENT_HOOKUP_OBJECT (remove, label16, "label16");
    RODENT_HOOKUP_OBJECT (remove, label20, "label20");
    RODENT_HOOKUP_OBJECT (remove, hbox9, "hbox9");
    RODENT_HOOKUP_OBJECT (remove, togglebutton1, "togglebutton1");
    RODENT_HOOKUP_OBJECT (remove, hbuttonbox1, "hbuttonbox1");
    RODENT_HOOKUP_OBJECT (remove, cancelbutton, "cancelbutton");
    RODENT_HOOKUP_OBJECT (remove, alignment1, "alignment1");
    RODENT_HOOKUP_OBJECT (remove, hbox27, "hbox27");
    RODENT_HOOKUP_OBJECT (remove, adicon, "adicon");
    RODENT_HOOKUP_OBJECT (remove, label21, "label21");
    RODENT_HOOKUP_OBJECT (remove, removebutton, "removebutton");
#ifdef GNU_SHRED
    RODENT_HOOKUP_OBJECT (remove, shredbutton, "shredbutton");
#endif
    RODENT_HOOKUP_OBJECT (remove, hbox28, "hbox28");
    RODENT_HOOKUP_OBJECT (remove, image1, "image1");
    RODENT_HOOKUP_OBJECT (remove, label22, "label22");

    gtk_widget_grab_focus (cancelbutton);
    return remove;
}

int
rodent_query_rm (widgets_t * widgets_p, const gchar * nfile, const gchar * ofile, gboolean overwrite) {
    int result;
    result = make_single_rm_dialog (widgets_p, nfile, ofile);
    if(result == RM_YES || result == RM_YES_ALL) {
        if(!overwrite) {
            if(unlink (nfile) < 0) {
                rfm_diagnostics (widgets_p, "xffm_tag/red", _("Error while deleting."), ": unlink(", nfile, ")--> ", strerror (errno), "\n", NULL);
            }
        }
        return TRUE;
    }
    return FALSE;
}

void
rodent_remove (widgets_t * widgets_p, GList * selection_list) {

    GList *remove_list = NULL;
    GList *tmp;

    if(!selection_list) {
        g_warning ("NULL selection list for remove");
        return;
    }

    if(remove_list) {
        g_list_free (remove_list);
        remove_list = NULL;
    }
    for(tmp = selection_list; tmp; tmp = tmp->next) {
        record_entry_t *en = rfm_copy_entry ((record_entry_t *) tmp->data);
        remove_list = g_list_append (remove_list, en);
    }
    if(!g_list_length (remove_list)) {
        rfm_status (widgets_p, "xffm/warning,", strerror (EINVAL), NULL);
        return;
    }

    while(remove_list != NULL && remove_list->data != NULL) {
        DBG ("REMOVE: remove dialog...\n");
        make_dialog (widgets_p, &remove_list);
    }
    if(remove_list) {
        g_list_free (remove_list);
    }
}
