
/*
 * Copyright (C) 2002-2011 Edscott Wilson Garcia
 * EMail: edscott@xfce.org
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/wait.h>
#include <limits.h>

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include "rodent_libs.h"

#define HISTORY_LIMIT 20
#define GLOB "fgr"
#define FIND "find"

/* this should be first 2 lines after headers: */
G_MODULE_EXPORT LIBRFM_MODULE 

typedef struct find_struct_t {
    widgets_t *widgets_p;
    void *find_combo_info;
    void *findpath_combo_info;
    void *findgrep_combo_info;
} find_struct_t;

#include "combobox-module.h"

#include "find-module.h"
#include "find-module_gui.h"

// this static variables mean we cannot have two or more threads
// running find dialogs. This is OK today because the dialog
// is transient and blocks the main thread from creating any more
// dialogs (dialog run in main thread, not generated thread)
// This may change in the future and for that we will need a method
// which is safe on multiple calls to destroy function...
static GtkWidget *find_dialog;
static find_struct_t *find_struct_p = NULL;
#include "find-module.i"
static filechooser_t filechooser_v;


G_MODULE_EXPORT void *
do_find (void *p, void *q) {
    record_entry_t *en = p;
    widgets_t *widgets_p = q;

    const char *path = NULL;
    if(en)
        path = en->path;
    DBG ("FIND: path=%s\n", en->path);
    //rfm_status(widgets_p, "xffm/warning", _("find dialog active"), NULL);

    GList *type_list = NULL;
	
    find_struct_p =
	    (find_struct_t *) malloc (sizeof (find_struct_t));
   
    memset (find_struct_p, 0, sizeof (find_struct_t));
    find_struct_p->widgets_p = widgets_p;

    GtkComboBoxEntry *combo;
    GtkComboBoxEntry *combopath;
    GtkComboBoxEntry *combogrep;

    TRACE ("do_find_path now, path is %s, widgets is 0x%lx\n", path, (long)widgets_p);
    find_dialog = create_find_dialog (widgets_p);
    DBG ("FGR: find_dialog=0x%lx\n", (unsigned long)find_dialog);

    combo = (GtkComboBoxEntry *) 
	rodent_get_widget_by_name (find_dialog, "filter_combo");
    if (!combo) {g_warning("cannot find widget filter_combo"); }
    combopath = (GtkComboBoxEntry *) 
	rodent_get_widget_by_name (find_dialog, "path_combo");
    if (!combopath) {g_warning("cannot find widget path_combo"); }
    combogrep = (GtkComboBoxEntry *) 
	rodent_get_widget_by_name (find_dialog, "grep_combo");
    if (!combogrep) {g_warning("cannot find widget grep_combo"); }
    gtk_widget_show (find_dialog);

    // transient, not for DESKVIEW_TYPE 
    view_t *view_p=widgets_p->view_p;
    if(view_p->type == ICONVIEW_TYPE) {
        gtk_window_set_transient_for (GTK_WINDOW (find_dialog), GTK_WINDOW (widgets_p->window));
    }

    g_signal_connect (G_OBJECT (rodent_get_widget_by_name (find_dialog, "apply_button")),
                      "clicked", G_CALLBACK (on_apply_clicked), (gpointer) find_struct_p);
    g_signal_connect (G_OBJECT (rodent_get_widget_by_name (find_dialog, "find_button")),
                      "clicked", G_CALLBACK (on_find_clicked), (gpointer) find_struct_p);
    g_signal_connect (G_OBJECT (rodent_get_widget_by_name (find_dialog, "close_button")),
                      "clicked", G_CALLBACK (on_find_close), (gpointer) widgets_p);
    g_signal_connect (G_OBJECT (rodent_get_widget_by_name (find_dialog, "togglebutton2")),
                      "toggled", G_CALLBACK (on_help_filter), widgets_p);
    g_signal_connect (G_OBJECT (rodent_get_widget_by_name (find_dialog, "togglebutton3")),
                      "toggled", G_CALLBACK (on_help_grep), widgets_p);

    gchar *f = g_build_filename (g_get_user_data_dir (), FIND_GREP_DBH_FILE, NULL);

    find_struct_p->findgrep_combo_info = COMBOBOX_init_combo (combogrep);
    if (find_struct_p->findgrep_combo_info == NULL) {
	g_warning("cannot initialize findgrep_combo_info");
    }

    COMBOBOX_set_activate_function(
	    find_struct_p->findgrep_combo_info, on_find_clicked_wrapper);
    COMBOBOX_set_activate_user_data(
	    find_struct_p->findgrep_combo_info, find_struct_p);

    COMBOBOX_read_history (find_struct_p->findgrep_combo_info, f);
    COMBOBOX_set_combo (find_struct_p->findgrep_combo_info);
    g_free (f);

    f = g_build_filename (g_get_user_data_dir (), FIND_DBH_FILE, NULL);

    find_struct_p->find_combo_info = COMBOBOX_init_combo (combo);
    COMBOBOX_set_activate_function(
	    find_struct_p->find_combo_info, on_find_clicked_wrapper);
    COMBOBOX_set_activate_user_data(
	    find_struct_p->find_combo_info, find_struct_p);

    COMBOBOX_read_history (find_struct_p->find_combo_info, f);
    COMBOBOX_set_combo (find_struct_p->find_combo_info);
    g_free (f);
    COMBOBOX_set_default(find_struct_p->find_combo_info);

    f = g_build_filename (g_get_user_data_dir (), GOTO_DBH_FILE, NULL);
    //  g_build_filename(g_get_user_data_dir(), FIND_PATHS_DBH_FILE, NULL);
    find_struct_p->findpath_combo_info = COMBOBOX_init_combo (combopath);

    COMBOBOX_set_activate_function(
	    find_struct_p->findpath_combo_info, on_find_clicked_wrapper);
    COMBOBOX_set_activate_user_data(
	    find_struct_p->findpath_combo_info, find_struct_p);

    COMBOBOX_read_history (find_struct_p->findpath_combo_info, f);
    COMBOBOX_set_combo (find_struct_p->findpath_combo_info);
    g_free (f);

    filechooser_v.folder = path;
    COMBOBOX_set_entry (find_struct_p->findpath_combo_info, (char *)path);

    filechooser_v.parent = find_dialog;
    filechooser_v.filechooser_action = GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER;
    filechooser_v.combo_info = find_struct_p->findpath_combo_info;

    g_signal_connect (rodent_get_widget_by_name (find_dialog, "fileselector"), "clicked",
                      G_CALLBACK (path_filechooser), (gpointer) (&filechooser_v));

    int i;
    for(i = 0; ftypes[i] != NULL; i++) {
	type_list = g_list_append (type_list, _(ftypes[i]));
    }

    fill_string_option_menu ((GtkComboBox *)
                             rodent_get_widget_by_name (find_dialog, "file_type_om"), 
			     type_list);

    gtk_label_set_text ((GtkLabel *) rodent_get_widget_by_name (find_dialog, "filter_help"),
                        _("Basic rules:\n" "\n"
                          "*  Will match any character zero or more times.\n"
                          "?  Will match any character exactly one time\n"
                          "[] Match any character within the [] \n"
                          "^  Match at beginning of string\n" "$  Match at end of string \n"));
    gtk_label_set_text ((GtkLabel *) rodent_get_widget_by_name (find_dialog, "regexp_help"),
                        _("Reserved characters for extended regexp are\n"
                          ". ^ $ [ ] ? * + { } | \\ ( ) : \n"
                          "In  basic regular expressions the metacharacters\n"
                          "?, +, {, |, (, and ) lose their special meaning.\n"
                          "\n"
                          "The  period  .   matches  any  single  character.\n"
                          "The caret ^ matches at the start of line.\n"
                          "The dollar $ matches at the end of line.\n" "\n"
                          "Characters within [ ] matches any single \n"
                          "       character in the list.\n"
                          "Characters within [^ ] matches any single\n"
                          "       character *not* in the list.\n"
                          "Characters inside [ - ] matches a range of\n"
                          "       characters (ie [0-9] or [a-z]).\n" "\n"
                          "A regular expression may be followed by one\n"
                          "       of several repetition operators:\n"
                          "?      The preceding item is optional and matched\n"
                          "       at most once.\n"
                          "*      The preceding item will be matched zero\n"
                          "       or more times.\n"
                          "+      The preceding item will be matched one or\n"
                          "       more times.\n"
                          "{n}    The preceding item is matched exactly n times.\n"
                          "{n,}   The preceding item is matched n or more times.\n"
                          "{n,m}  The preceding item is matched at least n times,\n"
                          "       but not more than m times.\n" "\n"
                          "To match any reserved character, precede it with \\. \n"
                          "\n"
                          "Two regular expressions may be joined by the logical or\n"
                          "       operator |.\n"
                          "Two regular expressions may be concatenated.\n" "\n"
                          "More information is available by typing \"man grep\"\n"));

    g_signal_connect(G_OBJECT (gtk_bin_get_child(GTK_BIN(combogrep))), "changed", G_CALLBACK (grep_options), (gpointer) find_struct_p);
    g_signal_connect (G_OBJECT (find_dialog), "destroy_event", G_CALLBACK (destroy_find), (gpointer) find_struct_p);
    g_signal_connect (G_OBJECT (find_dialog), "delete_event", G_CALLBACK (destroy_find), (gpointer) find_struct_p);


    return GINT_TO_POINTER (1);
}

G_MODULE_EXPORT const gchar *
g_module_check_init (GModule * module) {
#ifdef ENABLE_NLS
    bindtextdomain (GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);
# ifdef HAVE_BIND_TEXTDOMAIN_CODESET
    bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
# endif
    TRACE ("domain=%s", GETTEXT_PACKAGE);
#endif
    return NULL;
}

G_MODULE_EXPORT void *
module_init (void
    ) {
    return NULL;
}
