/*
 * Copyright (C) 2002-2011 Edscott Wilson Garcia
 * EMail: edscott@imp.mx
 *
 * Copyright (c) 2003 Benedikt Meurer <benedikt.meurer@unix-ag.uni-siegen.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif
#ifdef HAVE_CTYPE_H
# include <ctype.h>
#endif
#ifdef HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif
#ifdef HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifdef HAVE_SYS_PARAM_H
# include <sys/param.h>
#endif
#ifdef HAVE_SYS_UCRED_H
# include <sys/ucred.h>
#endif
#ifdef HAVE_SYS_MOUNT_H
# include <sys/mount.h>
#endif

#include <stdio.h>
#include <errno.h>
#ifdef HAVE_FSTAB_H
# include <fstab.h>
#endif
#include <limits.h>
#ifdef HAVE_MEMORY_H
# include <memory.h>
#endif
#ifdef HAVE_MNTENT_H
# include <mntent.h>
#endif
#ifdef HAVE_PATHS_H
# include <paths.h>
#endif
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>

#ifdef HAVE_SYS_FS_TYPES_H
# include <sys/fs_types.h>
#endif

#ifdef HAVE_SYS_MNTTAB_H
# include <sys/mnttab.h>
#endif

#ifdef HAVE_SYS_VFS_H
# include <sys/vfs.h>
#endif

#ifdef HAVE_SYS_VFSTAB_H
# include <sys/vfstab.h>
#endif

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include "rodent_libs.h"

#include "fstab-module.h"


/* this should be first 2 lines after headers: */
G_MODULE_EXPORT LIBRFM_MODULE
/* exported prototypes */
void *is_mounted (void *p);
void *is_in_fstab (void *p);
const gchar *module_name (void
    );

#include "fstab-module.i"

G_MODULE_EXPORT const gchar *
g_module_check_init (GModule * module) {
#ifdef ENABLE_NLS
    bindtextdomain (GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);
# ifdef HAVE_BIND_TEXTDOMAIN_CODESET
    bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
# endif
    TRACE ("domain=%s", GETTEXT_PACKAGE);
#endif
    return NULL;
}

/****************   void  functions  ***********************/

G_MODULE_EXPORT const gchar *
module_name (void
    ) {
    return "fstab";
}

G_MODULE_EXPORT const gchar *
submodule_dir (void
    ) {
    return "rfm";
}

G_MODULE_EXPORT void *
is_root_module (void
    ) {
#ifdef THIS_IS_BSD
// BSD functions are not thread safe... ergo, disable them for the time being.
    return NULL;
#else
    return GINT_TO_POINTER (1);
#endif
}

G_MODULE_EXPORT void *
module_init (void
    ) {
    return NULL;
}

G_MODULE_EXPORT void *
exec_name (void
    ) {
    return "rodent-fstab";
}

G_MODULE_EXPORT void *
private_popup (void *p, void *q) {
    return NULL;
    GtkWidget *w;
    widgets_t *widgets_p = (widgets_t *) q;
    view_t *view_p=widgets_p->view_p;
    gchar *label = _("Mount Point"),
        *altlabel = NULL;
    record_entry_t *en = (record_entry_t *) p;

    if(!en)
        return NULL;
    if(private_popup_widget)
        gtk_widget_destroy (private_popup_widget);

    TRACE ("en->st=0x%lx", (long)en->st);

    if(en && en->st && en->path) {
        label = en->path;
        if(strlen (label) > 30) {
            altlabel = g_strdup (en->path);
            altlabel[30] = 0;
            label = altlabel;
        }
    }

    private_popup_widget = rodent_mk_menu (widgets_p,   /* window */
                                           label,       /* label */
                                           NULL,        /* name */
                                           NULL,        /* parent */
                                           NULL, NULL); /* callback (or NULL) */
    g_free (altlabel);

    w = gtk_image_menu_item_new_with_mnemonic ("whatever");
    g_object_set_data (G_OBJECT (w), "widgets_p", widgets_p);
    rodent_mk_pixmap_menu (widgets_p, "xffm/stock_refresh", w, MENU_PIXMAP);
    gtk_container_add (GTK_CONTAINER (private_popup_widget), w);
//      g_signal_connect ((gpointer) w, "activate", G_CALLBACK (proc_reload), en);
    gtk_widget_show (w);

    gtk_menu_popup (GTK_MENU (private_popup_widget), NULL, NULL, NULL, NULL, 3, view_p->eventtime);

    return GINT_TO_POINTER (1);
}

/*******************  natural functions **********************/

G_MODULE_EXPORT void *
get_dnd_path (void *p) {
    static gchar *desktop = NULL;
    if(!desktop) {
        desktop = g_strdup_printf ("%s%c%s%c%s",
                                   PACKAGE_DATA_DIR, G_DIR_SEPARATOR, "applications", G_DIR_SEPARATOR, "Rodent-fstab.desktop");
    }
    return (void *)desktop;
}

G_MODULE_EXPORT void *
module_icon_id (void *p) {
    DBG ("fstab: resolving module icon id\n");
    return "inode/directory/composite/stock_insert-object";
}

G_MODULE_EXPORT void *
item_icon_id (void *p) {
    //DBG("FSTAB_ICON: resolving icon id\n");
    record_entry_t *en = (record_entry_t *) p;
    const gchar *icon_id = NULL;
    int type = en->type;
    if(IS_PARTITION_TYPE (type))
        icon_id = "xffm/disk";
    else if(IS_SLNK (type))
        icon_id = "inode/symlink";
    else if(IS_PROC_TYPE (type))
        icon_id = "xffm/chardevice";
    else if(IS_CDFS_TYPE (type))
        icon_id = "xffm/cdrom";
    //else if (IS_SMB_TYPE(type)) t = "network-fs";
    else if(strstr (en->path, "floppy"))
        icon_id = "xffm/floppy";
    else if(strstr (en->path, "cdrom"))
        icon_id = "xffm/cdrom";
    else if(strstr (en->path, "cdrw"))
        icon_id = "xffm/cdrom";
    else if(strstr (en->path, "dvd"))
        icon_id = "xffm/dvd";
    else if(is_in_fstab (en->path) && !IS_MTAB_TYPE (type)) {
        icon_id = "inode/mount-point";
    } else {
        icon_id = "inode/directory";
    }

    static gchar *id = NULL;
    g_free (id);
    if(is_mounted ((void *)(en->path))) {
        id = g_strconcat (icon_id, "/composite/stock_yes", NULL);
    } else {
        id = g_strconcat (icon_id, "/composite/stock_no", NULL);
    }
    return (void *)id;
}

G_MODULE_EXPORT void *
module_label (void *p) {
    return g_strdup (_("Mount Point"));
}

G_MODULE_EXPORT void *
is_selectable (void *p) {
    record_entry_t *en = (record_entry_t *) p;

    if(g_file_test (en->path, G_FILE_TEST_IS_DIR))
        return (void *)"Yes";
    if(IS_PARTITION_TYPE (en->type))
        return (void *)"Yes";

    return NULL;
}

G_MODULE_EXPORT void *
valid_drop_site (void *p) {
    record_entry_t *en = (record_entry_t *) p;
    if(!en)
        return NULL;
    if(IS_SDIR (en->type))
        return (void *)((long)1);
    return NULL;
}

G_MODULE_EXPORT void *
entry_tip (void *p) {
    record_entry_t *en = (record_entry_t *) p;
    if(!en)
        return NULL;
    gchar *module_text = NULL;
    gchar *mnt_dir = NULL;
    gchar *mnt_fsname = NULL;
    if(IS_PARTITION_TYPE (en->type)) {
        mnt_fsname = g_strdup (en->path);
        if(en->tag) {
            mnt_dir = g_strdup (en->tag);
        } else {
            mnt_dir = get_mnt_dir (mnt_fsname);
        }
    } else {
        mnt_dir = g_strdup (en->path);
        if(en->tag) {
            mnt_fsname = g_strdup (en->tag);
        } else {
            mnt_fsname = get_mnt_fsname (mnt_dir);
        }
    }

    if(!mnt_dir)
        mnt_dir = g_strdup ("");
    if(!mnt_fsname)
        mnt_fsname = g_strdup ("");
    module_text =
        g_strdup_printf ("%s %s\n%s: %s\n%s %s", 
                _("Mount point:"), mnt_dir, 
                _("Partition"), mnt_fsname,
                _("Partition type:"), 
                (IS_USER_TYPE (en->type)) ? _("user managed") : "root managed");
    g_free (mnt_dir);
    g_free (mnt_fsname);

    return module_text;
}

#if 0
/* NOTE: this function is current for linux only */
# if defined(HAVE_GETMNTENT) && defined(HAVE_STRUCT_MNTENT_MNT_DIR)
G_MODULE_EXPORT void *
entry_tip (void *p) {
    static gchar *tip = NULL;
    record_entry_t *en = p;
    const gchar *mnt_point;

    struct mntent *m;
    FILE *fstab_fd;

    g_free (tip);
    tip = NULL;
    if(!en || !en->path)
        return NULL;
    mnt_point = en->path;

    if((fstab_fd = setmntent ("/etc/mtab", "r")) == NULL) {
        fprintf (stderr, "Unable to open %s\n", "/etc/mtab");
        return (0);
    }
    while((m = getmntent (fstab_fd)) != NULL) {
        TRACE ("fstab=%s mountpoint=%s\n", m->mnt_dir, mnt_point);

        if(strcmp (m->mnt_dir, mnt_point) == 0) {
            endmntent (fstab_fd);
            /* gotcha */
            tip =
                g_strdup_printf (_
                                 ("Filesystem = %s\nMount type = %s\nMount options = %s"), m->mnt_fsname, m->mnt_type, m->mnt_opts);

            return tip;
        }
    }
    endmntent (fstab_fd);
    return NULL;
}
# endif
#endif
G_MODULE_EXPORT void *
plugin_info (void
    ) {
    return
        _("The FSTAB plugin keeps track of the mount points listed in /etc/fstab and allows you to easily mount or unmount these volumes with the popup menu, while navigating through the filesystem.");
}

/*******************  rational functions **********************/

#if 0
G_MODULE_EXPORT void *
double_click (void *p, void *q) {
    record_entry_t *en = (record_entry_t *) p;
    if(!en || !en->path || !g_file_test (en->path, G_FILE_TEST_EXISTS))
        return NULL;
    if(!is_in_fstab (en->path))
        return NULL;

    fstab_mount (p, q);

    return GINT_TO_POINTER (1);
}
#endif

static gchar *
get_options (gchar * mnt_src) {
    gchar *sudo = g_find_program_in_path ("sudo");
    if(!sudo)
        return NULL;
    gchar *device = g_strdup (mnt_src);
    while(strlen (device) && isdigit (device[strlen (device) - 1])) {
        device[strlen (device) - 1] = 0;
    }
    gchar *command = g_strdup_printf ("sudo -A fdisk -l %s", device);

    char line[2048];
    FILE *pipe;
    pipe = popen (command, "r");
    g_free (command);
    gboolean fat = FALSE;
    while(fgets (line, 2047, pipe) && !feof (pipe)) {
        if(strstr (line, mnt_src) && strstr (line, "FAT")) {
            fat = TRUE;
            break;
        }
    }
    pclose (pipe);
    gchar *options = NULL;
    if(fat) {
        options = g_strdup_printf ("uid=%d", geteuid ());
    }
    return options;
}

G_MODULE_EXPORT void *
do_properties (void *p, void *q) {
    widgets_t *widgets_p = p;
    if (!q){
	g_warning("!q");
	return NULL;
    }
    GList *tmp;
    GList * select_list=q;
    DBG("do_prop\n");
    
    record_entry_t *en = select_list->data;

    
    if(IS_PARTITION_TYPE (en->type)) {
        gchar *sudo = g_find_program_in_path ("sudo");
        if(!sudo){
	    g_warning("sudo command not found");
            goto done;
	}
        g_free (sudo);

        gchar *device = g_strdup (en->path);
	TRACE("*device = %s", device);
        while(strlen (device) && isdigit (device[strlen (device) - 1])) {
            device[strlen (device) - 1] = 0;
	    TRACE("device => %s", device);
        }
        gchar *argv[] = { "sudo", "-A", "fdisk", "-l", device, NULL };
	TRACE("%s -A fdisk -l %s", sudo, device);
	rfm_show_text (widgets_p);

        rfm_thread_run_argv (widgets_p, argv, FALSE);
        g_free (device);
        goto done;
    } else {
	DBG("IS_PARTITION_TYPE (en->type) == FALSE\n");
        return NULL;
    }
done:

    return GINT_TO_POINTER (1);

}

/*
G_MODULE_EXPORT void *show_local_menu_items( void *p, void *q){
    record_entry_t *en=p;
    =q;
    if (!is_mounted(en->path)) {
         SHOW_IT(view_p->widgets.paper, "mountP");
    }
    return GINT_TO_POINTER(1);
}
*/

/* use this function to hide menu elements of local paths 
 * when the path is provided by a module 
 * p is record entry and q is the name of the menu item*/

G_MODULE_EXPORT void *
hide_local_menu_items (void *p, void *q) {
    record_entry_t *en = p;
    widgets_t *widgets_p = q;
    view_t *view_p=widgets_p->view_p;
    if(!q)
        return NULL;

    // nothing selected:
    gchar *symbols_null[] = {
        "paste_menuitem",
        NULL
    };

    /*gchar *symbols[] = { "copy_menuitem", "cut_menuitem",
       "rename_menuitem", "duplicate_menuitem", "touch_menuitem",
       "symlink_menuitem", "remove_menuitem","autotype_Prun",
       "file_separator",    "ejecutar1",  "ejecutar2", "ejecutar3",
       "ejecutar4", "ejecutar5", "ejecutar6", "ejecutar7",
       "ejecutar8", "ejecutar9", "ejecutar10",
       "file_separator2", "navigation_separator",  "view_separator",
       "file_separator",
       NULL
       };  */
    gchar *symbols[] = {
        "cut_menuitem",
        "rename_menuitem",
        "autotype_Prun",
        "paste_menuitem",
        NULL
    };

    gchar *symbols_partition[] = {
        "open_with_menuitem",
        "copy_menuitem",
        "duplicate_menuitem",
        "symlink_menuitem",
        "touch_menuitem",
        "navigation_separator",
        "view_separator", "file_separator2",
        //"properties2",
        NULL
    };
    gchar *symbols_multiple_select[] = {
        "copy_menuitem",
        "remove_menuitem",
        "properties2", "find2",
        "open_with_menuitem",
        "ejecutar1", "ejecutar2", "ejecutar3",
        "ejecutar4", "ejecutar5", "ejecutar6", "ejecutar7",
        "ejecutar8", "ejecutar9", "ejecutar10",
        "file_separator2", "navigation_separator", "view_separator",
        "file_separator",

        NULL
    };
/*    gchar *symbols_mixed[] = { 
	"copy_menuitem",
	"remove_menuitem",
	NULL
    };*/

    gchar *symbols_dir[] = {
        "module1_menu",
        NULL
    };

    gchar **pp;
    if(!p) {
        for(pp = symbols_null; *pp; pp++) {
            HIDE_IT (widgets_p->paper, *pp);
        }
        return NULL;
    }

    for(pp = symbols; *pp; pp++) {
        HIDE_IT (widgets_p->paper, *pp);
    }
    for(pp = symbols_partition; *pp; pp++) {
        if(IS_PARTITION_TYPE (en->type)) {
            HIDE_IT (widgets_p->paper, *pp);
        }
    }
    for(pp = symbols_dir; *pp; pp++) {
        if(!IS_PARTITION_TYPE (en->type)) {
            HIDE_IT (widgets_p->paper, *pp);
        }
    }
    if(g_list_length (view_p->selection_list) > 1) {
        for(pp = symbols_multiple_select; *pp; pp++) {
            HIDE_IT (widgets_p->paper, *pp);
        }
    }
    return GINT_TO_POINTER (1);
}

static 
gchar *get_mnt_point(widgets_t *widgets_p, record_entry_t *en) {
#define MOUNT_DBH_FILE "rfm","mountdir-4.6.dbh"
        gchar *f = g_build_filename (g_get_user_data_dir (), MOUNT_DBH_FILE, NULL);
        gchar *device = g_path_get_basename (en->path);
        gchar *default_mount;
        if(g_file_test (en->tag, G_FILE_TEST_IS_DIR))
            default_mount = g_strdup (en->tag);
        else
            default_mount = g_build_filename (g_get_home_dir (), "mnt", device, NULL);
        g_free (device);
        const gchar *g = rodent_get_response_history (widgets_p, _("Mount Point"), _("Path"), f,
                                                      NULL,
                                                      default_mount,
                                                      NULL, NULL,
                                                      GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER,
                                                      "/");
        g_free (default_mount);
        g_free (f);
        if(g == NULL) {
            return NULL;
        }
        if(g_mkdir_with_parents (g, 0700) < 0) {
	    rfm_show_text(widgets_p);
            rfm_diagnostics (widgets_p, "xffm/error",  NULL);
            rfm_diagnostics (widgets_p, "xffm_tag/stderr", "mkdir(", g, "): ",strerror (errno), "\n", NULL);
            return NULL;
        }
	// is mount point already mounted?
	if (is_mounted((void *)g)){
	    rfm_show_text(widgets_p);
            rfm_diagnostics (widgets_p, "xffm/error", NULL);
            rfm_diagnostics (widgets_p, "xffm_tag/stderr", g, ": ", _("Failed to execute the mount command"), " (", _("address already in use"), ")", "\n", NULL);
            return NULL;
	}
        return g_strdup (g);

}


G_MODULE_EXPORT void *
fstab_mount (void *p, void *q) {
    gchar *argument[MAX_COMMAND_ARGS];
    gchar **ap;
    gchar *umount = "umount",
        *mount = "mount";
    record_entry_t *en = (record_entry_t *) p;
    gchar *mnt_src = NULL;
    gchar *mnt_tgt = NULL;
    widgets_t *widgets_p = (widgets_t *) q;

    SETWD();

    if(!widgets_p)
        return NULL;

    if(!en)
        return NULL;

    mnt_src = g_strdup (en->path);
    mnt_tgt = NULL;

    // user option check for mount point
    // this is necesary for mount/umount command outside of
    // the fstab module directory
    if(is_user_type (mnt_src)) {
        SET_USER_TYPE (en->type);
    }

    gchar *options = NULL;
    if(IS_PARTITION_TYPE (en->type) && is_mounted ((void *)mnt_src) == NULL && !IS_USER_TYPE (en->type)) {


	mnt_tgt=get_mnt_point(widgets_p, en);
	if (mnt_tgt == NULL) {
            g_free (mnt_src);
            return NULL;
	}
        options = get_options (mnt_src);
    } // end of partition type condition.

    // isofs:
    if ((en->mimetype && strstr(en->mimetype, "application/x-iso9660-image")) ||
	(en->mimemagic && strstr(en->mimemagic, "application/x-iso9660-image"))) {
	mnt_tgt=get_mnt_point(widgets_p, en);
	if (mnt_tgt == NULL) {
            g_free (mnt_src);
            return NULL;
	}
	options = g_strdup("loop");
    }
    DBG ("MOUNT: mnt_src=%s mnt_tgt=%s\n", mnt_src, mnt_tgt);

    ap = argument;

    // Sudo check...
    gboolean use_sudo = TRUE;
    // not for root
    if(!getuid ())
        use_sudo = FALSE;
    // not for general user mounts
    else if(IS_USER_TYPE (en->type))
        use_sudo = FALSE;
    // sudo not installed
    else {
        char *p = g_find_program_in_path ("sudo");
        if(p == NULL) {
            use_sudo = FALSE;
            rfm_diagnostics (widgets_p, "xffm/warning", strerror (ENOENT), ": sudo", "\n", NULL);
        } else {
            g_free (p);
        }
    }

    if(use_sudo) {
        *ap++ = "sudo";
        *ap++ = "-A";
    }

    if(is_mounted ((void *)mnt_src)) {
        *ap++ = umount;
    } else {
        *ap++ = mount;
        if(options) {
            *ap++ = "-o";
            *ap++ = options;
        }
    }

    *ap++ = (gchar *) mnt_src;
    if(mnt_tgt) {
        *ap++ = mnt_tgt;
    }
    *ap++ = NULL;
#if 10
    DBG ("MOUNT:");
    rfm_diagnostics (widgets_p, "emblem-urgent", NULL);
    int i;
    for(i = 0; argument[i]; i++) {
        DBG (" %s", argument[i]);
        if(strstr (argument[i], "username="))
            rfm_diagnostics (widgets_p, NULL, " username=*****", NULL);
        else
            rfm_diagnostics (widgets_p, NULL, " ", argument[i], NULL);
    }
    DBG ("\n");
    rfm_diagnostics (widgets_p, NULL, "\n", NULL);
#endif

    rfm_cursor_wait (widgets_p->window);
    rfm_thread_run_argv (widgets_p, argument, FALSE);
    rfm_cursor_reset (widgets_p->window);

    DBG ("fstab_mount %s done ", mnt_src);
    g_free (mnt_src);
    g_free (mnt_tgt);
    g_free (options);
    return GINT_TO_POINTER (1);
}

G_MODULE_EXPORT void *
double_click (void *p, void *q) {
    record_entry_t *en = (record_entry_t *) p;
    //widgets_t *widgets_p = (widgets_t *) q;
    if(IS_PARTITION_TYPE (en->type)) {
        fstab_mount (p, q);
        return GINT_TO_POINTER (1);
    }
    return NULL;
}

G_MODULE_EXPORT void *
module_data (void *p) {
    return ((void *)df ());
}

G_MODULE_EXPORT void *
module_xfdir_get (void *p) {
    xfdir_t *xfdir_p = p;
    private_get_xfdir (xfdir_p);
    return GINT_TO_POINTER (1);
}

G_MODULE_EXPORT void *
module_count (void *p) {
    int items = count_elements () + count_partitions ();
    return GINT_TO_POINTER (items);
}

G_MODULE_EXPORT void *
module_stat (void *p, void *q) {
    const gchar *path = p;
    // if p is NULL, then the function checks the top level stat file
    if(path == NULL) {
        path = "/etc/mtab";
    }
    struct stat st_v;
    struct stat *st = q;
    // if q is NULL, then the function just checks if the stat is possible.
    if(st == NULL) {
        st = &st_v;
    }
    if(!g_file_test (path, G_FILE_TEST_EXISTS))
        return NULL;
    if(stat (path, st) < 0) {
        g_warning ("stat(%s): %s", path, strerror (errno));
        return NULL;
    }
    return GINT_TO_POINTER (1);
}

G_MODULE_EXPORT void *
is_partition_type (void *p) {
    record_entry_t *en = p;
    if(!en)
        return NULL;
    if(IS_PARTITION_TYPE (en->type))
        return GINT_TO_POINTER (1);
    return NULL;

}
