
/*
 *  icontheme -   Gtkdoc comments in .h file.
 *
 *  Copyright (c) 2004-2010 Edscott Wilson Garcia <edscott@xfce.org>
 *
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <sys/types.h>
#include <sys/wait.h>

#include "rfm_libs.h"



G_MODULE_EXPORT gboolean
svg_supported (void) {
    static int support = 2;
    if(support < 2)
        return (gboolean) support;
    else
        support = 0;
    GSList *l;
    GSList *pix_formats = NULL;
    if(!pix_formats)
        pix_formats = gdk_pixbuf_get_formats ();
    for(l = pix_formats; l; l = l->next) {
        gchar **pix_mimetypes;
        int i;
        GdkPixbufFormat *fmt = l->data;
        pix_mimetypes = gdk_pixbuf_format_get_mime_types (fmt);
        for(i = 0; pix_mimetypes[i]; i++) {
            if(g_ascii_strcasecmp (pix_mimetypes[i], "image/svg") == 0) {
                support = 1;
                break;
            }
        }
        g_strfreev (pix_mimetypes);
    }
    g_slist_free (pix_formats);
    return (gboolean) support;
}

/*cache stuff*/
static int
check_dir (char *path) {
    struct stat st;
    if(stat (path, &st) < 0) {
        if(mkdir (path, 0770) < 0)
            return FALSE;
        return TRUE;
    }
    if(S_ISDIR (st.st_mode)) {
        if(access (path, W_OK) < 0)
            return FALSE;
        return TRUE;
    }
    return FALSE;
}

static gchar *
get_cache_path (void
    ) {

    gchar *cache_path = NULL;
    gchar *cache_dir;
    cache_dir = g_build_filename (g_get_user_cache_dir (), CACHE_DIR, NULL);

    if(!check_dir (cache_dir)) {
        g_free (cache_dir);
        return NULL;
    }
    cache_path = g_strdup_printf ("%s%c%s.%s.cache64.dbh",
                                  cache_dir, G_DIR_SEPARATOR, theme_name, (svg_supported ())? "svg" : "png");
    g_free (cache_dir);
    TRACE ("ICON: using cache: %s\n", cache_path);
    return cache_path;
}

static long long
recurse_basedir_sum (gchar * dir) {
    long long sum = 0;
    const char *file;
    GDir *gdir;
    TRACE ("ICON: recurse_basedir_sum(%s)\n", dir);
    if((gdir = g_dir_open (dir, 0, NULL)) != NULL) {
        while((file = g_dir_read_name (gdir))) {
            gchar *path = g_build_filename (dir, file, NULL);
            if(g_file_test (path, G_FILE_TEST_IS_SYMLINK)) {
                g_free (path);
                continue;
            }
            if(g_file_test (path, G_FILE_TEST_IS_DIR)) {
                struct stat st;
                /* recurse */
                sum += recurse_basedir_sum (path);
                if(stat (path, &st) == 0) {
                    sum += st.st_mtime;
                    sum += st.st_dev;
                }
            }
            g_free (path);
        }
        g_dir_close (gdir);
    }
    return sum;
}

static gchar *
get_base_dir (void
    ) {
    gchar *basedir = g_build_filename (PACKAGE_DATA_DIR, "icons", theme_name, NULL);
    return basedir;
}

/* must get info for all directories within */

static long long
get_basedir_sum (void
    ) {
    long long basedir_sum = 0;
    gchar *basedir = get_base_dir ();

    struct stat st;
    if(stat (basedir, &st) == 0) {
        TRACE ("ICON: stat ok\n");
        basedir_sum += st.st_mtime;
        basedir_sum += st.st_dev;
    }
    basedir_sum += recurse_basedir_sum (basedir);
    DBG ("ICON-DBH: basedir summing %s (%ld %ld)\n", basedir, (long)st.st_mtime, (long)st.st_dev);

    basedir_sum += SERIAL;
    g_free (basedir);
    return basedir_sum;
}

static const gchar *
get_supported_regex (void
    ) {
    GSList *pix_formats,
     *l;
    gchar **pix_extensions,
    **p;
    static gchar *reg = NULL,
        *r = NULL;
    /* check SVG format support */
    if((pix_formats = gdk_pixbuf_get_formats ()) != NULL) {
        for(l = pix_formats; l; l = l->next) {
            GdkPixbufFormat *fmt = l->data;
            pix_extensions = gdk_pixbuf_format_get_extensions (fmt);
            for(p = pix_extensions; *p; p++) {
                TRACE ("ICON: supported=%s\n", *p);
                if(reg) {
                    g_free (r);
                    r = reg;
                    reg = g_strconcat (r, "|", *p, NULL);
                } else
                    reg = g_strdup (*p);
            }
            g_strfreev (pix_extensions);
        }
        g_slist_free (pix_formats);
    }
    if(!reg)
        return "\\.(png|xpm)$)";
    g_free (r);
    r = g_strconcat ("\\.(", reg, ")$", NULL);
    g_free (reg);
    reg = NULL;

    TRACE ("ICON: regex=%s\n", r);
    return (const gchar *)r;
}

static void
inner_save_cache_info (gchar * version) {
    FILE *info;

    gchar *cache_file = get_cache_path ();
    gchar *cache_info_path = g_strconcat (cache_file, ".info", NULL);
    cache_info_t cache_info;
    if(!cache_file)
        return;
    cache_info_path = g_strconcat (cache_file, ".info", NULL);
    TRACE ("ICON: saving %s\n", cache_info_path);

    cache_info.basedir_sum = get_basedir_sum ();
    strncpy (cache_info.version, version, 64);
    TRACE ("ICON: sving cache_info.version=%s\n", version);
    strncpy (cache_info.supported_regex, get_supported_regex (), 255);
    cache_info.supported_regex[255] = 0;
    info = fopen (cache_info_path, "wb");
    if(info) {
        if(fwrite (&cache_info, sizeof (cache_info_t), 1, info) < 1) {
            g_warning ("cannot write to %s", cache_info_path);
        }
        fclose (info);
    } else
        g_warning ("cannot write to %s", cache_info_path);
    TRACE ("ICON: wrote basedir sum is %lld\n", cache_info.basedir_sum);
    g_free (cache_info_path);
    g_free (cache_file);
}

static void
save_cache_info () {
    inner_save_cache_info (BUILD);
}

static void
unset_cache_info () {
    inner_save_cache_info ("NONE");
}

static gboolean
compare_cache_version () {
    FILE *info;
    gchar *cache_info_path;
    cache_info_t disk_cache_info;
    size_t records;

    gchar *cache_file = get_cache_path ();
    DBG ("DBH-ICON: (%s) comparing cache version for %s \n", DBH_FILE_VERSION, theme_name);

    DBHashTable *test = dbh_open (cache_file);
    if(!test) {
        DBG ("DBH-ICON: cannot open %s\n", cache_file);
        g_free (cache_file);
        return FALSE;
    } else {
        DBG ("DBH-ICON: found version: %s\n", test->head_info->version);
        dbh_close (test);
    }

    cache_info_path = g_strconcat (cache_file, ".info", NULL);
    if((info = fopen (cache_info_path, "rb")) == NULL) {
        TRACE ("ICON: %s not exists\n", cache_info_path);
        g_free (cache_info_path);
        g_free (cache_file);
        return FALSE;
    }
    g_free (cache_file);

    records = fread (&disk_cache_info, sizeof (cache_info_t), 1, info);
    fclose (info);
    if(records < 1) {
        TRACE ("ICON: invalid info file: %s\n", cache_info_path);
        unlink (cache_info_path);
        g_free (cache_info_path);
        return FALSE;
    }
    g_free (cache_info_path);
    DBG ("DBH-ICON: disk_cache_info.version=%s\n", disk_cache_info.version);
    if(strcmp (disk_cache_info.version, BUILD) != 0) {
        return FALSE;
    }
    return TRUE;
}

/* criteria for cache regeneration:
 * 1- does not exist
 * 2- one of the base_dirs has been modified
 * 3- supported regexp has changed.
 *
 * This info will be kept in file "theme.cache.info"
 * */
static gboolean
compare_cache_info () {
    FILE *info;
    gchar *cache_info_path;
    cache_info_t cache_info,
      disk_cache_info;
    int records;

    gchar *cache_file = get_cache_path ();
    DBG ("ICON: comparing cache info for %s \n", theme_name);

    if(!g_file_test (cache_file, G_FILE_TEST_EXISTS)) {
        TRACE ("ICON: %s not exists\n", cache_file);
        g_free (cache_file);
        return FALSE;
    }

    cache_info_path = g_strconcat (cache_file, ".info", NULL);
    g_free (cache_file);
    if((info = fopen (cache_info_path, "rb")) == NULL) {
        TRACE ("ICON: %s not exists\n", cache_info_path);
        g_free (cache_info_path);
        return FALSE;
    }
    g_free (cache_info_path);

    records = fread (&disk_cache_info, sizeof (cache_info_t), 1, info);
    fclose (info);
    if(records < 1)
        return FALSE;

    DBG ("ICON: read basedir cache sum is %lld\n", disk_cache_info.basedir_sum);
    if(strcmp (disk_cache_info.version, BUILD) != 0) {
        TRACE ("ICON: %s != %s\n", disk_cache_info.version, BUILD);
        return FALSE;
    }

    cache_info.basedir_sum = get_basedir_sum ();
    DBG ("ICON: cache sum is %lld\n", cache_info.basedir_sum);
    if(cache_info.basedir_sum != disk_cache_info.basedir_sum) {
        TRACE
            ("ICON: cache_info.basedir_sum(%lld) != disk_cache_info.basedir_sum(%lld)\n",
             cache_info.basedir_sum, disk_cache_info.basedir_sum);
        return FALSE;
    }

    strncpy (cache_info.supported_regex, get_supported_regex (), 255);
    cache_info.supported_regex[255] = 0;
    if(strcmp (cache_info.supported_regex, disk_cache_info.supported_regex)) {
        TRACE ("ICON: cache_info.supported_regex != disk_cache_info.supported_regex\n");
        return FALSE;
    }
    return TRUE;
}

/*
 * open_theme()
 * 
 *  */

static gchar *
theme_index (gchar * base_dir) {
    gchar *index;
    //chop(base_dir);
    //index=g_build_filename(base_dir, theme_name, NULL);
    DBG ("ICON-DBH: index dir: %s\n", base_dir);
    if(!g_file_test (base_dir, G_FILE_TEST_IS_DIR)) {
        return NULL;
    }
    index = g_build_filename (base_dir, "index.theme", NULL);
    DBG ("ICON-DBH: index file %s\n", index);
    if(!g_file_test (index, G_FILE_TEST_EXISTS)) {
        TRACE ("ICON: %s does not exist\n", index);
        g_free (index);
        return NULL;
    }
    return index;
}

static const gchar **
parse_value (gchar * value) {
    if(!value || !strlen (value))
        return NULL;
    DBG ("ICONi: parsing: %s\n", value);
    int len = strlen (value);
    int count = 0;
    int i;
    for(i = 0; i < len; i++) {
        if(value[i] == ',')
            count++;
    }
    gchar **info;
    info = (gchar **) malloc ((count + 2) * sizeof (gchar *));
    memset (info, 0, (count + 2) * sizeof (gchar *));
    gchar *p;
    i = 0;
    info[i++] = p = value;
    while(strchr (p, ',')) {
        info[i++] = p = strchr (p, ',') + 1;
    }
    info[i] = NULL;
    for(i = 0; i < len; i++) {
        if(value[i] == ',')
            value[i] = 0;
    }
    for(i = 0; i < count + 1; i++) {
        DBG ("ICONi: --parsed[%d]=\"%s\"\n", i, info[i]);
    }
    return (const gchar **)info;
}

static gchar **
get_rc_info (gchar * index, gchar * rc_info) {
    GKeyFile *key_file;
    GError *error = NULL;
    gchar *value;
    const gchar **info;
    DBG ("ICONi: get_rc_info\n");
    key_file = g_key_file_new ();
    if(!g_key_file_load_from_file (key_file, index, 0, &error)) {
        g_warning ("%s (%s)", error->message, index);
        g_error_free (error);
        return NULL;
    }
    value = g_key_file_get_string (key_file, "Icon Theme", rc_info, &error);
    if(error) {
        g_warning ("%s (%s)", error->message, index);
        g_error_free (error);
    }
    g_key_file_free (key_file);
    DBG ("ICONi: value=%s\n", value);
    info = parse_value (value);
    return (gchar **) info;
}

static gboolean
read_icon_directory (const gchar * directory) {
    GDir *dir;
    static regex_t supported;
    static gboolean regex_compiled = FALSE;
    const gchar *name;
    gchar *key;

    if(!regex_compiled) {
        const gchar *regex = get_supported_regex ();
        if(regcomp (&supported, regex, REG_EXTENDED | REG_ICASE | REG_NOSUB) == 0)
            regex_compiled = TRUE;
        else
            regex_compiled = FALSE;
    }

    dir = g_dir_open (directory, 0, NULL);
    DBG ("ICON-DBH: reading %s...(%s)\n", directory, ((dir) ? "yes" : "no"));
    if(!dir)
        return FALSE;

    DBHashTable *cache = NULL;
    gchar *cache_file = get_cache_path ();
    DBG ("ICON-DBH: cache--> %s\n", cache_file);
    if((cache = dbh_open ((char *)cache_file)) == NULL) {
        g_warning ("cannot open %s for writing", cache_file);
        g_free (cache_file);
        return FALSE;
    }
    g_free (cache_file);
    while((name = g_dir_read_name (dir))) {
        /*check for supported types */
        if(regex_compiled && regexec (&supported, name, 0, NULL, 0)) {
            DBG ("ICON-DBH: %s not supported image type\n", name);
            continue;
        }
        key = g_strdup (name);
        // zap file extension
        if(strchr (key, '.'))
            *strrchr (key, '.') = 0;

        /* put in DBH cache  (if not there already) */
        GString *gs = g_string_new (key);
        memset (DBH_KEY (cache), 0, DBH_KEYLENGTH (cache));
        sprintf ((char *)DBH_KEY (cache), "%10u", g_string_hash (gs));
        g_string_free (gs, TRUE);
        TRACE ("ICON-DBH: add2cache, %s --> %s\n", (char *)DBH_KEY (cache), key);
        if(dbh_load (cache)) {
            TRACE ("ICON-DBH: add2cache, %s --> %s already in cache!\n", (char *)DBH_KEY (cache), key);
        } else {
            gchar *icon_path = g_build_filename (directory, name, NULL);
            strcpy ((gchar *) cache->data, icon_path);
            dbh_set_recordsize (cache, strlen (icon_path) + 1);
            if(!dbh_update (cache)) {
                TRACE ("ICON-DBH: !dbh_update failed!\n");
            }

        }

    }                           /* end while read dir */
    g_dir_close (dir);
    dbh_close (cache);
    return TRUE;
}

static gboolean
add_theme_directories (gchar * basedir) {
    TRACE ("ICON-DBH:  add_theme_directories() %s %s\n", basedir, theme_name);

    gchar *index = theme_index (basedir);
    if(!index) {
        g_warning ("ICON: no index  at %s", index);
        return FALSE;
    }

    gchar **directories;
    if((directories = get_rc_info (index, "Directories")) == NULL) {
        g_warning ("ICON: no directories field in index %s", index);
        g_free (index);
        return FALSE;
    }

    GError *error = NULL;
    GKeyFile *key_file = g_key_file_new ();
    if(!g_key_file_load_from_file (key_file, index, 0, &error)) {
        g_warning ("!g_key_file_load_from_file(%s): %s", index, error->message);
        g_error_free (error);
        return FALSE;
    }

    gchar **group;

    for(group = directories; *group; group++) {
        gchar *directory;

        directory = g_build_filename (basedir, *group, NULL);
        if(strstr (directory, "scalable") && svg_supported ()) {
            read_icon_directory (directory);
        }
        if(strstr (directory, "48x48")) {
            read_icon_directory (directory);
        }
        g_free (directory);
    }                           /* foreach directory */
    g_free (directories[0]);
    g_free (directories);

    g_free (index);
    return TRUE;
}

static int
open_theme (const gchar * theme) {
    if(!theme) {
        g_warning ("no icon theme defined");
        return -1;
    }
    DBG ("ICON: requested_theme is %s\n", theme);
    g_free (theme_name);
    theme_name = g_strdup (theme);
    {
        gchar *cache_file = get_cache_path ();
#ifndef DISABLE_OLD_CACHE
        gboolean version_match = compare_cache_version ();
        DBG ("ICON-DBH: cache version match=%d for %s\n", version_match, cache_file);
        if(version_match && cache_file && g_file_test (cache_file, G_FILE_TEST_EXISTS)) {
            /* checking whether the cache needs to be updated is time consuming
             * on a cold filesystem. So what we do is check whether cache needs
             * to be regenerated, and tag it so that on the next time it is 
             * called for, it will be regenerated. Thus we avoid the time wait
             * to check the cache. 
             */
            gchar *c_lock = g_build_filename (CACHE_CHECK_LOCK, NULL);
            DBG ("ICON-DBH: cache  exists...\n");
            if(!g_file_test (c_lock, G_FILE_TEST_EXISTS)) {
                if(fork () == 0) {
                    /* give child a head start */
                    fclose (fopen (c_lock, "w"));
                    sleep (1);
                    DBG ("ICON-DBH: BACKGROUND: comparing cache information for %s.\n", theme);
                    if(!compare_cache_info ()) {
                        DBG ("ICON-DBH: BACKGROUND: cache does not compare.\n");
                        unset_cache_info ();
                        unlink (c_lock);
                        DBG ("Icon cache is out of date. Will be updated on next restart");
                        //kill(getppid(),SIGUSR2);
                    } else {
                        DBG ("ICON-DBH: BACKGROUND: information for %s- is OK!.\n", theme);
                    }
                    exit (123);
                }
            } else {
                DBG ("ICON-DBH: c_lock=%s exists\n", c_lock);
            }
            g_free (c_lock);

        } else
#endif
        {
            /* create empty hash file */
            DBG ("ICON-DBH: create empty hash file\n");
            DBHashTable *cache = NULL;
            gchar *cache_file = get_cache_path ();
            DBG ("Creating cache file: %s", cache_file);
            if((cache = dbh_create ((char *)cache_file, 11)) == NULL) {
                unlink (cache_file);
                if((cache = dbh_create ((char *)cache_file, 11)) == NULL) {
                    g_warning ("cannot create hash file: %s", cache_file);
                    return FALSE;
                }
            }
            g_free (cache_file);
            dbh_close (cache);
            /* empty hash file created */
            gchar *basedir = NULL;
            basedir = get_base_dir ();
            DBG ("ICON-DBH: cache does not exist.\n");
            DBG ("Generating icon-module cache. This may take a bit...\n");
            add_theme_directories (basedir);
            save_cache_info ();

            DBG ("Icon-module cache complete!");
            g_free (basedir);
        }
        g_free (cache_file);
        /////////////////////
    }
    return 1;
}

static gchar *
mime_icon_find_pixmap_file (const gchar * key) {
    if (!key) return NULL;
    TRACE ("MIME ICON: icon lookup: %s\n", key);
    gchar *cache_file = get_cache_path ();
    if(!cache_file) {
        g_warning ("ICON: unable to get cache filename for %s", theme_name);
        return NULL;
    }
    /* shortcircuit with cache  */
    GString *gs;
    gchar *cache_hit = NULL;
    DBHashTable *cache = NULL;

    if((cache = dbh_open ((char *)cache_file)) == NULL) {
        g_warning ("cannot open cache file %s", cache_file);
        return NULL;
    } else {
        gs = g_string_new (key);
        sprintf ((char *)DBH_KEY (cache), "%10u", g_string_hash (gs));
        g_string_free (gs, TRUE);

        if(!dbh_load (cache)) {
            TRACE ("MIME ICON: icon not found: %s (%s)\n", key, (char *)DBH_KEY (cache));
            dbh_close (cache);
            return NULL;
        }
        cache_hit = g_strdup ((gchar *) cache->data);
        TRACE ("MIME ICON: cache hit: %s -> %s\n", key, cache_hit);
        dbh_close (cache);
        return cache_hit;
    }
}


