/*
 * Copyright (c) 2006-2011 Edscott Wilson Garcia <edscott@xfce.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA. 
 */

static GtkWidget *desktopcolor_button;
static GtkWidget *iconviewcolor_button;
static GtkWidget *desktopimage_button;
static GtkWidget *desktopdir_button;
static GtkWidget *combo_box;
static GtkWidget *terminal_box;
static GtkWidget *editor_box;
static gboolean disable_options = FALSE;
static GtkListStore *model;
static McsPlugin *mp = NULL;
#define SETTINGS_TIMERVAL 500
static int shm_settings_serial = 0;
static int settings_timer = 0;

static
  gboolean
transparency_changed (GtkRange * range, GtkScrollType scroll, gdouble value, gpointer user_data) {
    if(disable_options)
        return FALSE;
    int i = (int)((long)user_data);
    if(rfm_options[i].value)
        g_free (rfm_options[i].value);
    rfm_options[i].value = g_strdup_printf ("%lf", value);
    mcs_manager_set_string (mcs_manager, rfm_options[i].name, CHANNEL, rfm_options[i].value);
    mcs_manager_notify (mcs_manager, CHANNEL);

    return FALSE;
}

static gint
settings_monitor (gpointer data) {
    mcs_shm_t *mcs_shm_p = (mcs_shm_t *) mp->m;
    if(shm_settings_serial < 0)
        return FALSE;
    if(shm_settings_serial != mcs_shm_p->serial) {
        int i;
        DBG ("shm_settings_serial != mcs_shm_p->serial(%d!=%d)\n", shm_settings_serial, mcs_shm_p->serial);
        shm_settings_serial = mcs_shm_p->serial;
        for(i = 0; i < RFM_OPTIONS; i++) {
            if(mcs_shm_p->data[i].value) {
                g_free (rfm_options[i].value);
                rfm_options[i].value = g_strdup (mcs_shm_p->data[i].value);
                rfm_setenv (rfm_options[i].name, rfm_options[i].value, TRUE);
            } else {
                rfm_setenv (rfm_options[i].name, NULL, TRUE);
            }
        }
    } else {
        /*TRACE("shm_settings_serial OK\n"); */
    }
    return TRUE;
}

static int
mcs_shm_fileread (void
    ) {
    static gchar *rcfile = NULL;

    if(!rcfile) {
        rcfile = g_build_filename (MCS_SHM_PLUGIN_FILE, NULL);
    }

    if(g_file_test (rcfile, G_FILE_TEST_EXISTS)) {
        FILE *f;
        TRACE ("reading %s\n", rcfile);
        f = fopen (rcfile, "r");
        if(f) {
            if(fread (mp->m, sizeof (mcs_shm_t), 1, f) == 0)
                g_warning ("fread: %s", strerror (errno));
            fclose (f);
        }
        msync (mp->m, sizeof (mcs_shm_t), MS_SYNC);
        return 1;
    }
    return 0;
}

static int
mcs_shm_filewrite (void
    ) {
    FILE *f;
    static gchar *rcfile = NULL;

    if(!rcfile) {
        rcfile = g_build_filename (MCS_SHM_PLUGIN_FILE, NULL);
    }
    TRACE ("writing %s\n", rcfile);
    msync (mp->m, sizeof (mcs_shm_t), MS_SYNC);
    f = fopen (rcfile, "w");
    if(f) {
        mcs_shm_t *mcs_shm_p = (mcs_shm_t *) mp->m;
        mcs_shm_p->serial++;
        if(fwrite (mp->m, sizeof (mcs_shm_t), 1, f) < 1)
            g_warning ("fwrite: %s", strerror (errno));
        fclose (f);
        return 1;
    }
    msync (mp->m, sizeof (mcs_shm_t), MS_SYNC);
    return 0;
}

static void
mcs_shm_bringforth (void
    ) {
    int i,
      j;
    mcs_shm_t *mcs_shm_p = (mcs_shm_t *) mp->m;
    TRACE ("mcs_shm_bringforth\n");
    for(i = 0; i < RFM_OPTIONS; i++) {
        g_free (rfm_options[i].value);
        rfm_options[i].value = g_strdup ("");
    }
    for(i = 0; i < RFM_OPTIONS; i++)
        for(j = 0; j < RFM_OPTIONS; j++) {
            if(rfm_options[i].name && mcs_shm_p->data[j].name) {
                if(strcmp (rfm_options[i].name, mcs_shm_p->data[j].name))
                    continue;
                g_free (rfm_options[i].value);
                rfm_options[i].value = g_strdup (mcs_shm_p->data[j].value);
                break;
            }
        }
}

#if 0
static void
mcs_shm_sendback (void
    ) {
    int i;
    mcs_shm_t *mcs_shm_p = (mcs_shm_t *) mp->m;
    TRACE ("mcs_shm_sendback\n");
    for(i = 0; i < RFM_OPTIONS; i++) {
        memset (mcs_shm_p->data[i].name, 0, SHM_STRING_LENGTH);
        memset (mcs_shm_p->data[i].value, 0, SHM_VALUE_LENGTH);
        strncpy (mcs_shm_p->data[i].name, rfm_options[i].name, SHM_STRING_LENGTH);
        if(rfm_options[i].value)
            strncpy (mcs_shm_p->data[i].value, rfm_options[i].value, SHM_VALUE_LENGTH);
    }
    msync (mp->m, sizeof (mcs_shm_t), MS_SYNC);
}
#endif

static void
mcs_shm_init (void
    ) {
    int i;
    mcs_shm_t *mcs_shm_p = (mcs_shm_t *) mp->m;
    TRACE ("mcs_shm_init\n");
    for(i = 0; i < RFM_OPTIONS; i++) {
        memset (mcs_shm_p->data[i].name, 0, SHM_STRING_LENGTH);
        memset (mcs_shm_p->data[i].value, 0, SHM_VALUE_LENGTH);
        strncpy (mcs_shm_p->data[i].name, rfm_options[i].name, SHM_STRING_LENGTH);
        if(rfm_options[i].value) {
            strncpy (mcs_shm_p->data[i].value, rfm_options[i].value, SHM_VALUE_LENGTH);
            mcs_shm_p->data[i].value[SHM_VALUE_LENGTH - 1] = 0;
        }
        DBG ("*SETTINGS: initialising %s->%s\n", rfm_options[i].name, rfm_options[i].value);
    }
    msync (mp->m, sizeof (mcs_shm_t), MS_SYNC);
}

#if 0
static gboolean
mcs_manager_save_channel_to_file (McsManager * mcs_m, const gchar * channel, const gchar * rcfile) {
    TRACE ("mcs_manager_save_channel_to_file");
    /* here we dump the shmblock to a file */
    mcs_shm_sendback ();
    mcs_shm_filewrite ();
    return TRUE;
}
#endif

static void
start_desktop (gboolean on) {
    Window xid;
    GError *error = NULL;
    if(on) {
        Atom selection_atom = XInternAtom (GDK_DISPLAY (), "RODENT_DESK_ATOM", False);
        //TRACE("enabling desktop=%d\n",enable_desktop);
        if((xid = XGetSelectionOwner (GDK_DISPLAY (), selection_atom))) {
            TRACE ("rodent-desk already running\n");
        } else if(!g_spawn_command_line_async ("rodent-desk", &error)) {
            g_warning ("%s", error->message);
        }
    } else {
        if(!g_spawn_command_line_async ("killall rodent-desk", &error)) {
            g_warning ("%s", error->message);
            g_error_free (error);
        }
    }

}

static void
mcs_manager_set_string (McsManager * mcs_m, const gchar * setting_name, const gchar * channel, const gchar * setting_value) {
    mcs_shm_t *mcs_shm_p = (mcs_shm_t *) mp->m;
    int i;
    DBG ("mcs_manager_set_string\n");

    for(i = 0; i < RFM_OPTIONS; i++) {
        DBG ("%d/%d %s==%s?\n", i+1, RFM_OPTIONS, setting_name, mcs_shm_p->data[i].name);
        if(strcmp (setting_name, mcs_shm_p->data[i].name) == 0) {
            memset (mcs_shm_p->data[i].value, 0, SHM_VALUE_LENGTH);
            if(setting_value) {
                DBG ("%s -> %s (%zd)", setting_name, setting_value, strlen (setting_value));
                if(i == RFM_ENABLE_DESKTOP) {
                    if(strlen (setting_value) > 0) {
                        TRACE ("Starting desktop (%ld)\n", (long)strlen (setting_value));
                        start_desktop (TRUE);
                    } else {
                        TRACE ("Terminating desktop (%ld)\n", (long)strlen (setting_value));
                        start_desktop (FALSE);
                    }
                }

                strncpy (mcs_shm_p->data[i].value, setting_value, SHM_VALUE_LENGTH);
                mcs_shm_p->data[i].value[SHM_VALUE_LENGTH - 1] = 0;
                g_free (rfm_options[i].value);
                rfm_options[i].value = g_strdup (setting_value);
            }
        }
    }
    msync (mp->m, sizeof (mcs_shm_t), MS_SYNC);
    /* here we update the shmblock */
    return;
}

static void
mcs_manager_notify (McsManager * mcs_m, const gchar * channel) {
    /* here we ++ the shmblock serial (or something), this should be
     * monitored by other instances so that changes are ACK.*/
    TRACE ("mcs_manager_notify\n");
    mcs_shm_t *mcs_shm_p = (mcs_shm_t *) mp->m;
    mcs_shm_p->serial++;
    msync (mp->m, sizeof (mcs_shm_t), MS_SYNC);
    mcs_shm_filewrite ();
    return;
}

/* the dialog ******************************************/
static void
add_spacer (GtkBox * box) {
    GtkWidget *eventbox = gtk_alignment_new (0, 0, 0, 0);

    gtk_widget_set_size_request (eventbox, SKIP, SKIP);
    gtk_widget_show (eventbox);
    gtk_box_pack_start (box, eventbox, FALSE, TRUE, 0);
}

static void update_tree (gint id);
static gboolean
environment_changed (GtkCellRendererText * cell, const gchar * path_string, const gchar * new_text, gpointer data) {
    if(disable_options)
        return FALSE;

    GtkTreeModel *model = (GtkTreeModel *) data;
    GtkTreePath *path = gtk_tree_path_new_from_string (path_string);
    GtkTreeIter iter;

    gint *column;

    column = g_object_get_data (G_OBJECT (cell), "column");

    gtk_tree_model_get_iter (model, &iter, path);

    if(GPOINTER_TO_INT (column) == COLUMN_VALUE) {
        TRACE ("column=%d path=%s newtext=%s", GPOINTER_TO_INT (column), path_string, new_text);

        gint i;
        gchar *old_text;

        gtk_tree_model_get (model, &iter, column, &old_text, -1);
        g_free (old_text);
        old_text = NULL;
        i = gtk_tree_path_get_indices (path)[0];

        if(new_text && strlen (new_text) >= SHM_VALUE_LENGTH) {
            g_warning ("strlen($%s) <  %d not met. Ignoring it.", rfm_options[i].name, SHM_VALUE_LENGTH);
            gtk_list_store_set (GTK_LIST_STORE (model), &iter, column, rfm_options[i].value, -1);
            return FALSE;
        }
        /* test for valid commands */
        if(i == TERMINAL_CMD || i == EDITOR_CMD) {
            gboolean dump = FALSE;
            if(!new_text)
                dump = TRUE;
            else {
                gchar *cmd = g_find_program_in_path (new_text);
                if(!cmd)
                    dump = TRUE;
                g_free (cmd);
            }
            if(dump) {
                gtk_list_store_set (GTK_LIST_STORE (model), &iter, column, rfm_options[i].value, -1);
                return FALSE;
            }
        }

        if(rfm_options[i].value) {
            g_free (rfm_options[i].value);
            rfm_options[i].value = NULL;
        }
        if(new_text && strlen (new_text)) {
            rfm_options[i].value = g_strdup (new_text);
        } else {
            rfm_options[i].value = g_strdup ("");
        }
        gtk_list_store_set (GTK_LIST_STORE (model), &iter, column, rfm_options[i].value, -1);
        mcs_manager_set_string (mcs_manager, rfm_options[i].name, CHANNEL, rfm_options[i].value);
        mcs_manager_notify (mcs_manager, CHANNEL);
        /*write_options(); */
    }
    gtk_tree_path_free (path);

    return FALSE;
}

/* the dialog */
static void
dialog_delete (GtkWidget * dialog, gpointer data) {
    GError *error = NULL;
    if((int)((long)data) == 1) {
	// confirm
        g_spawn_command_line_async ("rodent-mime", &error);
    } else if((int)((long)data) == 2) {
        /* directory cache */
	gchar *t=g_strdup_printf("%s: %s",_("Restart"), rfm_global_p->argv[0]);
	if (rfm_confirm (NULL, GTK_MESSAGE_QUESTION, t, _("Cancel"), _("Yes, restart"))) {
          gchar *cache_dir = g_build_filename (g_get_user_cache_dir (), "rfm", "cache", NULL);
          DBG ("clearing directory cache: %s", cache_dir);
          gchar *command = g_strdup_printf ("rm -rf \"%s\"", cache_dir);
          g_spawn_command_line_async (command, &error);
          g_free (cache_dir);
          g_free (command);
          /* thumbnail cache */
          cache_dir = g_build_filename (g_get_user_cache_dir (), "rfm", "thumbnails", NULL);
          DBG ("clearing image cache: %s", cache_dir);
          command = g_strdup_printf ("rm -rf \"%s\"", cache_dir);
          g_spawn_command_line_async (command, &error);
          g_free (cache_dir);
          g_free (command);

          /* pixmap cache */
          cache_dir = g_build_filename (g_get_user_cache_dir (), "rfm", "modules", NULL);
          DBG ("clearing dbh caches: %s", cache_dir);
          command = g_strdup_printf ("rm -rf \"%s\"", cache_dir);
          g_spawn_command_line_async (command, &error);
          g_free (cache_dir);
          g_free (command);
	  execvp(rfm_global_p->argv[0],rfm_global_p->argv);
	}
	g_free(t);
    }
    gtk_widget_destroy (dialog);
    dialog = NULL;
}

static void
update_combo (GtkComboBox * combo_box, gint id) {
    gchar *g = gtk_combo_box_get_active_text (combo_box);
    if(rfm_options[id].value && strlen (rfm_options[id].value)) {
        if(g && strcmp (g, rfm_options[id].value)) {
            gtk_combo_box_insert_text (combo_box, 0, rfm_options[id].value);
        }
    }
//    else { gtk_combo_box_insert_text(combo_box,0,""); }
    gtk_combo_box_set_active (combo_box, 0);
    g_free (g);
}

static void
update_tree (gint id) {
    GtkTreePath *treepath = gtk_tree_path_new_from_indices (id, -1);
    GtkTreeIter iter;
    if(gtk_tree_model_get_iter ((GtkTreeModel *) model, &iter, treepath)) {
        gtk_list_store_set (model, &iter, COLUMN_VALUE, g_strdup (rfm_options[id].value), COLUMN_EDITABLE, TRUE, -1);
    }
    gtk_tree_path_free (treepath);
}

GtkWidget *toggle_button[RFM_OPTIONS];
static void
set_option_buttons (void
    ) {
    int i;
    disable_options = TRUE;

    for(i = 0; i < RFM_OPTIONS; i++)
        if(toggle_button[i]) {
            if(rfm_options[i].value && strlen (rfm_options[i].value)) {
                gtk_toggle_button_set_active ((GtkToggleButton *)
                                              toggle_button[i], TRUE);
            } else {
                gtk_toggle_button_set_active ((GtkToggleButton *)
                                              toggle_button[i], FALSE);
            }
            GtkTreePath *treepath = gtk_tree_path_new_from_indices (i, -1);
            GtkTreeIter iter;
            if(gtk_tree_model_get_iter ((GtkTreeModel *) model, &iter, treepath)) {
                gtk_list_store_set (model, &iter, COLUMN_VALUE, g_strdup (rfm_options[i].value), COLUMN_EDITABLE, TRUE, -1);
            }
            gtk_tree_path_free (treepath);
        }

    int index = 0;
    if(strcmp (rfm_options[VERSION_CONTROL].value, "numbered") == 0)
        index = 1;
    else if(strcmp (rfm_options[VERSION_CONTROL].value, "existing") == 0)
        index = 2;
    else if(strcmp (rfm_options[VERSION_CONTROL].value, "none") == 0)
        index = 3;
    else if(strcmp (rfm_options[VERSION_CONTROL].value, "simple")) {
        g_free (rfm_options[VERSION_CONTROL].value);
        rfm_options[VERSION_CONTROL].value = g_strdup ("simple");
    }
    gtk_combo_box_set_active ((GtkComboBox *) combo_box, index);

    update_combo ((GtkComboBox *) terminal_box, TERMINAL_CMD);
    update_combo ((GtkComboBox *) editor_box, EDITOR_CMD);
    if(getenv ("RFM_DESKTOP_DIR") && strlen (getenv ("RFM_DESKTOP_DIR"))) {
        gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (desktopdir_button), getenv ("RFM_DESKTOP_DIR"));
    }
    if(getenv ("RFM_DESKTOP_IMAGE") && strlen (getenv ("RFM_DESKTOP_IMAGE"))) {
        gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (desktopimage_button), getenv ("RFM_DESKTOP_IMAGE"));
    }
    if(getenv ("RFM_DESKTOP_COLOR") && strlen (getenv ("RFM_DESKTOP_COLOR"))) {
        GdkColor color;
        if(gdk_color_parse (getenv ("RFM_DESKTOP_COLOR"), &color)) {
            gtk_color_button_set_color ((GtkColorButton *) desktopcolor_button, &color);
        }
    }
    if(getenv ("RFM_ICONVIEW_COLOR") && strlen (getenv ("RFM_ICONVIEW_COLOR"))) {
        GdkColor color;
        if(gdk_color_parse (getenv ("RFM_ICONVIEW_COLOR"), &color)) {
            gtk_color_button_set_color ((GtkColorButton *) iconviewcolor_button, &color);
        }
    }

    update_tree (VERSION_CONTROL);
    update_tree (TERMINAL_CMD);
    update_tree (EDITOR_CMD);
    update_tree (RFM_DESKTOP_DIR);
    update_tree (RFM_DESKTOP_IMAGE);
    update_tree (RFM_DESKTOP_COLOR);

    disable_options = FALSE;

}

static void
option_toggled (GtkToggleButton * togglebutton, gpointer user_data) {
    if(disable_options)
        return;
    int i = (int)((long)user_data);

    if(rfm_options[i].value) {
        g_free (rfm_options[i].value);
    }
    if(gtk_toggle_button_get_active (togglebutton)) {
        rfm_options[i].value = g_strdup ("yes");
    } else {
        rfm_options[i].value = g_strdup ("");
    }
    mcs_manager_set_string (mcs_manager, rfm_options[i].name, CHANNEL, rfm_options[i].value);
    mcs_manager_notify (mcs_manager, CHANNEL);
    /*write_options(); */

}

static void
switch_page (GtkNotebook * notebook, GtkNotebookPage * page, guint page_num, gpointer user_data) {
    //    if (page_num < )
    {
        //DBG("pagenum=%d",page_num);
        set_option_buttons ();

    }
}

static void
path_changed (GtkComboBox * combo_box, gint id) {
    if(disable_options)
        return;
    gchar *path_selected = gtk_combo_box_get_active_text (combo_box);
    gchar *path = g_find_program_in_path (path_selected);
    if(!path) {
        gchar *m = g_strdup_printf ("%s (%s): %s", strerror (ENOEXEC), path_selected,
                                    strerror (ENOENT));
        rfm_confirm (NULL, GTK_MESSAGE_ERROR, m, NULL, NULL);
        g_free (m);
        return;
    }
    g_free (path);

    if(rfm_options[id].value)
        g_free (rfm_options[id].value);
    rfm_options[id].value = path_selected;

    mcs_manager_set_string (mcs_manager, rfm_options[id].name, CHANNEL, path_selected);
    mcs_manager_notify (mcs_manager, CHANNEL);
}

static void
terminal_changed (GtkComboBox * combo_box, gpointer user_data) {
    path_changed (combo_box, TERMINAL_CMD);
}

static void
editor_changed (GtkComboBox * combo_box, gpointer user_data) {
    path_changed (combo_box, EDITOR_CMD);
}

static void
combo_changed (GtkComboBox * combo_box, gpointer user_data) {
    if(disable_options)
        return;
    int i = (int)((long)VERSION_CONTROL);
    if(rfm_options[i].value)
        g_free (rfm_options[i].value);

    rfm_options[i].value = g_strdup (gtk_combo_box_get_active_text ((GtkComboBox *) combo_box));

    mcs_manager_set_string (mcs_manager, rfm_options[i].name, CHANNEL, rfm_options[i].value);
    mcs_manager_notify (mcs_manager, CHANNEL);
}

static void
folder_changed (GtkFileChooser * chooser, gpointer user_data) {
    if(disable_options)
        return;
    int i = (int)((long)user_data);
    if(rfm_options[i].value)
        g_free (rfm_options[i].value);
    rfm_options[i].value = gtk_file_chooser_get_filename (chooser);
    mcs_manager_set_string (mcs_manager, rfm_options[i].name, CHANNEL, rfm_options[i].value);
    mcs_manager_notify (mcs_manager, CHANNEL);
}

static void
color_changed (GtkColorButton * chooser, gpointer user_data) {
    if(disable_options)
        return;
    int i = (int)((long)user_data);
    if(rfm_options[i].value)
        g_free (rfm_options[i].value);
    GdkColor color;
    gtk_color_button_get_color ((GtkColorButton *) chooser, &color);
    rfm_options[i].value = gdk_color_to_string (&color);
    mcs_manager_set_string (mcs_manager, rfm_options[i].name, CHANNEL, rfm_options[i].value);
    mcs_manager_notify (mcs_manager, CHANNEL);
}
