/**************************************************************/
/////////////////////  4.6.2 ///////////////////////////////////
/**************************************************************/

#ifndef __RFM_TYPES_H__
#define __RFM_TYPES_H__
#include <string.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <semaphore.h>
#include <gmodule.h>

/* increment this MACRO whenever any change is done to structures
 * (to avoid incorrect memory accesses by modules)
 * Must be an integer. 
 * */
#define LIBRFM_SERIAL 157
#define  LIBRFM_MODULE \
void *module_sanity(void){ \
    return (GINT_TO_POINTER(LIBRFM_SERIAL)); \
}

typedef struct rfm_global_t
{
  gint 		argc; 		/* partial argument list */
  gchar 	**argv; 	/* partial argument list */
}
rfm_global_t;

/**************************************************************/
/////////////////////  record_entry_t //////////////////////////
/**************************************************************/

typedef struct record_entry_t	/* was: record_entry_t */ 
{
    gint type;			
    struct stat *st;
    gchar *mimetype;
    gchar *mimemagic;
    gchar *filetype;
    gchar *encoding;
    gchar *tag;
    gchar *path;
    const gchar *module;	/* module name, or NULL for local files */
    const gchar *parent_module;	/* parent module name, or NULL for local files */
}
record_entry_t;


/**************************************************************/
/////////////////////  xfdir_t /////////////////////////////////
/**************************************************************/

typedef struct dir_t
{
    gchar		*pathv;		/* matched pathname.  */
    record_entry_t	 *en;

}
dir_t;


typedef struct xfdir_t
{
    size_t 		pathc;		/* Count of paths matched */
    dir_t 		*gl;
    gchar		path[1024];
    void 		*data;
    gint		type;
    struct stat		st;
    off_t 		tama;
    guint		monitor_id;
    guint		view_id;
    //flags
    gboolean		reload;
    gboolean		invalidate_icons;
    gint		ascending;
    gint		sort_column;
    struct view_t	*view_p;
    GThread 		*thread;
    struct xfdir_t	*xfdir_p;

}
xfdir_t;

/**************************************************************/
/////////////////////   population_t ///////////////////////////
/**************************************************************/


typedef struct population_t {
    gint 		population_serial;
    struct view_t	*view_p;
    record_entry_t 	*en;
    gboolean 		preview_done;
    // deskview uses x,y and gridview uses row, column
    union {
	gint 		x;
	gint 		row;
    };
    union {
	gint 		y;
	gint 		column;
    };
    gint 		pixbufW;
    gint 		pixbufH;
    gint 		labelX;
    gint 		labelY;
    // this is a key value to keep threads in sync.
    gint 		serial;  

    GdkRectangle 	rect;    

    GdkPixbuf 		*pixbuf; // no reference here
    GdkPixbuf 		*base_pixbuf; // no reference here

    gboolean  		preview_busy;  // for tooltip and properties preview
    gboolean  		normal_tip; // preview was tried but not accomplished
    time_t    		preview_pixbuf_timestamp;
    GdkPixbuf 		*preview_pixbuf;  // referenced: for tooltip and properties preview
    GdkPixbuf 		*initial_preview_pixbuf;  // referenced: for tooltip and properties preview
    GdkPixbuf 		*normal_pixbuf; // reference owned by pixbuf hash
    GdkPixbuf 		*saturated_pixbuf; // referenced
    GdkPixbuf 		*selected_pixbuf; // reference not clear, probably not
    GdkPixbuf 		*cut_pixbuf; // referenced

    gboolean 		selected;
    PangoLayout 	*layout;
    PangoRectangle 	logical_rect;
    PangoLayout 	*layout2;
    PangoRectangle 	logical_rect2;
} population_t;


/**************************************************************/
/////////////////////  widgets_t ///////////////////////////////
/**************************************************************/

typedef struct widgets_t { 
    struct view_t	*view_p;		// view_t * parent

    GtkWidget 		*window; 		// main  window 
    GtkWidget 		*paper; 		// drawing area
    GtkWidget 		*diagnostics;		// diagnostics text area
    GtkWidget 		*diagnostics_window;	// window to place diagnostics
    GtkWidget 		*status;		// status text area
    GtkWidget 		*progress;		// load progress bar
    GtkWidget 		*remove;		// remove/overwrite dialog
    GtkWidget 		*popup;			// popup menu
    GtkWidget 		*rename;		// rename entry box
    GtkWidget 		*vpane;			// drawing/diagnostics pane
    GtkWidget 		*button_space;		// little button space
    GtkWidget 		*clear_button;		// clear text area button
    GtkWidget 		*navigation_popup;	// navigation popup menu
    GtkScrolledWindow 	*scrolled_window;	// viewport for drawing area

    // notebook is only for gridview
    GtkWidget 		*notebook; 		// notebook
    GtkWidget 		*page_child; 		// notebook page
    GtkWidget 		*page_icon; 		// notebook tab icon
    GtkWidget 		*page_label; 		// notebook tab label
    GtkWidget 		*page_button; 		// notebook remove button
    GtkWidget 		*page_menu_label; 	// notebook remove button

} widgets_t;


/**************************************************************/
/////////////////////   view_t /////////////////////////////////
/**************************************************************/

typedef struct view_t {
    gint 		serial;
    guint 		type;			// iconview type:
						// 1. DESKVIEW_TYPE 
						// 2. ICONVIEW_TYPE
    gint 		preferences;		//
    gchar		*workdir;		// current workdir
    gchar 		*desktop_dir;		// used by deskview
    record_entry_t 	*en;			// folder entry
    xfdir_t 		*xfdir_p;
    widgets_t 		widgets;
    
    gboolean 		monitor_enabled;
    gboolean 		protected;
    gboolean 		thumbnailer_active;
    gboolean 		thumbnailer_exit;
    gboolean		disable_toggle_refresh;	// flag for popup setup

    gboolean 		tooltip_active;		// tooltip status   
    record_entry_t	*tooltip_entry;  	// tooltip record_entry
    GdkPixbuf		*tooltip_pixbuf;	// tooltip image preview
    gchar 		*tooltip_text;		// tooltip status   

    gint 		paperX;			// configuration parameter
    gint 		paperY;			// configuration parameter
    gint 		single_click_mode_state;
    gint 		status;

    // only for desktop
    Pixmap 		background;
    GdkPixmap 		*pixmap;		// offscreen drawing area
   

    GList 		*selection_list;
    GList 		*go_list;
    GList 		*children;
    GList 		*sh_command;
    gint 		sh_command_counter;
    
    gint		exit_status;
    sem_t 		*population_semaphore; 
    gint		population_serial;
    gint		population_lock_count;
    GMutex		*population_serial_mutex;
    GMutex		*population_lock_mutex;

    gint 		monitor_id;
    GMutex		*monitor_id_mutex; 
    gboolean		refresh;
    
    // mutex for active xfdir structure:
    GMutex		*monitor_loop_mutex;

    GMutex		*thumbnail_loop_mutex; 

    GdkDragAction 	drag_action;
    gchar 		*drag_src;
    GdkGC 		*penGC;
    GdkColormap 	*cmap;
    GdkEventDND 	drag_event;
    GdkPixbuf 		*dnd_pixbuf;
    GtkTargetList	*target_list;

    guint32 		eventtime;  	/* eventtime for module popup menu */  
    long 		pasteboard_serial;
    gint 		icon_size;
    gint 		progress_count;
    gint 		progress_total;

    gchar 		*xbuffer; 
    gchar 		*deepest_dir;
	

    char 		*grid_flags;
    gint 		grid_extent;
    gint 		grid_rows;
    gint 		grid_columns;   
    gint 		grid_area;	// 
    gint 		max_elements;	//
    gint 		text_height;	//
    gint 		column_width;	//
    gint 		current_mouseX;	//
    gint 		current_mouseY;	//
    gint 		old_X;		//
    gint 		old_Y;		//
    gint 		boxX;		//
    gint 		boxY;		//
    gint 		tip_timer;	//
    gint 		sortcolumn;	//
    gint 		count; 		// to monitor load progress
    gint 		root_w;
    gint 		root_h;
    
    gdouble 		initial_x;
    gdouble 		mouseX;
    gdouble 		initial_y;
    gdouble 		mouseY;
    gdouble 		offset_x;
    gdouble 		offset_y;

    gboolean 		dragstate;
    gboolean 		redraw_pixmap; // redraw deskview buffer pixbuf

    population_t 	**population_pp;
    population_t 	*selected_p;
    population_t 	*doing_drag_p;
    population_t 	*saturated_p;
    population_t 	*label_p;	


    union { //either name is OK to use. same thing...
	const gchar *module_name;
	const gchar *module;
    };
    union { //either name is OK to use. same thing...
	const gchar *submodule_name;
	const gchar *submodule;
    };
        
    /* constructor functions: */
    struct view_t	*(*constructor)(record_entry_t *en); 		
    void 		(*child_constructor)(struct widgets_t *widgets_p, 
				    gchar *path);
    void 		(*popup_constructor)(struct widgets_t *widgets_p, 
				    gchar *path);
    void 		(*tab_constructor)(struct widgets_t *widgets_p, 
				    const gchar *path);

} view_t;
    

#endif
